package sos.util;


import java.util.*;
import java.io.*;



/**
 * Title: SOS File Utilities
 * Description: Helfer-Klassen zur Dateiverarbeitung
 * Copyright: Copyright (c) 2003
 * Company: SOS GmbH
 * @author <a href="mailto:ghassan.beydoun@sos-berlin.com">Ghassan Beydoun</a>
 * @version $Id: SOSFile.java 961 2005-06-06 09:23:25Z ap $
 */


public class SOSFile {
	
	/**
	 * Ersetzt einen Platzhalter in einer Dateispezifikation.
	 * @param  filespec Dateispezifikation, Platzhalter sind via [...] enthalten
	 * @param  substitute Subsitut des Platzhalters
	 * @return String Dateiname
	 */
	public static String subFileMask(String filespec, String substitute) throws IOException{
		
		if (filespec == null) {
			throw new IOException("file specification is null.");
		}
		
		String retVal = new String();
		int ipos1 = filespec.indexOf("[");
		int ipos2 = filespec.lastIndexOf("]");
		if ((ipos1 == -1) || (ipos2 == -1)){
			return filespec;
		}
		
		String midStr   = new String();
		String startStr = new String();
		String endStr   = new String();
		
		if (ipos1 != 0) {
			startStr = filespec.substring(0, ipos1);
		}
		
		midStr = substitute.concat(filespec.substring(ipos2 + 1, filespec.length()));
		retVal = startStr.concat(midStr).concat(endStr);
		return retVal;
	}
	
	
	/**
	 * liefert Dateiliste des eingegebenen Verzeichnis zurück.
	 * @return Vector Dateiliste
	 * @param regexp ein regul&auml;er Ausdruck
	 * @param flag ein Integer-Wert: CASE_INSENSITIVE, MULTILINE, DOTALL, UNICODE_CASE, and CANON_EQ
	 * @see <a href="http://java.sun.com/j2se/1.4.2/docs/api/constant-values.html#java.util.regex.Pattern.UNIX_LINES">Constant Field Values</a>
	 */
	public static Vector getFilelist(String folder, String regexp, int flag) throws Exception {
		
		Vector filelist = new Vector();
		
			if(folder == null || folder.length() == 0) throw new FileNotFoundException("empty directory not allowed!!");
			
			File f = new File(folder);
			if(!f.exists()) throw new FileNotFoundException("directory does not exist: " + folder);
			filelist        = new Vector();
			File[] files    = f.listFiles(new SOSFilelistFilter(regexp,flag));
			for(int i=0; i < files.length;i++) {
				if(files[i].isDirectory()) {
				} else if ( files[i].isFile()) {
					filelist.add(files[i]);
				} else {
					// unknown
				}
			}
		
		return filelist;
	}
	
	/**
	 * liefert Dateiliste des eingegebenen Verzeichnis zurück.
	 * @return Vector Dateiliste
	 * @param regexp ein regul&auml;er Ausdruck
	 * @param flag ein Integer-Wert: CASE_INSENSITIVE, MULTILINE, DOTALL, UNICODE_CASE, and CANON_EQ
	 * @see <a href="http://java.sun.com/j2se/1.4.2/docs/api/constant-values.html#java.util.regex.Pattern.UNIX_LINES">Constant Field Values</a>
	 */
	public static Vector getFolderlist(String folder, String regexp, int flag) throws Exception {
		
		Vector filelist = new Vector();
		
			if(folder == null || folder.length() == 0) throw new FileNotFoundException("empty directory not allowed!!");
			
			File f = new File(folder);
			if(!f.exists()) throw new FileNotFoundException("directory does not exist: " + folder);
			filelist        = new Vector();
			File[] files    = f.listFiles(new SOSFilelistFilter(regexp,flag));
			for(int i=0; i < files.length;i++) {
				if (!files[i].getName().equals(".") && !files[i].getName().equals(".."))
					filelist.add(files[i]);
			}
		
		return filelist;
	}
	
	/**
	 * 
	 * liefert Dateiliste des eingegebenen Verzeichnis zurück.
	 * @return Vector Dateiliste
	 * @param regexp ein regul&auml;er Ausdruck
	 * @param flag ein Integer-Wert: CASE_INSENSITIVE, MULTILINE, DOTALL, UNICODE_CASE, and CANON_EQ
	 * @param withSubFolder 
	 * @see <a href="http://java.sun.com/j2se/1.4.2/docs/api/constant-values.html#java.util.regex.Pattern.UNIX_LINES">Constant Field Values</a>
	 */
	public static Vector getFilelist(String folder, String regexp, int flag, boolean withSubFolder) throws Exception {
		Vector filelist = new Vector();
		File file = null;
		File[] subDir = null;
		
			file = new File(folder);
			subDir = file.listFiles(); 
			filelist.addAll(getFilelist(folder, regexp, flag));
			if (withSubFolder) {
				for (int i = 0; i < subDir.length; i++) {  			
					if (subDir[i].isDirectory()) {  					
						filelist.addAll(getFilelist(subDir[i].getPath(), regexp, flag, true)); 
					}
				}
			}
		return filelist;
	}
	
	/**
	 * 
	 * liefert Dateiliste des eingegebenen Verzeichnis zurück.
	 * @return Vector Dateiliste
	 * @param regexp ein regul&auml;er Ausdruck
	 * @param flag ein Integer-Wert: CASE_INSENSITIVE, MULTILINE, DOTALL, UNICODE_CASE, and CANON_EQ
	 * @param withSubFolder 
	 * @see <a href="http://java.sun.com/j2se/1.4.2/docs/api/constant-values.html#java.util.regex.Pattern.UNIX_LINES">Constant Field Values</a>
	 */
	public static Vector getFolderlist(String folder, String regexp, int flag, boolean withSubFolder) throws Exception {
		Vector filelist = new Vector();
		File file = null;
		File[] subDir = null;
		
			file = new File(folder);
			subDir = file.listFiles(); 
			filelist.addAll(getFolderlist(folder, regexp, flag));
			if (withSubFolder) {
				for (int i = 0; i < subDir.length; i++) {  			
					if (subDir[i].isDirectory()) {  					
						filelist.addAll(getFolderlist(subDir[i].getPath(), regexp, flag, true)); 
					}
				}
			}
		return filelist;
	}
	
	/**
	 * Kopiert eine Datei von Quelle zum Ziel.
	 * Die Parametern geben die absoluter Pfad und Dateiname an, 
	 * in dem die Datei kopiert werden sollen
	 * 
	 * @param dest  
	 * @param source
	 * @throws IOException
	 * @author mo
	 */
	public static boolean copyFile(String source, String dest) throws Exception {
		return copyFile(new File(source), new File(dest));
	}
	
	/**
	 * Kopiert eine Datei von Quelle zum Ziel
	 * 
	 * @param dest
	 * @param source
	 * @throws IOException
	 * @author mo
	 */
	public static boolean copyFile(File source, File dest) throws Exception {
		int size = (int) source.length();
		int bytes_read = 0;
		byte[] data=null;
		FileInputStream in=null;
		FileOutputStream out=null;
		boolean retVal = true;
		try {
			in   = new FileInputStream(source);
			out  = new FileOutputStream(dest,true);
			data = new byte[size];
			while (bytes_read < size)
				bytes_read += in.read(data, bytes_read, size-bytes_read);
			out.write(data);
			return retVal;
			
		} catch (Exception e) {
			throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
		} finally {														 
			if (in != null)  in.close();
			if (out != null) out.close();
		}
	}        
	
	/**
	 * Liest die Dateiinhalt aus und gibt diese als String
	 * 
	 * @param dest
	 * @param source
	 * @throws IOException
	 * @author mo
	 */
	public static String readFile(File source) throws Exception {
		int size 			= (int) source.length();
		int bytes_read 		= 0;
		byte[] data			= null;
		FileInputStream in	= null;
		String retsVal 		= "";		
		try {
			
			in   = new FileInputStream(source);
			data = new byte[size];
			while (bytes_read < size) 
				bytes_read += in.read(data, bytes_read, size-bytes_read);			
			retsVal = retsVal + new String(data);			
			return retsVal;
			
		} catch (Exception e) {
			throw new Exception ("\n -> ..error in " + SOSClassUtil.getMethodName() + " " + e);
		} finally {														 
			if (in != null)  in.close();			
		}
	}
	
	/**
	 * Bewegt eine Datei von Quelle zum Zielverzeichnis.
	 * 
	 * @param source
	 * @param dest
	 * @param sosLogger
	 * @return boolean
	 * @throws Exception
	 * @author mo
	 */
	public static boolean renameTo(String source, String dest, SOSStandardLogger sosLogger) throws Exception {
		return renameTo(new File(source), new File(dest), sosLogger);
	}
	
	/**
	 * Bewegt eine Datei von der Quelle zum Zielverzeichnis.
	 * @param dest
	 * @param source
	 * @throws IOException
	 * @author mo
	 */
	public static boolean renameTo(File source, File dest, SOSStandardLogger sosLogger) throws Exception {
		boolean retVal = true; 
		try {
			if (sosLogger != null) {
				sosLogger.debug1("..trying to move File " + source.getAbsolutePath() + " to " + dest.getAbsolutePath());
			}
			copyFile(source, dest);
			if (!source.delete()) {
				retVal = false;
			}
			return retVal;
		} catch (Exception e) {
			throw new Exception("\n ..error in " + SOSClassUtil.getMethodName() + " " + e);
		}		
	}
	
	/**
	 * �erprft, ob die Datei in von einem anderen Prozess verwendet wird.
	 * Wenn ja, dann versucht es nochmal nach einer Sekunde. Die Anzahl der 
	 * Versuche sind fnfmal.
	 * @author mo
	 */
	public static boolean canReadFile(File file, SOSStandardLogger sosLogger) throws Exception {       
		boolean rep = true;        
		int repeatCount = 5; // 5 mal versuchen		
		
		while (rep && repeatCount != -1) {
			try {                
				if (!file.exists()) {
					rep = false;
					throw new Exception("..file does not exist: " + file.getAbsolutePath());
				}
				java.io.FileOutputStream f = new java.io.FileOutputStream(file, true);
				f.close();
				rep = false;
			} catch (Exception e) {
				repeatCount--;
				if (repeatCount == -1) { 
					throw new Exception("..file cannot be accessed: " + file.getAbsolutePath() + ": " + e); }
				if (sosLogger != null) {
					sosLogger.debug1("trial " + (5 - repeatCount)
							+ " of 5 to access order file: "
							+ file.getAbsolutePath());
				}
				Thread.sleep(1000);
			}
		}
		return !rep;
	}
	
}
