// $Id: document.cxx 3546 2005-04-03 08:30:29Z jz $

#include "document_factory.h"


namespace sos {
namespace document_factory {

//-------------------------------------------------------------------------------------------static

Document::Class_descriptor Document::class_descriptor   ( &typelib, "sos.DocumentFactory.Document" );

const string subdocument_select_fields = "`NR`, `TEMPLATE`, `LANGUAGE`, `HEAD`";

//-------------------------------------------------------------------------------Document::Document

Document::Document( Template* template_ )
:
    Idispatch_implementation( &class_descriptor ),
    Registered( static_cast<Template_base*>( template_ )->_session ),
    _zero_(this+1),
    _template( template_ )
{
    init();
    _table = &_session->_documents_table;

    _new                            = true;
    _number_of_subdocuments_known   = true;
    _owner_id                       = _session->user()->id();
    _template_id                    = template_->id();

    register_object( _session->_settings.get_new_id( _session->_documents_table ) );
}

//-------------------------------------------------------------------------------Document::Document

Document::Document( Session* session, int id )
:
    Idispatch_implementation( &class_descriptor ),
    Registered( session ),
    _zero_(this+1)
{
    init();
    _table = &_session->_documents_table;

    hostole::Record record = _session->db_get_single_or_null( "select `TEMPLATE`, `USER`, `LANGUAGE`, `STATUS`, `MODEL`, `STATUS_TEXT`, `HEAD`, `FILENAME`, `TITLE`, `HAS_INCLUDES`  from LF_LETTERS  where `LETTER`=" + as_string(id) );
    if( !record )  throw_xc( "FACTORY-1009", id );

    _template_id      = record.as_int   ( "template"     );
    _owner_id         = record.as_int   ( "user"         );
    _language         = record.as_string( "language"     );
    _workflow_id      = record.as_int   ( "model"        );
    _state            = record.as_int   ( "status"       );
    _state_text       = record.as_string( "status_text"  );
    _head_id          = record.as_int   ( "head"         , 0 );
    _path             = record.as_string( "filename"     );
    _title            = record.as_string( "title"        );
    _has_subdocuments = record.as_bool  ( "has_includes" );
    _template = _session->_template_register[ _template_id ];       // Falls das Template schon geladen ist, sonst NULL.

    register_object( id );
}

//-------------------------------------------------------------------------------Document::Document
// Subdokument

Document::Document( Document* parent_document, Template* template_ )
:
    Idispatch_implementation( &class_descriptor ),
    Registered( parent_document->_session, &parent_document->_subdocuments_register ),
    _zero_(this+1),
    _template( template_ )
{
    init();
    _table                          = &_session->_subdocuments_table;
    _parent_document                = parent_document;   // Elterndokument setzt _parent_document auf NULL, wenn es geschlossen wird.

    _new                            = true;
    _number_of_subdocuments_known   = true;
    _owner_id                       = _session->user()->id();
    _template_id                    = template_->id();

    parent_document->number_of_subdocuments();     // Anzahl feststellen und _subdocuments anlegen
    parent_document->_has_subdocuments = true;
    parent_document->_subdocuments.push_back( this );
    _subdocument_index = (int)parent_document->_subdocuments.size();   // Eins hher als Index in _subdocuments.

    register_object( _subdocument_index );
    _parent_document = parent_document;   // Elterndokument setzt _parent_document auf NULL, wenn es geschlossen wird.
}

//-------------------------------------------------------------------------------Document::Document
// Subdokument

Document::Document( Document* parent_document, int index )
:
    Idispatch_implementation( &class_descriptor ),
    Registered( parent_document->_session, &parent_document->_subdocuments_register ),
    _zero_(this+1)
{
    init();
    _table = &_session->_subdocuments_table;
    _parent_document = parent_document;   // Elterndokument setzt _parent_document auf NULL, wenn es geschlossen wird.

    string select = S() << "select " << subdocument_select_fields << 
                           "  from " << _table->_table_name <<
                           "  where `LETTER`=" << parent_document->id() << 
                           " and  `NR`=" << index;

    hostole::Record record = _session->db_get_single_or_null( select );
    if( !record )  throw_xc( "FACTORY-1809", index );

    init_subdocument_from_record( record );
}

//-------------------------------------------------------------------------------Document::Document
// Subdokument

Document::Document( Document* parent_document, hostole::Record record )
:
    Idispatch_implementation( &class_descriptor ),
    Registered( parent_document->_session, &parent_document->_subdocuments_register ),
    _zero_(this+1)
{
    _table = &_session->_subdocuments_table;
    _parent_document = parent_document;   // Elterndokument setzt _parent_document auf NULL, wenn es geschlossen wird.
    init_subdocument_from_record( record );
}

//------------------------------------------------------------------------------Document::~Document

Document::~Document()
{
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << "()  " << obj_name() << "\n" );

    try
    {
        close();
    }
    catch( const exception& x )  { Z_LOG2( "factory",  __PRETTY_FUNCTION__ << ": " << x << "\n" ); }
}

//-----------------------------------------------------------------------------------Document::init

void Document::init()
{
    _language = "";   // Default: Ohne Sprache, damit Uwes AOK-Applikation es leichter hat (ist kompatibel zum alten Verfahren ohne Sprache)
    //TODO _language = _session->setting( ... );
}

//-----------------------------------------------------------Document::load_subdocument_from_record

void Document::init_subdocument_from_record( hostole::Record record )
{
    init();

    _subdocument_index = record.as_int( "nr" );
    _template_id    = record.as_int   ( "template"    );
    _owner_id       = _parent_document->_owner_id;
    _language       = record.as_string( "language"    );
    _workflow_id    = _parent_document->_workflow_id;
  //_state          = parent_document->_state;
  //_state_text     = record.as_string( "status_text" );
    _head_id        = record.as_int   ( "head"        , 0 );
  //_path           = record.as_string( "filename"    );
  //_title          = record.as_string( "title"       );

    _template = _session->_template_register[ _template_id ];       // Falls das Template schon geladen ist, sonst NULL.

    _parent_document->_has_subdocuments = true;
    if( (int)_parent_document->_subdocuments.size() < _subdocument_index )  _parent_document->_subdocuments.resize( _subdocument_index );
    _parent_document->_subdocuments[ _subdocument_index - 1 ] = this;

    register_object( _subdocument_index );
}

//----------------------------------------------------------------------------------Document::close

void Document::close()
{
    for( uint i = 0; i < _dialog_script_objects.size(); i++ )
    {
        if( _dialog_script_objects[ i ] )  
        {
            try
            {
                _dialog_script_objects[ i ].close();
            }
            catch( const exception& x )  { Z_LOG2( "factory",  __PRETTY_FUNCTION__ << ": " << x << "\n" ); }

            _dialog_script_objects[ i ] = NULL;
        }
    }

    _forms.clear();
    _form_ids.clear();

    Z_FOR_EACH( Postprocessing_order_registers, _postprocessing_order_registers, reg )  reg->second->close();
    _postprocessing_order_registers.clear();

    _subdocuments_register.close();

    unregister_object();
    _parent_document = NULL;  // Wenn vom Elterndokument gerufen: Verweis kappen!
}

//-------------------------------------------------------------------------------------Document::id

int Document::id() const
{
    if( is_subdocument() )  throw_xc( "FACTORY-1021", "id" );
    
    return Registered::id();
}

//-------------------------------------------------------------------------------Document::obj_name

string Document::obj_name() const
{ 
    S result;

    if( is_subdocument() )
    {
        result << "Subdokument ";
        if( _parent_document )  result << _parent_document->id() << "/";
        result << _subdocument_index << " ";
    }
    else  
    {
        result << "Dokument " << id() << " von ";
    }


    if( _template )  result << _template->obj_name();
               else  result << "Vorlage " << _template_id;

    int c = reduce( reduce_count_non_null, 0, _dialog_script_objects );
    if( c > 0 )  result << ", " << c << " Dialogskripte aktiv"; 

    c = reduce( reduce_count_non_null, 0, _subdocuments );
    if( c > 0 )  result << ", " << c << " Subdokumente geladen"; 


    return result;
}

//--------------------------------------------------------------------Document::parameters_filename

string Document::parameters_filename()
{ 
    S result;
    
    result << _session->_db_name << "-table=" << _table->_table_name << " -blob=parameters  where";
    
    if( is_subdocument() )
    {
        result << " `letter`=" << parent_document()->id();
        result << " and `nr`=" << _subdocument_index;
    }
    else
    {
        result << " `letter`=" << id(); 
    }

    return result;
}

//----------------------------------------------------------------------------------Document::store

void Document::store()
{
    try
    {
        Transaction ta ( _session );

            sql::Update_stmt stmt ( _table );

            //stmt.add_key( "letter" );

            stmt[ "template"     ] = _template_id;
            stmt[ "head"         ] = _head_id;
            stmt[ "language"     ] = _language;

            if( is_subdocument() )
            {
                stmt[ "letter"       ] = parent_document()->id();
                stmt[ "nr"           ] = _subdocument_index;
            }
            else
            {
                stmt[ "letter"       ] = id();
                stmt[ "model"        ] = _workflow_id;
                stmt[ "status"       ] = _state;
                stmt[ "status_text"  ] = _state_text;
                stmt[ "status_date"  ] = sql::Direct_value( "%now" );
                stmt[ "user"         ] = _owner_id;
                stmt[ "filename"     ] = _path;
                stmt[ "title"        ] = _title;
                stmt[ "has_includes" ] = _has_subdocuments;
            }

            _session->db_execute( _new? stmt.make_insert_stmt() : stmt.make_update_stmt() );


            if( _new )  parameters();   // Parameter noch nicht angerhrt? Dann Parameter initialisieren.

            if( _parameters )           // Nicht geladene Parameter speichern wir nicht. Die sind schon gespeichert.
            {
                //_session->insert_or_update_blob( blob_object_document, id(), blob_nr_document_parameters, _parameters.xml(), "", _new );
                hostole::File blob ( "-out -binary " + parameters_filename() );
                blob.put_line( _parameters.xml() );
                blob.close();
            }


            if( _delete_subdocuments  &&  _has_subdocuments  &&  !_new )
            {
                int count = _session->db_execute( S() << "DELETE from " << _session->_subdocuments_table._table_name << 
                                                        " where letter=" << id() );
                Z_LOG2( "factory", count << " Subdokumente gelscht" );

                _delete_subdocuments = false;
            }
        

            Z_FOR_EACH( vector< ptr<Document> >, _subdocuments, sub )  if( *sub )  (*sub)->store();


            store_postprocessing_orders( &ta );
        
        ta.commit();
        _new = false;
    }
    catch( Xc& x )
    {
        x.append_text( obj_name() );
        throw;
    }
}

//------------------------------------------------------------Document::store_postprocessing_orders

void Document::store_postprocessing_orders( Transaction* ta )
{
    Z_FOR_EACH( Postprocessing_order_registers, _postprocessing_order_registers, reg )
    {
        Z_FOR_EACH( Postprocessing_order::Register::Id_map, reg->second->_id_map, p )
        {
            p->second->store( ta );
        }
    }
}

//----------------------------------------------------------Document::postprocessing_order_register

Postprocessing_order_register* Document::postprocessing_order_register( const string& type_name )
{
    Postprocessing_order_registers::iterator it = _postprocessing_order_registers.find( type_name );
    if( it == _postprocessing_order_registers.end() )
    {
        ptr<Postprocessing_order_register> reg = Z_NEW( Postprocessing_order_register() );
        _postprocessing_order_registers[ type_name ] = reg;
        return reg;
    }
    else
        return it->second;
}

//---------------------------------------------------------------------------Document::real_head_id

int Document::real_head_id()
{
    return _head_id == 0                ? _template->_head_id :
           _head_id == use_users_head_id? owner()->_head_id :
           _head_id == use_no_head_id   ? 0
                                        : _head_id;
}

//--------------------------------------------------------------------------Document::head_filename

string Document::head_filename()
{ 
    return _session->blob_filename( blob_object_head, real_head_id(), blob_nr_template_text, "" ); 
}

//----------------------------------------------------------------------------------Document::owner

User* Document::owner()
{
    if( !_owner )  
    {
        _owner = _session->user( _owner_id );
    }

    return _owner;
}

//---------------------------------------------------------------------------Document::get_template

Template* Document::get_template()
{
    if( !_template )  
    {
        _template = _session->get_template( Variant( _template_id ) );
    }

    return _template;
}

//-----------------------------------------------------------------------------Document::parameters

hostole::Variables Document::parameters()
{
    if( !_parameters )
    {
        _parameters.create_instance();
        
        if( _new )
        {
            _parameters.set( "uservar", Variant( _session->user()->parameters().clone() ) );
        }
        else
        {
            Bstr xml_bstr;
            //_session->_hostware.file_as_string( Bstr( _session->blob_filename( id(), blob_object_document, blob_nr_document_parameters ) ), &xml_bstr );
            _session->_hostware.file_as_string( Bstr(parameters_filename()), &xml_bstr._bstr );
            if( xml_bstr.length() > 0 )  _parameters.set_xml( xml_bstr );
        }
    }

    return _parameters;
}

//--------------------------------------------------------------------------Document::dialog_script
/*
const string& Document::dialog_script()
{
    if( !_dialog_script_loaded )
    {
        _dialog_script = get_template()->dialog_script();
    }

    return _dialog_script;
}
*/
//-----------------------------------------------------------Document::dialog_script_object_or_null

Iscript_object* Document::dialog_script_object_or_null( int index )
{
    if( index >= (int)_dialog_script_objects_loaded.size()  ||  !_dialog_script_objects_loaded[ index ] )
    {
        ptr<Form>   form;
        int         n = 1;

        if( index != 0 )
        {
            if( _form_ids.size() > 0 )
            {
                n = 1 + (int)_form_ids.size();
                if( index < 0  ||  index >= n )  throw_xc( "FACTORY-1019", as_string(index) );

                form = _forms[ index - 1 ];
                if( !form )  form = _forms[ index - 1 ] = _session->form( _form_ids[ index - 1 ] );
            }
            else
            {
                n = 1 + get_template()->form_count();
                if( index < 0  ||  index >= n )  throw_xc( "FACTORY-1012", get_template()->_path, as_string(index) );

                form = get_template()->form( index - 1 );
            }

        }

        if( (int)_dialog_script_objects.size() < n )  
        {
            _dialog_script_objects       .resize( n );
            _dialog_script_objects_loaded.resize( n );
        }


        string script = index == 0? get_template()->dialog_script()
                                  : form->dialog_script();

        if( !script.empty() )
        {
            hostole::Script_object script_object;
            script_object.create_instance();
            script_object.set_language( _template->_script_language );

            _dialog_script_objects[ index ] = script_object;        // Vor parse(), denn parse() kann factory_document.dialog_script_object_or_null rufen.

            if( _session->_add_parameters_as_variables_in_dialog_scripts )  script_object.add_variables( parameters() );

            script_object.add_object( static_cast<Isession* >( _session ), "factory"          );
            script_object.add_object( static_cast<Idocument*>( this     ), "factory_document" );

            script_object.parse( _session->start_script( get_template()->_script_language ) );
            if( script_object.name_exists( "start_script" ) )  script_object.parse( "start_script()" );

            script_object.parse( script );
        }

        _dialog_script_objects_loaded[ index ] = true;
    }

    return _dialog_script_objects[ index ];     // NULL, wenn es kein Dialogskript gibt.
}

//-----------------------------------------------------------------Document::number_of_subdocuments

int Document::number_of_subdocuments()
{
    if( !is_subdocument()         // Ein Subdokument kann keine Subdokumente enthalten
     && _has_subdocuments )
    {
        if( !_number_of_subdocuments_known )
        {
            /* Nur die Anzahl bestimmen:
            string select = S() << "select max(`NR`)  from " << _session->_subdocuments_table._table_name << " where letter=" << id();
            Variant max_nr = _session->db_get_single_value( select );
            if( max_nr.is_null_or_empty_string() )  max_nr = 0;

            int count = max_nr.as_int();
            if( count )  _has_subdocuments = false;      // korrigieren

            _subdocuments.resize( count );
            */


            // Wir lesen gleich alle Subdokumente ein. Wir werden sie sowieso brauchen.


            string select = S() << "select " << subdocument_select_fields << 
                                   "  from " << _session->_subdocuments_table._table_name <<
                                   "  where `LETTER`=" << id() << 
                                   "  order by `NR`";

            hostole::File result_set ( "-in " + _session->_db_name + select );
            
            if( result_set.eof() )
            {
                _has_subdocuments = false;
            }
            else
            {
                while( !result_set.eof() )
                {
                    Z_NEW( Document( this, result_set.get() ) );   // Trgt sich selbst in _subdocuments ein.
                }

            }

            _number_of_subdocuments_known = true;

            for( uint i = 0; i < _subdocuments.size(); i++ )  if( !_subdocuments[i] )  throw_xc( "FACTORY-1809", i+1 );
        }
    }

    return (int)_subdocuments.size();
}

//----------------------------------------------------------------------------Document::subdocument

Document* Document::subdocument( int index )
{
    if( index < 1  ||  index > number_of_subdocuments() )  throw_xc( "FACTORY-1020", index, number_of_subdocuments() );

    Document* result = _subdocuments.at( index - 1 );

    if( result == NULL )
    {
        result = _subdocuments.at( index - 1 ) 
               = Z_NEW( Document( this, index ) );
    }

    return result;
}

//------------------------------------------------------------------------Document::parent_document

Document* Document::parent_document()
{
    if( !is_subdocument() )  throw_xc( "FACTORY-1022" );
    if( !_parent_document )  throw_xc( "FACTORY-1023" );

    return _parent_document;
}

//----------------------------------------------------------------Document::get_Number_of_documents

STDMETHODIMP Document::get_Number_of_documents( int* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        *result = number_of_subdocuments();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//----------------------------------------------------------------------------------Document::Close

STDMETHODIMP Document::Close()
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        close();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//---------------------------------------------------------------------------Document::get_Template

STDMETHODIMP Document::get_Template( Itemplate** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        *result = get_template();
        if( *result )  (*result)->AddRef();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//------------------------------------------------------------------------Document::get_Template_id
/*
STDMETHODIMP Document::get_Template_id( int** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;
    return E_NOTIMPL;

    HRESULT hr = S_OK;

    try
    {
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}
*/
//------------------------------------------------------------------------------Document::get_Owner

STDMETHODIMP Document::get_Owner( Iuser** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        *result = owner();
        if( *result )  (*result)->AddRef();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------Document::get_Parameters

STDMETHODIMP Document::get_Parameters( Ivariables2** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        parameters();
        *result = _parameters.copy();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//----------------------------------------------------------------------Document::putref_Parameters

STDMETHODIMP Document::putref_Parameters( Ivariables2* parameters )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    
    HRESULT hr = S_OK;
    
    try
    {
        if( !parameters )  return E_POINTER;        // Das wrde die Parameter aus der Datenbank neu lesen lassen.
        _parameters = parameters;        
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    
    return hr;
}

//--------------------------------------------------------------------------Document::put_Parameter

STDMETHODIMP Document::put_Parameter( BSTR name, VARIANT* value )
{
    Z_LOGI2( "factory.parameter", __PRETTY_FUNCTION__ << "(" << name << ")\n" );
    
    HRESULT hr = S_OK;
    
    try
    {
        parameters().set( name, *value );    
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    
    return hr;
}

//--------------------------------------------------------------------------Document::get_Parameter

STDMETHODIMP Document::get_Parameter( BSTR name, VARIANT* result )
{
    Z_LOGI2( "factory.parameter", __PRETTY_FUNCTION__ << "(" << name << ")\n" );
    
    HRESULT hr = S_OK;
    
    try
    {
        parameters().get( name, result );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    
    return hr;
}

//------------------------------------------------------------------------------Document::put_Param
// Nur fr AOK (10.2004)

STDMETHODIMP Document::put_Param( BSTR name, VARIANT* value )
{ 
    if( value->vt == VT_EMPTY )  // Fr AOK Bremen soll VT_EMPTY wie "" wirken (fr ur 27.10.04)
    {
        Variant empty_string = "";
        return put_Parameter( name, &empty_string );
    }
    else
    {
        return put_Parameter( name, value ); 
    }
}

//---------------------------------------------------------------------------Document::put_Filename

STDMETHODIMP Document::put_Filename( BSTR filename )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << filename << ")\n" );

    HRESULT hr = E_NOTIMPL;

/*
    try
    {
        if( is_subdocument() )  throw_xc( "FACTORY-1021" );

        //_path = ...;
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
*/

    return hr;
}

//----------------------------------------------------------------------------------Document::Store

STDMETHODIMP Document::Store()
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        store();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------Document::get_Dialog_script_object_or_null

STDMETHODIMP Document::get_Dialog_script_object_or_null( int index, Iscript_object** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << index << ")\n" );
    
    HRESULT hr = S_OK;
    
    try
    {
        *result = dialog_script_object_or_null( index );
        if( *result )  (*result)->AddRef();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    
    return hr;
}

//-----------------------------------------------------------------------Document::Create_processor
/* Vor Process() kann Create_processor() gerufen werden, 
   um ans Prozessor-Objekt ranzukommen.
   process() verwendet das Objekt und gibt zuerst die Referenz frei,
   damit im Fehlerfall kein Verweis von Document auf den Prozessor brigbleibt (also kein Zirkel).
   Der Aufrufer kann den Prozessor solange halten, wie er will.
*/
STDMETHODIMP Document::Create_processor( Ifactory_processor** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        _processor = NULL;
        _processor.create_instance();
        *result = _processor.copy();
    }
    catch( const exception& x )   { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}
//--------------------------------------------------------------------------------Document::Process

STDMETHODIMP Document::Process()
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    hostole::Factory_processor processor = _processor;
    _processor = NULL;

    try
    {
        get_template();


        // AUSGABEDATEI LSCHEN

        int err = unlink( _path.c_str() );
        if( err  &&  errno != ENOENT )  throw_errno( errno, _path.c_str() );


        // FACTORY PROCESSOR: ERSTE EINSTELLUNGEN

        if( !processor )  processor.create_instance();      // Wenn nicht mit Create_processor() angelegt

        processor.set_document_filename( _path );
        if( _has_subdocuments )
        {
            _session->_hostware.need_version( "1.6.111" );
            processor.set_merge_documents( true );
        }


        process_subdocument( processor );


        for( int i = 1; i <= number_of_subdocuments(); i++ )
        {
            subdocument( i )->process_subdocument( processor );
        }

        processor.close_output_file();
        processor.close();
    }
    catch( const exception& x )   { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    try
    {
        if( processor )  processor.close();
    }
    catch( exception& ) {}

    return hr;
}

//--------------------------------------------------------------------Document::process_subdocument
// Auch fr das Hauptdokument

void Document::process_subdocument( const hostole::Factory_processor& processor )
{
    if( get_template()->_text_type != "rtf" )  throw_xc( "FACTORY-1010", _template->_type );

    processor.set_language( get_template()->_script_language );
    processor.init_engine();
    processor.add_obj( static_cast<Isession* >( _session ), "factory"          );
    processor.add_obj( static_cast<Idocument*>( this     ), "factory_document" );
    processor.set_parameters( parameters().clone() );


    // BRIEFBOGEN

    if( real_head_id() == 0 )  // ohne Briefbogen
    {
        processor.set_head_filename( "" );
        processor.set_template_filename( get_template()->template_filename( _language ) );
    }
    else
    {
        processor.set_head_filename( head_filename() );
        processor.set_template_filename( get_template()->_path );
    }


    processor.add_parameters();     // Parameter fr Startskript und Vorlaufskript bernehmen.
                                    // Ab jetzt wirkt set_parameter() nicht mehr.


    // STARTSKRIPT

    processor.set_param( "-source-name=Startskript" );
    processor.parse( processor.preprocess( _session->start_script( processor.language() ) ) );

    if( processor.name_exists( "start_script" ) )  processor.parse( "start_script()" );


    // VORLAUFSKRIPT

    processor.set_param( "-source-name", "Vorlaufskript der Vorlage " + get_template()->_path );
    processor.parse( processor.preprocess( get_template()->forerun_script() ) );


    // AUFBEREITEN

    processor.set_param( "-source-name", "Vorlage " + _template->_path );
    processor.process();
}

//---------------------------------------------------------------------------------Document::get_Id

STDMETHODIMP Document::get_Id( int* result )
{ 
    //Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    
    HRESULT hr = S_OK;
    
    try
    {
        *result = id();  
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    
    return hr;
}

//--------------------------------------------------------------------------Document::Head_filename

STDMETHODIMP Document::Head_filename( VARIANT* path, BSTR* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << *path << ")\n" );
    
    HRESULT hr = S_OK;
    
    try
    {
        hr = String_to_bstr( _session->head( *path )->template_filename( _language ), result );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    
    return hr;
}

//----------------------------------------------------------------------Document::Template_filename

STDMETHODIMP Document::Template_filename( VARIANT* path, BSTR* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << *path << ")\n" );
    
    HRESULT hr = S_OK;
    
    try
    {
        hr = String_to_bstr( _session->get_template( *path )->template_filename( _language ), result );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    
    return hr;
}

//--------------------------------------------------------------------------Documnet::Text_filename

STDMETHODIMP Document::Text_filename( VARIANT* path, BSTR* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << *path << ")\n" );

    HRESULT hr = S_OK;

    try
    {
        hr = String_to_bstr( _session->text( *path )->template_filename( _language ), result );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//------------------------------------------------------------Document::Create_postprocessing_order

STDMETHODIMP Document::Create_postprocessing_order( BSTR type, Ipostprocessing_order** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << type << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        if( is_subdocument() )  throw_xc( "FACTORY-1021" );

        Postprocessing_order_type* t = _session->postprocessing_order_type( string_from_bstr( type ) );
        ptr<Postprocessing_order>  p = Z_NEW( Postprocessing_order( t, this ) );     // Registriert sich in _postprocessing_registers

        *result = p.take();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//---------------------------------------------------------------Document::get_postprocessing_order

STDMETHODIMP Document::get_Postprocessing_order( BSTR type, int number, Ipostprocessing_order** )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << type << "," << number << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        return E_NOTIMPL;
        if( is_subdocument() )  throw_xc( "FACTORY-1021" );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//----------------------------------------------------------------------------Document::Add_form_id

STDMETHODIMP Document::Add_form_id( int form_id )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << form_id << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        if( _dialog_script_objects.size() > 0  )  throw_xc( "FACTORY-1018" );

        _form_ids.push_back( form_id );
        _forms.push_back( NULL );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//---------------------------------------------------------------------------Document::Add_document

STDMETHODIMP Document::Add_new_document( VARIANT* template_name_or_id, Idocument** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << debug_string_from_variant( *template_name_or_id ) << "\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        if( is_subdocument() )  throw_xc( "FACTORY-1021" );

        ptr<Document> subdocument = Z_NEW( Document( this, _session->get_template( *template_name_or_id ) ) );
        // subdocument hat sich selbst in _subdocuments eingetragen.

        *result = subdocument.take();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//---------------------------------------------------------------------------Document::get_Document

STDMETHODIMP Document::get_Document( int index, Idocument** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << index << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        if( is_subdocument() )  throw_xc( "FACTORY-1021" );

        *result = subdocument( index );
        (*result)->AddRef();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-----------------------------------------------------------------------Document::Delete_documents

STDMETHODIMP Document::Delete_documents()
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        if( is_subdocument() )  throw_xc( "FACTORY-1021" );

        _subdocuments_register.close();
        _subdocuments.clear();

        _delete_subdocuments = true;        // store() soll lschen
        _number_of_subdocuments_known = true;
        _has_subdocuments = false;
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------------------------------

} //namespace document_factory
} //namespace sos


