// $Id: document_factory.cxx 3546 2005-04-03 08:30:29Z jz $

#include "document_factory.h"


#ifdef Z_MICROSOFT
#   include <crtdbg.h>
#   include <olectl.h>          // DllRegisterServer
#endif


using namespace zschimmer;
using namespace zschimmer::com;

//-------------------------------------------------------------------------------------------------

namespace sos {
namespace document_factory {

//-------------------------------------------------------------------------------------------static

Ihostware*                      hostware_ptr;               // Nicht Klasse Hostware, damit Destruktor und dadurch entladene hostole.dll nicht gerufen wird
zschimmer::Mutex                document_factory_mutex      ( "Document Factory" );
Typelib_ref                     typelib;
const char                      needed_hostware_version[]   = "1.6.97";


static Error_code_text error_codes[] =
{
    { "FACTORY-1001", "Datenbankname (db_name) fehlt" },
    { "FACTORY-1002", "Mandant, Benutzer oder Kennwort ist nicht bekannt" },
    { "FACTORY-1003", "Pfad fr $1 ist leer" },
    { "FACTORY-1004", "Pfad fr $2 $1 gibt es nicht" },
    { "FACTORY-1005", "Diese Version der Document Factory kann nur eine Session erzeugen" },
    { "FACTORY-1006", "$2-Pfad bezeichnet ein Verzeichnis: $1" },
    { "FACTORY-1007", "Parameter ist kein Pfad oder keine Kennung" },
    { "FACTORY-1008", "$2 ist nicht bekannt: $1" },
    { "FACTORY-1009", "Dokument $1 ist nicht bekannt" },
    { "FACTORY-1010", "Vorlagentyp $1 kann nicht von Document.process() verarbeitet werden" },
    { "FACTORY-1011", "ID $1 ist ungltig" },
    { "FACTORY-1012", "Vorlage $1 hat kein Formular Nummer $2" },
    { "FACTORY-1013", "Benutzer knnen nicht ber einen Pfad geladen werden" },
    { "FACTORY-1014", "Diese Methode gilt nicht fr echte Templates, nicht fr Head oder Text" },
    { "FACTORY-1015", "Postprocessing_order typ=$2 id=$3 fr Dokument $1 ist nicht bekannt" },
    { "FACTORY-1016", "Postprocessing_order typ=$1 nicht definiert" },
    { "FACTORY-1017", "Postprocessing_order: Tabelle $2 hat keine Spalte $1" },
    { "FACTORY-1018", "Document.Add_form() nach Dialog_script_object_or_null() nicht mglich" },
    { "FACTORY-1019", "Das Dokument hat nicht $1 Formulare" },
    { "FACTORY-1020", "Ungltiger Subdokumentindex $1. Das Dokument hat $2 Subdokumente" },
    { "FACTORY-1021", "Die Opertion kann nicht auf ein Subdokument ausgefhrt werden" },
    { "FACTORY-1022", "Das Dokument ist kein Subdokument" },
    { "FACTORY-1023", "Das bergeordnete Dokument ist nicht mehr geladen" },

    // Interne Fehler:
    { "FACTORY-1801", "Objekt-Kennung $1 ist doppelt" },
    { "FACTORY-1802", "UPDATE hat nicht genau einen Satz gendert, sondern $1: $2" },
    { "FACTORY-1803", "SELECT hat mehr als einen Satz geliefert: $1" },
    { "FACTORY-1804", "SELECT hat keinen Satz geliefert: $1" },
    { "FACTORY-1805", "Ungltiger Setting-Pfad $1" },
    { "FACTORY-1806", "Setting $1 fehlt" },
    { "FACTORY-1807", "Fehler beim Lesen des Integer-Setting $1: " },
    { "FACTORY-1808", "Fehler beim Lesen des Blob-Setting $1: " },
    { "FACTORY-1809", "In der Datenbank fehlt das Subdokument $1" },
    {}
};

//-------------------------------------------------------------------------------------------Z_INIT

Z_INIT( document_factory )
{
    add_error_code_texts( error_codes ); 
}

//------------------------------------------------------------------------------------------DllMain

#ifdef Z_COM
#   define DllMain document_factory_DllMain
#endif

extern "C" BOOL WINAPI DllMain( HANDLE hinstance, DWORD ul_reason_being_called, void* )
{
    BOOL result = TRUE;

    switch( ul_reason_being_called )
	{
		case DLL_PROCESS_ATTACH: 
        {
#           ifdef Z_MICROSOFT
                _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );   // Speicherprfung bei _DEBUG
#           endif

            zschimmer_init();

            typelib.set_hinstance( (HINSTANCE)hinstance );

            break;
        }


    	case DLL_PROCESS_DETACH: 
        {
            zschimmer_terminate();
            //if( hostware_ptr )  hostware_ptr->Release(),  hostware_ptr = NULL;    // Ist das gefhrlich? Kann die hostole.dll schon entladen sein? Ja.
            break;
        }


      //case DLL_THREAD_ATTACH : break;
      //case DLL_THREAD_DETACH : break;

/*
        case Z_DLL_COM_ATTACH: 
        {
            Com_module_params* params = (Com_module_params*)param;
            Log_ptr::set( params->_log_stream, params->_log_mutex );
            set_com_context( params->_com_context );
            break;
        }
*/
		default: break;
	} 

    return result;
}

//------------------------------------------------------------------------------------Init_hostware

static HRESULT Init_hostware( bool show_msg = false )
{
    HRESULT hr = S_OK;

    Z_MUTEX( document_factory_mutex )
    if( !hostware_ptr )
    {
        try 
        {
            hostole::Hostware hostware;
            
            if( show_msg )  
            {
                hostware.create_instance();         // Lst Exception COM-xxxx aus mit hr=DISP_E_EXCEPTION
            }
            else
            {
                hr = hostware.Create_instance();
                if( FAILED(hr) )  return hr;        // Damit HRESULT-Fehler zurckgegeben wird (DISP_E_EXCEPTION wird vom Aufrufer nicht ausgewertet)
            }

            hostware.use_log();
            hostware.need_version( needed_hostware_version );

            hostware_ptr = hostware.take();  // hostole_ptr halten, solange use_log() aktiv ist, also Log_ptr Verweise auf Hostole hat.
        }
        catch( const exception& x )
        { 
            hr = Com_set_error( x, "Hostware" ); 
            if( show_msg )  
            {
                string msg = "Zum Betrieb der DocumentFactory bitte HostOLE registrieren!\n\n";
                msg += x.what();
#               ifdef Z_WINDOWS
                    MessageBox( NULL, msg.c_str() , "DocumentFactory", 0 );
#               else
                    fprintf( stderr, "%s\n"; msg.c_str() );
#               endif
            }
        }
    }

    return hr;
}

//----------------------------------------------------------------------------DllGetClassObject

#ifdef Z_COM //COM_STATIC
#   define DllGetClassObject document_factory_DllGetClassObject
#endif

extern "C" HRESULT APIENTRY DllGetClassObject( const CLSID& clsid, const IID& iid, void** result )
{
    HRESULT hr = S_OK;
  //MessageBox( NULL, "Get_class_object", "DllGetClassObject", 0 );

    if( !hostware_ptr )  
    {
        hr = Init_hostware();
        if( FAILED(hr) )  return hr;
    }

  //MessageBox( NULL, "Get_class_object", "DocumentFactory", 0 );
    hr = typelib.Get_class_object( clsid, iid, result );
  //MessageBox( NULL, ("Get_class_object hr=" + string_from_hresult(hr)).c_str(), "DocumentFactory", 0 );

    return hr;
}

//------------------------------------------------------------------------------DllCanUnloadNow

extern "C" int WINAPI DllCanUnloadNow()
{
    return com_can_unload_now()? S_OK : S_FALSE;
}

//----------------------------------------------------------------------------DllRegisterServer
#ifdef SYSTEM_HAS_COM

extern "C" HRESULT WINAPI DllRegisterServer()
{
    if( !hostware_ptr )  
    {
        Init_hostware( true );      // Fehler wird gemeldet, aber ignorieren wir.
    }

    return typelib.Register_server();
}

//--------------------------------------------------------------------------DllUnregisterServer

extern "C" HRESULT WINAPI DllUnregisterServer()
{
    if( !hostware_ptr )  
    {
        Init_hostware( false );      // Fehler wird nicht gemeldet, und ignorieren wir.
    }

    return typelib.Unregister_server();
}

#endif
//-------------------------------------------------------------------------------------------------

} //namespace document_factory
} //namespace sos
