// $Id: factory.cxx 3546 2005-04-03 08:30:29Z jz $

#include "document_factory.h"
#include "version.h"
#include "../zschimmer/z_windows_registry.h"


namespace sos {
namespace document_factory {

//-------------------------------------------------------------------------------------------static

Factory::Class_descriptor       Factory::class_descriptor   ( &typelib, "sos.DocumentFactory" );

//---------------------------------------------------------------------------------Factory::Factory

Factory::Factory()
:  
    Creatable_idispatch_implementation( &class_descriptor ), 
    _zero_(this+1)
{
    Z_LOG( VER_FILEDESC_STR << " " << VER_PRODUCTVERSION_STR << "\n" );

    _configuration_path = "factory.ini";

#   ifdef Z_WINDOWS
    {
        windows::Registry_key key;
        if( key.try_open( HKEY_LOCAL_MACHINE, "Software\\SOS Software", KEY_QUERY_VALUE ) )  _configuration_path = key.get_string( "factory.ini", _configuration_path );

        char buffer[ 4000 ];
        int len = GetPrivateProfileString( "factory", "db", "", buffer, sizeof buffer, _configuration_path.c_str() );
        _db_name.assign( buffer, len );
    }
#   endif
}

//----------------------------------------------------------------------Factory::put_Log_categories

STDMETHODIMP Factory::put_Log_categories( BSTR categories )
{
    Z_LOGI( __PRETTY_FUNCTION__ << "(" << categories << ")\n" );

    HRESULT hr = S_OK;

    try
    {
        log_categories.set_multiple( string_from_bstr( categories ) );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-----------------------------------------------------------------------------Factory::put_Db_name

STDMETHODIMP Factory::put_Db_name( BSTR hostware_filename )
{
    Z_MUTEX( document_factory_mutex )
    {
        if( _session_count > 0 )  return E_ACCESSDENIED;            // Der Datenbankname ist schon verwendet worden.

        _db_name = string_from_bstr( hostware_filename ) + " ";
    }

    return S_OK;
}

//-----------------------------------------------------------------------------------Factory::Login

STDMETHODIMP Factory::Login( BSTR unit_bstr, BSTR user_bstr, BSTR password_bstr, Isession** result )
{
    HRESULT hr = S_OK;

    try
    {
        if( _session_count > 0 )  throw_xc( "FACTORY-1005" );

        _hostware.create_instance();

        ptr<Session> session = Z_NEW( Session( this,
                                               string_from_bstr( unit_bstr ),
                                               string_from_bstr( user_bstr ),
                                               string_from_bstr( password_bstr ) ) );
        *result = session.take();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------------------------------

} //namespace document_factory
} //namespace sos
