// $Id: postprocessing_order.cxx 3546 2005-04-03 08:30:29Z jz $

#include "document_factory.h"


namespace sos {
namespace document_factory {

//-------------------------------------------------------------------------------------------static

Postprocessing_order::Class_descriptor Postprocessing_order::class_descriptor   ( &typelib, "sos.DocumentFactory.Postprocessing_order" );

//-------------------------------------------------------Postprocessing_order::Postprocessing_order

Postprocessing_order::Postprocessing_order( Postprocessing_order_type* type, Document* document )
:
    Idispatch_implementation( &class_descriptor ),
    Registered( document->_session, document->postprocessing_order_register( type->_name ) ),
    _zero_(this+1),
    _type(type),
    _document(document)
{
    //Transaction ta ( _session );

    if( document->is_subdocument() )  throw_xc( "FACTORY-1021" );

    _new = true;

    _type->init();

    Postprocessing_order_register* reg = _document->postprocessing_order_register( type->_name );

    if( !reg->_new_id )
    {
        string select = S() << "select max(`NR`)  from " << _type->_table._table_name << " where letter=" << document->id();
        Variant max_id = _session->db_get_single_value( select );
        if( max_id.is_null_or_empty_string() )  max_id = 0;

        reg->_new_id = int_from_variant( max_id ) + 1;
    }

    int id = reg->_new_id++;

    //ta.commit();


    register_object( id );
}

//-------------------------------------------------------Postprocessing_order::Postprocessing_order

Postprocessing_order::Postprocessing_order( Postprocessing_order_type* type, Document* document, int id )
:
    Idispatch_implementation( &class_descriptor ),
    Registered( document->_session, document->postprocessing_order_register( type->_name ), id ),
    _zero_(this+1),
    _type(type),
    _document(document)
{
    if( document->is_subdocument() )  throw_xc( "FACTORY-1021" );

    string select = S() << "select *  from " << _type->_table._table_name << "  where `LETTER`=" << document->id() << " and nr=" << id;
    hostole::Record record = _session->db_get_single_or_null( select );
    if( !record )  throw_xc( "FACTORY-1009", as_string( document->id() ), _type->_name, as_string( id ) );

    for( int i = 0; i < record.field_count(); i++ )
    {
        _attributes[ lcase( record.field_name( i ) ) ] = record[i];         
    }

    register_object( id );
}

//------------------------------------------------------Postprocessing_order::~Postprocessing_order

Postprocessing_order::~Postprocessing_order()
{
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << "()  " << id() << "\n" );

    try
    {
        close();
    }
    catch( const exception& x )  { Z_LOG2( "factory",  __FUNCTION__ << ": " << x << "\n" ); }

    unregister_object();
    _document = NULL;
}

//----------------------------------------------------------------------Postprocessing_order::close

void Postprocessing_order::close()
{
    _type     = NULL;
}

//-------------------------------------------------------------------Postprocessing_order::obj_name

string Postprocessing_order::obj_name() const
{ 
    S text;

    text << "Postprocessing_order(";
    if( _type )  text << _type->_name;
    text << "," << id() << ")";

    return text;
}

//----------------------------------------------------------------------Postprocessing_order::store

void Postprocessing_order::store( Transaction* )
{
    try
    {
        sql::Update_stmt stmt ( &_type->_table );

        stmt[ "letter"  ] = _document->id();
        stmt[ "nr"      ] = id();
        stmt[ "created" ] = sql::Direct_value( "%now" );

        Z_FOR_EACH( Attributes, _attributes, a )
        {
            if( variant_is_integer( a->second ) )  stmt[ a->first ] = int_from_variant   ( a->second );
                                             else  stmt[ a->first ] = string_from_variant( a->second );
        }
        //stmt.add_key( "letter" );

        _session->db_execute( _new? stmt.make_insert_stmt() : stmt.make_update_stmt() );

        _new = false;
    }
    catch( Xc& x )
    {
        x.append_text( obj_name() );
        throw;
    }
}

//----------------------------------------------------------------------Postprocessing_order::Close

STDMETHODIMP Postprocessing_order::Close()
{
    HRESULT hr = S_OK;

    try
    {
        close();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//--------------------------------------------------------------Postprocessing_order::put_Attribute

STDMETHODIMP Postprocessing_order::put_Attribute( BSTR name_bstr, VARIANT* value )
{
    HRESULT hr = S_OK;

    try
    {
        string name = lcase( string_from_bstr( name_bstr ) );
        _type->check_attribute_name( name );

        _attributes[ name ] = *value;
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//--------------------------------------------------------------Postprocessing_order::get_Attribute

STDMETHODIMP Postprocessing_order::get_Attribute( BSTR name_bstr, VARIANT* result )
{
    HRESULT hr = S_OK;

/*
    if( !name_bstr )
    {
        VariantInit( result );
        result->vt = VT_BSTR;
        return get_Debug_state( V_BSTR( result ) );
    }
*/

    try
    {
        string name = lcase( string_from_bstr( name_bstr ) );
        _type->check_attribute_name( name );


        Attributes::iterator it = _attributes.find( name );
        if( it == _attributes.end() )
        {
            VariantInit( result );
        }
        else
        {
            hr = VariantCopy( result, &it->second );
        }
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//------------------------------------------------------------------Postprocessing_order_type::init

void Postprocessing_order_type::init()
{
    if( !_initialized )
    {
        string          select = S() << "select *  from " << _table._table_name << "  where 1=0";
        hostole::File   columns ( "-in columns | " + _session->_db_name + select );

        while( !columns.eof() )
        {
            hostole::Record record = columns.get();
            string name = lcase( record.as_string( "column_name" ) );
            _field_names.insert( name  );
        }

        _field_names.erase( "letter"  );
        _field_names.erase( "nr"      );
        _field_names.erase( "created" );

        _initialized = true;
    }
}

//--------------------------------------------------Postprocessing_order_type::check_attribute_name

void Postprocessing_order_type::check_attribute_name( const string& name )
{
    if( _field_names.find( name ) == _field_names.end() )  throw_xc( "FACTORY-1017", name, _table._table_name );
}

//-------------------------------------------------------------------------------------------------

} //namespace document_factory
} //namespace sos
