// $Id: session.cxx 3546 2005-04-03 08:30:29Z jz $

#include "document_factory.h"


namespace sos {
namespace document_factory {

//-------------------------------------------------------------------------------------------static

Session::Class_descriptor       Session::class_descriptor   ( &typelib, "sos.DocumentFactory.Session" );

//---------------------------------------------------------------------------------Session::Session

Session::Session( Factory* factory, const string& unit, const string& user, const string& password )
:
    Idispatch_implementation( &class_descriptor ),
    _zero_(this+1),
    _factory( factory ),
    _hostware( factory->_hostware ),
    _settings( this ),
    _db_descr( sql::flag_uppercase_names | sql::flag_quote_names | sql::flag_write_empty_as_null ),
    _head_table             ( &_db_descr, "LF_HEADS"          , "id"                         , "Briefbogen"   ),
    _template_table         ( &_db_descr, "LF_TEMPLATES"      , "template"                   , "Vorlage"      ),
    _text_table             ( &_db_descr, "LF_TEXTS"          , "id"                         , "Textbaustein" ),
    _form_table             ( &_db_descr, "LF_FORMS"          , "id"                         , "Formular"     ),
    _documents_table        ( &_db_descr, "LF_LETTERS"        , "letter"                     , "Dokument"     ),
    _blobs_table            ( &_db_descr, "LF_BLOBS"          , "id,object,nr"               , "Blob"         ),
    _settings_table         ( &_db_descr, "SETTINGS"          , "application,section,name"   , "Setting"      ),
    _subdocuments_table     ( &_db_descr, "LF_LETTER_INCLUDES", "letter,nr"                  , "Teildokument" )
{
    _settings._application_default = "document_factory";
    _document_register._unique = false;     // Dasselbe Dokument (mit gleicher ID) darf mehrmals geladen werden. Mit Load_document().

    // -session=ID in _db_name einfgen. ID wird mit einem Hostware-Aufruf generiert (ein Wunsch darf angegeben werden)
    // string session_id = _factory->_hostware->get_session_id( "document_factory" );
    _db_name = factory->_db_name + " ";

    if( rtrim( _db_name ).empty() )  throw_xc( "FACTORY-1001" );

    _db.create_instance();
    _db.open( "-in -out " + _db_name );
    _db.set_write_empty_as_null( _db_descr._write_empty_as_null );

    _user = this->user( unit, user, password );


    // POSTPROCESSING_ORDER_TYPES
    /*
    add_postprocessing_order_type( "print"  , "lf_prints"   );
    add_postprocessing_order_type( "mail"   , "lf_mails"    );
    add_postprocessing_order_type( "archive", "lf_archives" );

    // [postprocessing_order_types]
    // print = lf_prints
    // mail  = lf_emails

    hostole::File profile ( "-in profile -section=postprocessing_order " + _factory->_configuration_path );
    while( !profile.eof() )
    {
        hostole::Record record = profile.get();
        add_postprocessing_order_type( record.as_string( "entry" ), record.as_string( "value" ) );
    }
    */


    InterlockedIncrement( &_factory->_session_count );
}

//--------------------------------------------------------------------------------Session::~Session

Session::~Session()
{
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << "()\n" );

    if( _factory )  InterlockedDecrement( &_factory->_session_count );

    try
    {
        close();
    }
    catch( const exception& x )  { Z_LOG2( "factory", __FUNCTION__ << ": " << x << "\n" ); }
}

//-----------------------------------------------------------------------------------Session::close

void Session::close()
{
    _document_register.close();
    _form_register    .close();
    _head_register    .close();
    _template_register.close();
    _text_register    .close();
    _user = NULL;
    _user_register    .close();

    Z_FOR_EACH( Postprocessing_order_types, _postprocessing_order_types, t )  t->second->close();

    _db.close();
}

//-----------------------------------------------------------Session::add_postprocessing_order_type

Postprocessing_order_type* Session::add_postprocessing_order_type( const string& type_name, const string& table_name )
{
    sql::Table_descriptor table_descriptor ( &_db_descr, table_name, "letter,nr" );

    ptr<Postprocessing_order_type> t = Z_NEW( Postprocessing_order_type( this, type_name, table_descriptor ) );

    _postprocessing_order_types[ type_name ] = t;

    return t;
}

//------------------------------------------------------------------------------Session::id_of_path

int Session::id_of_path( const string& path, const Table_descriptor& t )
{
    if( path == "/" )  
    {
        return 0;
    }
    else
    {
        string my_path = path;

        //If IsNumeric(my_path) Then template_id_of_path = CLng(my_path): Exit Function

        if( my_path.empty() )  throw_xc( "FACTORY-1003", t._object_name );
        if( my_path[0] == '/' )  my_path.erase( 0, 1 );
        if( my_path.empty() )  throw_xc( "FACTORY-1003", t._object_name );

        string npath = normalized_path( my_path );

        string stmt = "select " + sql::quoted_name( t.key_name(0) ) + " from " + t._table_name + "  where `normalized_path`=" + sql::quoted( npath );
        hostole::Record record = db_get_single_or_null( stmt );
        if( !record )  throw_xc( "FACTORY-1004", my_path, t._object_name );

        return record.as_int(0);
    }
}

//------------------------------------------------------------------------------Session::id_of_path

int Session::id_of_path( const VARIANT& path, const Table_descriptor& t )
{
    int id = 0;

    if( path.vt == VT_BSTR )
    {
        id = id_of_path( string_from_variant( path ), t ); 
    }
    else
    if( variant_is_integer( path ) )
    {
        id = int_from_variant( path );
    }
    else
        throw_com( DISP_E_TYPEMISMATCH, "id_of_path" );

    return id;
}

//------------------------------------------------------------------------------Session::db_execute

int Session::db_execute( const string& stmt )
{
    try
    {
        _db.put_line( stmt );
    }
    catch( Xc& x )
    {
        x.append_text( stmt );
        throw;
    }

    return _db.row_count();
}

//-----------------------------------------------------------------------Session::db_execute_single

void Session::db_execute_single( const string& update_stmt )
{
    _db.put_line( update_stmt );
    if( _db.row_count() != 1 )  throw_xc( "FACTORY-1802", as_string( _db.row_count() ), update_stmt );
}

//---------------------------------------------------------------------------Session::db_get_single

hostole::Record Session::db_get_single( const string& select_stmt )
{
    hostole::Record result = db_get_single_or_null( select_stmt );
    if( !result )  throw_xc( "FACTORY-1804", select_stmt );

    return result;
}

//-------------------------------------------------------------------Session::db_get_single_or_null

hostole::Record Session::db_get_single_or_null( const string& select_stmt )
{
    hostole::File select ( "-in -seq " + _db_name + select_stmt );
    if( select.eof() )  return NULL;

    hostole::Record result = select.get();
    if( !select.eof() )  throw_xc( "FACTORY-1803", select_stmt );

    return result;
}

//-------------------------------------------------------------------------------Session::blob_name

string Session::blob_name( Blob_object_code blob_object_code, int id, Blob_nr_code blob_nr_code, string rest )
{
    S name;

    switch( blob_object_code )
    {
        case blob_object_template: name << "template";  break;
        case blob_object_text    : name << "text";      break;
        case blob_object_queue   : name << "queue";     break;
        case blob_object_printer : name << "printer";   break;
        case blob_object_query   : name << "query";     break;
        case blob_object_form    : name << "form";      break;
        case blob_object_document: name << "document";  break;
        case blob_object_head    : name << "head";      break;
        default:                   name << (int)blob_object_code;  break;
    }

    name << "/" << id << "/";

    switch( blob_nr_code )
    {
        case blob_nr_template_text          : name << "text";            break;
      //case blob_nr_head_head              : name << "text";            break;     == blob_nr_template_text
        case blob_nr_template_forerun_script: name << "forerun_script";  break;
        case blob_nr_template_dialog_script : name << "dialog_script";   break;
        case blob_nr_template_archivscript  : name << "archive_script";  break;
        
        case blob_nr_template_asp_main      : name << ( blob_object_code == blob_object_template? "asp_main" :
                                                        blob_object_code == blob_object_form?     "dialog_script" :
                                                                                                  as_string( blob_nr_code ) );
        case blob_nr_template_asp_help      : name << "asp_help";        break;
        case blob_nr_template_asp_js        : name << "asp_js";          break;
        case blob_nr_template_asp_mask      : name << "asp_mask";        break;
        case blob_nr_document_parameters    : name << "parameters";      break;

        // Fr blob_object_head:

        // Fr blob_object_form:
        //case blob_nr_form_dialog_script     : name << "dialog_script"

        default: name << (int)blob_nr_code;
    }

    if( rest != "" )  name << "/" << rest;

    return name;
}

//---------------------------------------------------------------------------Session::blob_filename

string Session::blob_filename( Blob_object_code blob_object_code, int id, Blob_nr_code blob_nr_code, string rest, bool commit )
{

    S where;
    where << " where `object`=" << blob_object_code;
    where <<   " and `id`="       << id;
    where <<   " and `nr`="       << blob_nr_code;

    if( rest != "" )
    {
        S name;
        name << blob_object_code << "/" << id << "/" << blob_nr_code;
        if( rest != "" )  name << "/" << rest;
        where << " and `name`="   << sql::quoted( blob_name( blob_object_code, id, blob_nr_code, rest ) );
    }


    return "-binary " + _use_cache_prefix + _db_name + 
           ( commit? " -commit" : "" ) + 
           " -table=LF_BLOBS -blob=VALUE" +
           where;
}

//-------------------------------------------------------------------Session::insert_or_update_blob

void Session::insert_or_update_blob( Blob_object_code o, int id, Blob_nr_code nr, const string& rest, const string& value, bool insert )
{
    if( insert )
    {
        sql::Insert_stmt insert ( &_blobs_table );

        insert[ "object" ] = o;
        insert[ "id"     ] = id;
        insert[ "nr"     ] = nr;
        insert[ "name"   ] = blob_name( o, id, nr, rest );

        db_execute( insert );
    }

    hostole::File blob ( "-out -binary " + blob_filename( o, id, nr, rest ) );
    blob.put_line( value );
    blob.close();
}

//---------------------------------------------------------------------------Session::blob_filename
/*
string Session::setting_filename( const string& name )
{
    return _use_cache_prefix + _db_name + " -table=LF_SETTINGS -blob=LONG_VALUE"
           "  where name=" + sql::quoted( name );
}
*/
//------------------------------------------------------------------------------------Session::user

ptr<User> Session::user( const VARIANT& id_or_path )
{
    ptr<User> user = _user_register[ id_or_path ];
    
    if( user == NULL )
    {
        user = Z_NEW( User( this, id_or_path, "", false ) );
    }

    return user;
}

//------------------------------------------------------------------------------------Session::user

ptr<User> Session::user( const string& unit, const string& name, const string& password )
{
    ptr<User> user;  // = _user_register[ id ];
    
    //if( user == NULL )
    {
        user = Z_NEW( User( this, unit, name, password ) );   // Muss ein neuer, noch nicht geladener Benutzer sein! (Sowieso, nach Login)
    }

    return user;
}

//------------------------------------------------------------------------------------Session::head

ptr<Head> Session::head( const VARIANT& path_or_id )
{
    ptr<Head> head = _head_register[ path_or_id ];
    
    if( head == NULL )
    {
        Z_LOG2( "factory", __PRETTY_FUNCTION__ << "( " << variant_type_name( path_or_id ) << ":" << path_or_id << " ) -> neu  " );
        head = Z_NEW( Head( this, path_or_id ) );
        Z_LOG2( "factory", __PRETTY_FUNCTION__ << " id=" << head->id() << "\n" );
    }
    else
        Z_LOG2( "factory", __PRETTY_FUNCTION__ << "( " << variant_type_name( path_or_id ) << ":" << path_or_id << " ) -> vorhanden  id=" << head->id() << "\n" );


    return head;
}

//----------------------------------------------------------------------------Session::get_template

ptr<Template> Session::get_template( const VARIANT& path_or_id )
{
    ptr<Template> templat = _template_register[ path_or_id ];
    
    if( templat == NULL )
    {
        Z_LOG2( "factory", __PRETTY_FUNCTION__ << "( " << variant_type_name( path_or_id ) << ":" << path_or_id << " ) -> neu  " );
        templat = Z_NEW( Template( this, path_or_id ) );
        Z_LOG2( "factory", __PRETTY_FUNCTION__ << " id=" << templat->id() << "\n" );
    }
    else
        Z_LOG2( "factory", __PRETTY_FUNCTION__ << "( " << variant_type_name( path_or_id ) << ":" << path_or_id << " ) -> vorhanden  id=" << templat->id() << "\n" );


    return templat;
}

//------------------------------------------------------------------------------------Session::text

ptr<Text> Session::text( const VARIANT& path_or_id )
{
    ptr<Text> text = _text_register[ path_or_id ];
    
    if( text == NULL )
    {
        text = Z_NEW( Text( this, path_or_id ) );
    }

    return text;
}

//------------------------------------------------------------------------------------Session::form

ptr<Form> Session::form( const VARIANT& path_or_id )
{
    ptr<Form> form = _form_register[ path_or_id ];
    
    if( form == NULL )
    {
        form = Z_NEW( Form( this, path_or_id ) );
    }

    return form;
}

//--------------------------------------------------------------------------------Session::document

ptr<Document> Session::document( int id )
{
    ptr<Document> document = _document_register[ id ];
    
    if( document == NULL )
    {
        document = Z_NEW( Document( this, id ) );
    }

    return document;
}

//----------------------------------------------------------------------------Session::start_script

string Session::start_script( const string& script_language )
{
    if( !_start_script_read )
    {
        try
        {
            _start_script = _settings.get_blob_or_default( "blobs/factory_start_script." + lcase( script_language ), "\x01" );
            
            if( _start_script == "\x01" )
            {
                _start_script = _settings.get_blob_or_default( "blobs/factory_start_script", "" );
            }

            _start_script_read = true;
        }
        catch( Xc& x ) { x.append_text( __FUNCTION__ );  throw; }
    }

    return _start_script;
}

//---------------------------------------------------------------Session::postprocessing_order_type

Postprocessing_order_type* Session::postprocessing_order_type( const string& type_name )
{
    Postprocessing_order_types::iterator it = _postprocessing_order_types.find( type_name );
    if( it == _postprocessing_order_types.end() )
    {
        return add_postprocessing_order_type( type_name, type_name );
        //throw_xc( "FACTORY-1016", type_name );
    }

    return it->second;
}

//-----------------------------------------------------------------------------------Session::Close

STDMETHODIMP Session::Close()
{
    Z_LOGI2( "factory",  "\n" << __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        close();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------Session::get_Debug_state

STDMETHODIMP Session::get_Debug_state( BSTR* result )
{
    HRESULT hr = S_OK;

    try
    {
        string text;

        text += "Geladene Benutzer:\n";
        text += _user_register.state_text();
        text += "\n";

        text += "Geladene Vorlagen:\n";
        text += _template_register.state_text();
        text += "\n";

        text += "Geladene Formulare:\n";
        text += _form_register.state_text();
        text += "\n";

        text += "Geladene Dokumente:\n";
        text += _document_register.state_text();
        text += "\n";

        return String_to_bstr( text, result );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//---------------------------------------------------------------------------Session::put_Use_cache

STDMETHODIMP Session::put_Use_cache( VARIANT_BOOL use_cache )
{
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << '(' << use_cache << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        _use_cache_prefix = use_cache? "cache | " : "";
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//---------------------------------------------------------------------------Session::get_Use_cache

STDMETHODIMP Session::get_Use_cache( VARIANT_BOOL* result )
{
    *result = !_use_cache_prefix.empty();
    return S_OK;
}

//-----------------------------------------------------------------------------Session::Flush_cache

STDMETHODIMP Session::Flush_cache()
{
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        hostole::File( "cache -flush-all" );

        _start_script_read = false;
        _start_script = "";

        Z_FOR_EACH( object_register<Template>::Id_map, _template_register._id_map, it )  it->second->flush_cache();
        Z_FOR_EACH( object_register<Form>    ::Id_map, _form_register    ._id_map, it )  it->second->flush_cache();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------Session::Head_id_of_path

STDMETHODIMP Session::Head_id_of_path( BSTR path_bstr, int* result )
{
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << "(" << path_bstr << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        *result = id_of_path( string_from_bstr( path_bstr ), _head_table );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//---------------------------------------------------------------------Session::Template_id_of_path

STDMETHODIMP Session::Template_id_of_path( BSTR path_bstr, int* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << path_bstr << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        *result = id_of_path( string_from_bstr( path_bstr ), _template_table );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------Session::Text_id_of_path

STDMETHODIMP Session::Text_id_of_path( BSTR path_bstr, int* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << path_bstr << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        *result = id_of_path( string_from_bstr( path_bstr ), _text_table );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------Session::Form_id_of_path

STDMETHODIMP Session::Form_id_of_path( BSTR path_bstr, int* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << path_bstr << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        *result = id_of_path( string_from_bstr( path_bstr ), _form_table );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//--------------------------------------------------------------------------------Session::get_User

STDMETHODIMP Session::get_User( VARIANT* id_or_name_or_none, Iuser** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << id_or_name_or_none << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        if( id_or_name_or_none->vt == VT_ERROR  &&  id_or_name_or_none->scode == DISP_E_PARAMNOTFOUND 
         || id_or_name_or_none->vt == VT_EMPTY )
        {
            *result = _user;
            if( *result )  (*result)->AddRef();
        }
        else
        {
            return E_NOTIMPL;
            //*result = user( *id_or_name_or_none ).take();
        }
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//----------------------------------------------------------------------------Session::get_Template

STDMETHODIMP Session::get_Template( VARIANT* path_or_id, Itemplate** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << *path_or_id << ")\n" );

    HRESULT hr = S_OK;

    try
    {
        *result = get_template( *path_or_id ).take();        
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//--------------------------------------------------------------------------------Session::get_Form

STDMETHODIMP Session::get_Form( VARIANT* path_or_id, Iform** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << *path_or_id << ")\n" );

    HRESULT hr = S_OK;

    try
    {
        *result = form( *path_or_id ).take();        
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//----------------------------------------------------------------------------Session::get_Document

STDMETHODIMP Session::get_Document( int document_id, Idocument** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << document_id << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        *result = document( document_id ).take();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//---------------------------------------------------------------------------Session::Load_document

HRESULT Session::Load_document( int id, Idocument** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << id << ")\n" );
    HRESULT hr = S_OK;

    try
    {
        // berschreibt in _document_register eingetragenes Dokument mit gleicher ID.

        ptr<Document> document = Z_NEW( Document( this, id ) );
        *result = document.take();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------Session::Create_document

STDMETHODIMP Session::Create_document( VARIANT* template_name_or_id, Idocument** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << *template_name_or_id << ")\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    ptr<Itemplate> templat;

    hr = get_Template( template_name_or_id, templat.pp() );
    if( FAILED(hr) )  return hr;

    hr = templat->Create_document( result );

/*
    try
    {
        _session->get_template( template_name_or_id )->
        int id = id_of_path( *template_name_or_id );

        ptr<Document> document = Z_NEW( Document( this ) );
        document->create( id );

        *result = document.take();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
*/
    return hr;
}

//----------------------------------------------------------------------------Session::Check_script

STDMETHODIMP Session::Check_script( BSTR script_bstr, BSTR language )
{
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << "()\n" );
    HRESULT hr = S_OK;

    try
    {
        hostole::Script_object script_object;
        script_object.create_instance();
        script_object.set_language( language );

        script_object.parse_only( script_bstr, scripttext_isvisible );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------Session::Normalized_path

STDMETHODIMP Session::Normalized_path( BSTR path, BSTR* result )
{ 
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << "()\n" );
    HRESULT hr = S_OK;

    try
    {
        return String_to_bstr( normalized_path( string_from_bstr( path ) ), result );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------------------------------

} //namespace document_factory
} //namespace sos
