// $Id: standard.cxx 3546 2005-04-03 08:30:29Z jz $

#include "document_factory.h"


#ifdef Z_MICROSOFT
#   include <crtdbg.h>
#   include <olectl.h>          // DllRegisterServer
#endif


using namespace zschimmer;
using namespace zschimmer::com;

//-------------------------------------------------------------------------------------------------

namespace sos {
namespace document_factory {

//-------------------------------------------------------------------------------------------static
/*
const extern Table_descriptor head_table      = { "lf_texts"    , "id"        , "Vorlagen"    };
const extern Table_descriptor template_table  = { "lf_templates", "template"  , "Vorlagen"    };
const extern Table_descriptor text_table      = { "lf_heads"    , "id"        , "Briefbogen"  };
const extern Table_descriptor form_table      = { "lf_forms"    , "id"        , "Formular"    };
const extern Table_descriptor documents_table = { "lf_letters"  , "letter"    , "Dokument"    };
*/
//---------------------------------------------------------------------------------------left_align

string left_align( int value, int length )
{
    return left_align( as_string( value ), length );
}

//---------------------------------------------------------------------------------------left_align

string left_align( const string& value, int length )
{
    if( (int)value.length() < length )
        return value + string( length - value.length(), ' ' );
    else
        return value;
}

//-------------------------------------------------------------------------Session::normalized_path

string normalized_path( const string& path )
{
    string result = lcase( path );

    for( uint i = 0; i < result.length(); i++ )
    {
        switch( result[i] )
        {
            case ' ':
            case '_':   result[i] = '-';    break;

            case '%':   result[i] = '@';    break;

            case '\\':  result[i] = '/';    break;

            default:;
        }
    }

    return result;
}

//--------------------------------------------------------------------------------------equal_paths

bool equal_paths( const string& path1, const string& path2 )
{
    return normalized_path( path1 ) == normalized_path( path2 );
}

//-------------------------------------------------------------------------Transaction::Transaction

Transaction::Transaction( Session* session )
: 
    _session(session), 
    _outer_transaction(session->_current_inner_transaction) 
{
    _session->_current_inner_transaction = this;
}

//------------------------------------------------------------------------Transaction::~Transaction

Transaction::~Transaction()
{ 
    try
    { 
        rollback(); 
    } 
    catch( exception& x ) { Z_LOG2( "factory", "ROLLBACK: " << x << "\n" ); } 
}

//------------------------------------------------------------------------------Transaction::commit

void Transaction::commit()
{ 
    _session->_current_inner_transaction = _outer_transaction;

    if( !_outer_transaction )
    {
        _session->_db.put_line( "COMMIT" );  
    }

    _session = NULL; 
}

//----------------------------------------------------------------------------Transaction::rollback

void Transaction::rollback()
{ 
    if( _session )  
    {
        _session->_current_inner_transaction = _outer_transaction;

        _session->_db.put_line( "ROLLBACK" );
        _session = NULL; 

        if( _outer_transaction )  throw_xc( "ROLLBACK", "Rollback in innerer Transaktion setzt die uere Transaktion zurck" );
    }
}

//----------------------------------------------------------------tring_from_file::String_from_file
/*
String_from_file::String_from_file( Session* s, const string& filename, const string& title ) 
: 
    _hostware(s->_hostware), 
    _filename(filename), 
    _read(false), 
    _title(title) 
{
}
*/
//----------------------------------------------------------------------------String_from_file::get
/*
const string& String_from_file::get()
{
    if( !_read )
    {
        try
        {
            _string = _hostware.file_as_string( _filename );
            _read = true;
        }
        catch( Xc& x )  { x.append_text( _title );  throw; }
    }

    return _string;
}
*/

//--------------------------------------------------------------------------Settings::path_to_names
/*
void Settings::path_to_names( const string& path, string* application, string* section, string* name )
{
    int n = split( "/", path, application, section, name );
    
    if( n < 3 )
    {
        *name       = *section;
        *section    = *application;
        *applicatin = _application_default;
    }
}
*/
//------------------------------------------------------------------------Settings::names_from_path

vector<string> Settings::names_from_path( const string& path )
{
    vector<string> result;

    vector_split( "/", path, 3, &result );
    
    if( result.size() < 3 )
    {
        result.resize( 3 );
        result[2] = result[1];
        result[1] = result[0];
        result[0] = _application_default;
    }

    if( result.size() != 3 )  throw_xc( "FACTORY-1805", path );

    return result;
}

//----------------------------------------------------------------------------------Settings::where

string Settings::where( const string& path )
{
    vector<string> names = names_from_path( path );

    return " where `application`=" + sql::quoted( names[0] ) +
             " and `section`="     + sql::quoted( names[1] ) +
             " and `name`="        + sql::quoted( names[2] );
}

//------------------------------------------------------------------------------------Settings::get

string Settings::get( const string& path )
{
    hostole::Record record = _session->db_get_single_or_null( "select `value`  from SETTINGS " + where( path ) );
    if( !record )  throw_xc( "FACTORY-1806", path );

    return record.as_string( 0 );
}

//--------------------------------------------------------------------------------Settings::get_int

int Settings::get_int( const string& path )
{
    try
    {
        hostole::Record record = _session->db_get_single_or_null( "select `value`  from SETTINGS " + where( path ) );
        if( !record )  throw_xc( "FACTORY-1806", path );

        return record.as_int( 0 );
    }
    catch( exception& x )
    {
        throw_xc( "FACTORY-1807", path, x.what() );
    }
}

//-----------------------------------------------------------------------Settings::get_with_default

string Settings::get_with_default( const string& path, const string& deflt )
{
    hostole::Record record = _session->db_get_single_or_null( "select `value`  from SETTINGS " + where( path ) );
    if( !record )  return deflt;

    return record.as_string( 0 );
}

//-------------------------------------------------------------------------------Settings::get_blob

string Settings::get_blob( const string& path )
{
    try
    {
        return _session->_hostware.file_as_string( _session->_db_name + "-table=SETTINGS -blob=long_value " + where( path ) );
    }
    catch( exception& x )
    {
        throw_xc( "FACTORY-1808", path, x.what() );
    }
}

//--------------------------------------------------------------------Settings::get_blob_or_default

string Settings::get_blob_or_default( const string& path, const string& deflt )
{
    try
    {
        return _session->_hostware.file_as_string( _session->_db_name + "-table=SETTINGS -blob=long_value " + where( path ) );
    }
    catch( Xc& x )
    {
        if( x.name() == "NOTEXIST" )  return deflt;
        throw_xc( "FACTORY-1808", path, x.what() );
    }
}

//---------------------------------------------------------------------------------Settings::insert

void Settings::insert( const string& path, const string& value )
{
    Transaction ta ( _session );

    vector<string> names = names_from_path( path );

    sql::Insert_stmt insert ( &_session->_settings_table );

    insert[ "application" ] = names[0];
    insert[ "section"     ] = names[1];
    insert[ "name"        ] = names[2];
    insert[ "value"       ] = value;

    insert[ "forced"      ] = 0;
    insert[ "created"     ].set_direct( "%now" );
    insert[ "created_by"  ] = "DocumentFactory";
    insert[ "modified"    ].set_direct( "%now" );
    insert[ "modified_by" ] = "DocumentFactory";

    _session->db_execute( insert );

    ta.commit();
}

//-----------------------------------------------------------------------------Settings::get_new_id

int Settings::get_new_id( const Table_descriptor& table )
{
    Transaction ta ( _session );

    if( table._key_names.size() != 1 )  throw_xc( "get_new_id", table._table_name );

    string path = _application_default + "/counter/" + lcase( table._table_name );  // Tabellenname soll kleingeschrieben sein, fr Uwe, 2.11.04
    string where = this->where( path );
    string stmt  = "UPDATE  SETTINGS  set `value`=`value`+1 " + where;

    int row_count = _session->db_execute( stmt );

    if( row_count == 0 )
    {
        Variant max_id = _session->db_get_single_value( "select max(" + sql::quoted_name(table.key_name(0)) + ") from " + table._table_name );
        if( max_id.is_null_or_empty_string() )  max_id = 0;
        insert( path, max_id.as_int() + 1 );
    }

    int id = get_int( path );

    ta.commit();

    return id;
}

//-------------------------------------------------------------------------------------------------

} //namespace document_factory
} //namespace sos
