// $Id: template_base.cxx 3546 2005-04-03 08:30:29Z jz $

#include "document_factory.h"


namespace sos {
namespace document_factory {

//-------------------------------------------------------------------------------------------static

Template_base::Class_descriptor      Template_base::class_descriptor  ( &typelib, "sos.DocumentFactory.Template" );

//---------------------------------------------------------------------Template_base::Template_base

Template_base::Template_base( Session* session, Template_kind kind, const Table_descriptor* table_description, const VARIANT& id_or_path )
:
    Idispatch_implementation( &class_descriptor ),
  //Registered( session ),
    _zero_(this+1),
    _session(session),
    _template_kind(kind),
    _table_description( table_description )
{
    string id_name = _table_description->key_name( 0 );

    string stmt = "select `" + id_name + "`, `path`, `leaf`, `title`, `type`";
    if( _template_kind == template_kind_template )  stmt += ", `head`";
    stmt += "  from " + string(_table_description->_table_name) + "  where ";

    if( is_path( id_or_path ) )  stmt += "`normalized_path`=" + sql::quoted( normalized_path( string_from_variant( id_or_path ) ) );
    else
    if( is_id  ( id_or_path ) )  stmt += "`template`=" + string_from_variant( id_or_path );
    else 
        throw_xc( "FACTORY-1007" );

    hostole::Record record = _session->db_get_single_or_null( stmt );
    if( !record )  throw_xc( "FACTORY-1008", string_from_variant( id_or_path ), "Vorlage" );

    _id      = record.as_int   ( id_name );
    _path    = record.as_string( "path"     );
    _title   = record.as_string( "title"    );
    _type    = record.as_string( "type"     );

    if( _template_kind == template_kind_template )  
    _head_id = record.as_int   ( "head"     , 0 );

    _normalized_path = normalized_path( _path );
    split( "/", _type, &_text_type, &_script_language );
    if( _type == "frame/host" )  _script_language = "VBScript";    // Dialogskripte sind in VBScript codiert, fr Uwe 12.8.04

    if( !record.as_bool( "leaf" ) )  throw_xc( "FACTORY-1006", _path, "Vorlagen" );
}

//--------------------------------------------------------------------Template_base::~Template_base

Template_base::~Template_base()
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()  " << _path << "\n" );

    try
    {
        close();
    }
    catch( const exception& x )  { Z_LOG2( "factory",  __FUNCTION__ << ": " << x << "\n" ); }
}

//-----------------------------------------------------------------------------Template_base::close

void Template_base::close()
{
}

//--------------------------------------------------------------------------Template_base::obj_name

string Template_base::obj_name() const
{ 
    string result = "Vorlage " + as_string( id() ) + " " + _path;
    return result;
}

//--------------------------------------------------------------------Template_base::forerun_script

const string& Template_base::forerun_script()
{
    if( !_forerun_script_read )
    {
        try
        {
            _forerun_script = _session->string_from_blob_or_empty( blob_object_template, id(), blob_nr_template_forerun_script, "" );
            _forerun_script_read = true;
        }
        catch( Xc& x )  { x.append_text( __FUNCTION__ );  throw; }
    }

    return _forerun_script;
}

//------------------------------------------------------------------------------Template_base::form

Form* Template_base::form( int index )
{
    if( !_form_ids_loaded )
    {
        load_form_ids();
    }

    if( index <= 0  ||  (uint)index >= _form_ids.size() ||  _form_ids[index] == 0 )  throw_xc( "FACTORY-1012", _path, as_string(index) );

    int form_id = _form_ids[ index ];
    if( form_id == 0 )  throw_xc( "FACTORY-1012", _path, as_string(index) );

    if( !_forms_loaded[ index ] )
    {
        _forms[ index ] = _session->form( form_id );
        _forms_loaded[ index ] = true;
    }

    Form* form = _forms[ index ];
    if( !form )  throw_xc( "FACTORY-1012", _path, as_string(index) );   // Sollte nicht passieren.

    return form;
}

//------------------------------------------------------------------------------Template_base::form

void Template_base::load_form_ids()
{
    if( !_form_ids_loaded )
    {
        try
        {
            string stmt = "select `nr`, `form`  from LF_TEMPLATE2FORMS"
                        "  where `template`=" + as_string( id() ) +
                        "  order by `nr` desc";

            hostole::File select ( "-in -seq " + _session->_db_name + stmt );

            while( !select.eof() )
            {
                hostole::Record record = select.get();
                int nr = record.as_int( "nr" );
                if( _form_ids.size() <= (uint)nr )  _form_ids.resize( nr+1 );
                _form_ids[ nr ] = record.as_int( "form" );
            }

            select.close();


            _forms       .resize( _form_ids.size() );
            _forms_loaded.resize( _form_ids.size() );

            _form_ids_loaded = true;
        }
        catch( Xc& x )  { x.append_text( __FUNCTION__ );  throw; }
    }
}

//---------------------------------------------------------------------Template_base::dialog_script

const string& Template_base::dialog_script()
{
    if( !_dialog_script_read )
    {
        try
        {
            _dialog_script = _session->string_from_blob_or_empty( blob_object_template, id(), blob_nr_template_dialog_script, "" );
            _dialog_script_read = true;
        }
        catch( Xc& x )  { x.append_text( __FUNCTION__ );  throw; }
    }

    return _dialog_script;
}

//-----------------------------------------------------------------------Template_base::flush_cache

void Template_base::flush_cache()
{
    _dialog_script_read = false;
    _dialog_script = "";

    _forerun_script_read = false;
    _forerun_script = "";
}

//-----------------------------------------------------------------------------Template_base::Close

STDMETHODIMP Template_base::Close()
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    
    HRESULT hr = S_OK;

    try
    {
        close();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------Template_base::get_Owner
/*
STDMETHODIMP Template_base::get_Owner( Iuser** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    return E_NOTIMPL;

    HRESULT hr = S_OK;

    try
    {

    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}
*/

//--------------------------------------------------------------------------Template_base::get_Form

STDMETHODIMP Template_base::get_Form( int index, Iform** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "(" << index << ")\n" );
    
    HRESULT hr = S_OK;
    
    try
    {
        if( _template_kind != template_kind_template )  throw_xc( "FACTORY-1014" );

        *result = form( index );
        if( *result )  (*result)->AddRef();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    
    return hr;
}

//----------------------------------------------------------------Template_base::get_Forerun_script

STDMETHODIMP Template_base::get_Forerun_script( BSTR* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );

    HRESULT hr = S_OK;

    try
    {
        if( _template_kind != template_kind_template )  throw_xc( "FACTORY-1014" );

        hr = String_to_bstr( forerun_script(), result );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//------------------------------------------------------------------Template_base::get_Dialog_script

STDMETHODIMP Template_base::get_Dialog_script( BSTR* result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );

    HRESULT hr = S_OK;

    try
    {
        if( _template_kind != template_kind_template )  throw_xc( "FACTORY-1014" );

        hr = String_to_bstr( dialog_script(), result );
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//--------------------------------------------------------------------Template_base::Create_document

STDMETHODIMP Template_base::Create_document( Idocument** result )
{
    Z_LOGI( __PRETTY_FUNCTION__ << "()\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = S_OK;

    try
    {
        if( _template_kind != template_kind_template )  throw_xc( "FACTORY-1014" );

        ptr<Document> document = Z_NEW( Document( DYNAMIC_CAST( Template*, this ) ) );

        *result = document.take();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//----------------------------------------------------------------Template_base::blob_filename

string Template_base::blob_filename( Blob_nr_code nr, string language, bool commit )
{ 
    Blob_object_code o;
    
    switch( _template_kind )
    {
        case template_kind_head:        o = blob_object_head;      break;
        case template_kind_template:    o = blob_object_template;  break;
        case template_kind_text:        o = blob_object_text;      break;
        default:                        throw_xc( __FUNCTION__, (int)_template_kind );
    }

    return _session->blob_filename( o, id(), nr, language, commit ); 
}

//-------------------------------------------------------------------------------------------------

} //namespace document_factory
} //namespace sos
