// $Id: user.cxx 3546 2005-04-03 08:30:29Z jz $

#include "document_factory.h"


namespace sos {
namespace document_factory {

//-------------------------------------------------------------------------------------------static

User::Class_descriptor          User::class_descriptor  ( &typelib, "sos.DocumentFactory.User" );

//---------------------------------------------------------------------------------------User::User

User::User( Session* session, const string& unit, const string& name, const string& password )
:  
    Idispatch_implementation(&class_descriptor), 
    Registered(session), 
    _zero_(this+1)
{
    load( "where `name`=" + sql::quoted(name) + 
           " and `unit`=" + sql::quoted(unit), password );
}

//---------------------------------------------------------------------------------------User::User

User::User( Session* session, const VARIANT& id_or_path, const string& password, bool check_password )
:  
    Idispatch_implementation(&class_descriptor), 
    Registered(session), 
    _zero_(this+1)
{
    if( !variant_is_integer( id_or_path ) )  throw_xc( "FACTORY-1013" );

    load( "where `id`=" + string_from_variant( id_or_path ), password, check_password );
}

//--------------------------------------------------------------------------------------User::~User

User::~User()
{
    Z_LOGI2( "factory",  __PRETTY_FUNCTION__ << "()  " << id() << "\n" );

    try
    {
        close();
    }
    catch( const exception& x )  { Z_LOG2( "factory",  __PRETTY_FUNCTION__ << ": " << x << "\n" ); }
}

//--------------------------------------------------------------------------------------User::close

void User::close()
{
}

//-----------------------------------------------------------------------------------User::obj_name

string User::obj_name() const
{ 
    string result = "Benutzer " + left_align( id(), 6 ) + " " + left_align( _unit + "/" + _name, 20 ) + " " + _full_name;
    return result;
}

//---------------------------------------------------------------------------------------User::User

void User::load( const string& where, const string& password, bool check_password )
{
    string stmt = "select `id`, `unit`, `parent`, `name`, `full_name`, `head`, `password`  from USERS  " + where;
    
    hostole::Record record = _session->db_get_single_or_null( stmt );
    
    if( !record )  throw_xc( "FACTORY-1002" );
    
    if( check_password )
    {
        string md5 = password == ""? "" : _session->_hostware.hex_md5_from_bytes( password );
        if( stricmp( md5.c_str(), record.as_string( "password" ).c_str() ) != 0 )  throw_xc( "FACTORY-1002" );
    }


    _password_checked = check_password;

    _unit       = record.as_string( "unit"      );
    _parent_id  = record.as_int   ( "parent"    );
    _name       = record.as_string( "name"      );
    _full_name  = record.as_string( "full_name" );
  //_head_id    = record.as_int   ( "head"      , 0 );

    register_object( record.as_int( "id" ) );
}

//---------------------------------------------------------------------------------User::parameters

hostole::Variables User::parameters()
{
    if( !_parameters )
    {
        hostole::Variables parameters;   parameters.create_instance();

        string stmt = "select  `name`, `value`  from USER_VARIABLES "
                      "  where `user_id`=" + as_string( id() );

        hostole::File select ( "-in -seq " + _session->_db_name + stmt );

        while( !select.eof() )
        {
            hostole::Record record = select.get();
            parameters.set( record.as_string(0), record[1] );
        }
        
        select.close();

        _parameters = parameters;
    }

    return _parameters;
}

//--------------------------------------------------------------------------------------User::Close

STDMETHODIMP User::Close()
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );
    
    HRESULT hr = S_OK;

    try
    {
        close();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-----------------------------------------------------------------------------User::get_Parameters

STDMETHODIMP User::get_Parameters( Ivariables2** result )
{
    Z_LOGI2( "factory", __PRETTY_FUNCTION__ << "()\n" );

    HRESULT hr = S_OK;

    try
    {
        *result = parameters().copy();
    }
    catch( const exception&  x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { hr = Set_excepinfo( x, __FUNCTION__ ); }

    return hr;
}

//-------------------------------------------------------------------------------------------------

} //namespace document_factory
} //namespace sos
