// $Id: factory_processor.cxx 3979 2005-10-27 12:41:25Z jz $          Joacim Zschimmer

// 1818

// Der Factory-Prozessor (Version 2)

#include "precomp.h"

#include <wchar.h>

#ifdef SYSTEM_WIN
#   include <io.h>
#   include <crtdbg.h>
#endif

#include <fcntl.h>
#include <strstream>

#include "../kram/scripting_engine.h"

#include "../kram/sosstrng.h"
#include "../kram/sos.h"

#include "../kram/sosappl.h"
#include "../kram/sosctype.h"
#include "../kram/soslimtx.h"
#include "../kram/sosopt.h"
#include "../file/anyfile.h"
#include "../kram/com_simple_standards.h"
#include "../kram/sosscrpt.h"
#include "../kram/sosctype.h"
#include "../kram/tabucase.h"
#include "../kram/log.h"

#include "../hostole/variables.h"

#include "../zschimmer/log.h"
#include "../zschimmer/z_com.h"
#include "../zschimmer/z_com_server.h"
#include "../zschimmer/z_sql.h"

#include "factory_processor.h"
#include "factory_text.h"
#include "factory_rtf.h"

using namespace zschimmer;
using namespace zschimmer::com;

//-------------------------------------------------------------------------------------------

namespace sos {

// sysdep.cxx:
extern string extension_of_path( const string& path );  // sysdep.cxx
extern string module_filename();
extern string directory_of_path( const string& path );
extern string basename_of_path( const string& path );

//-------------------------------------------------------------------------------------------

namespace factory {

//-------------------------------------------------------------------------------Typbibliothek

DESCRIBE_CLASS_CREATABLE( &hostole_typelib, Processor, factory_processor, CLSID_Factory_processor, "hostWare.Factory_processor", "1.0" );

//--------------------------------------------------------------------------------------const

#define     SCRIPT_CONSTRUCTOR_NAME   "factory_constructor"
//#define     SCRIPT_CONSTRUCTOR_NAME   "factory_global"
const char* factory_manager_name    = "factory";        // Alte sosfact.dll (Basic)
const char* factory_document_name   = "factory_document";   // Neue documentfactorydll (C++)
const int   max_rtf_line_length     = 255;      // Max. Lnge einer RTF-Zeile (Ausgabe)
const int   rtf_buffer_size         = 1024;     // Max. Lnge einer eingelesenen RTF-Zeile
const int   script_line_size        = 1024;     // Max. Lnge einer Script-Zeile (#include)
const int   script_size_increment   = 64*1024;
const int   script_size             = 512*1024; // 1MB (unicode)
const int   max_include_nesting     = 50;       // Max. Verschachtelung der #include

//--------------------------------------------------------------------------------hex_to_digit
/*
static int hex_to_digit( char hex )
{
    if( sos_isdigit( hex ) )  return hex - '0';
    return sos_toupper( hex ) - 'A' + 10;
}
*/
//----------------------------------------------------------------------factory_processor_dummy

extern void factory_processor_dummy()
{
    // Damit hostole eine Referenz zum Einbinden hat.
}

//--------------------------------------------------------------Constructor::Constructor

Constructor::Constructor( Ole_class_descr* class_descr, IUnknown* this_ )
:
    Sos_ole_object( class_descr, this_ ),
    _zero_(this+1)
{
}

//----------------------------------------------------------Constructor::factory_put_bstr

STDMETHODIMP Constructor::Put_bstr( BSTR text, VARIANT_BOOL raw )
{
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = NOERROR;

    try {
        put_string( bstr_as_string( text ), raw != 0 );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Constructor::put_bstr" ); }

    return hr;
}

//----------------------------------------------------------------Constructor::factory_put

STDMETHODIMP Constructor::Put( VARIANT* text, VARIANT_BOOL raw )
{
    Z_COM_MY_THREAD_ONLY;

    // Prozedur put fr das Skript. Schreibt etwas (RTF) in die Ausgabedatei

    HRESULT hr = NOERROR;

    if( V_VT( text ) == VT_BSTR ) {
        hr = Put_bstr( V_BSTR( text ), raw );
    } 
    else
    if( V_VT( text ) == VT_NULL ) {
        // nix ausgeben
    }
    else 
    {
        VARIANT vt;
        VariantInit( &vt );
        HRESULT hr = VariantChangeTypeEx( &vt, text, (LCID)0, 0, VT_BSTR );         //MAKELCID(MAKELANGID(0x09, 0x01),SORT_DEFAULT)
        if( FAILED( hr ) ) {
            LOG( "Constructor::put()  VariantChangeTypeEx() failed hr=" << hex << hr << dec << '\n' );
            return hr;
        }

        hr = Put_bstr( V_BSTR( &vt ), raw );

        VariantClear( &vt );
    }

    return hr;
}

//-----------------------------------------------------------Constructor::put_formatted

STDMETHODIMP Constructor::Put_formatted( VARIANT* text, Ausgabeformat format )
{
    HRESULT hr = NOERROR;

    Z_LOG2( "factory.processor.put", "Constructor::put_formatted " << *text << "\n" );
    Z_COM_MY_THREAD_ONLY;

    Variant variant;

    if( V_VT( text ) == VT_DISPATCH )
    {
        DISPPARAMS  dispparams;
        EXCEPINFO   excepinfo;
        uint        error;

        dispparams.rgvarg            = NULL;              // Array of arguments.
        dispparams.rgdispidNamedArgs = NULL;	          // Dispatch IDs of named arguments.
        dispparams.cArgs             = 0;                 // Number of arguments.
        dispparams.cNamedArgs        = 0;                 // Number of named arguments.

        hr = V_DISPATCH( text)->Invoke( DISPID_VALUE, IID_NULL, LOCALE_SYSTEM_DEFAULT, DISPATCH_PROPERTYGET,
                                        &dispparams, &variant, &excepinfo, &error );
        if( FAILED( hr ) ) {
            try {
                throw_ole_excepinfo( hr, &excepinfo, "Constructor::put_formatted", "DISPID_VALUE" );  // Gibt excepinfo frei und protokolliert
            }
            catch( const Xc& )  { goto ENDE; }
        }

        text = &variant;
    }

    if( V_VT( text ) == VT_BSTR ) 
    {
        hr = Put_bstr( V_BSTR( text ), false );
    } 
    else
    if( V_VT( text ) == VT_NULL ) 
    {
        // nix ausgeben
    }
    else 
    {
        switch( format ) 
        {
            case format_std: 
            {
                VARIANT vt;
                VariantInit( &vt );
                // Achtung: VariantChangeTypeEx liefert bei Date-Typen das Jahr nur zweistellig!!!
                //jz 4.10.98 HRESULT hr = VariantChangeTypeEx( &vt, text, MAKELCID(MAKELANGID(LANG_NEUTRAL,SUBLANG_DEFAULT),SORT_DEFAULT), 0, VT_BSTR );
                HRESULT hr = VariantChangeTypeEx( &vt, text, (LCID)0, 0, VT_BSTR );
                if( FAILED( hr ) )  goto ENDE;

                hr = Put_bstr( V_BSTR( &vt ), false );

                VariantClear( &vt );
                break;
            }

            case format_waehrung:
            {
                Sos_limited_text<100> buffer;
                VARIANT vt;

                VariantInit( &vt );

                // Achtung: VariantChangeTypeEx liefert bei Date-Typen das Jahr nur zweistellig!!!
                // Variante sollte sowieso numerisch sein!
                HRESULT hr = VariantChangeTypeEx( &vt, text, MAKELCID(MAKELANGID(LANG_ENGLISH,SUBLANG_DEFAULT),SORT_DEFAULT), 0, VT_BSTR );
                if( FAILED( hr ) )  goto ENDE;

                try {
                    string str = bstr_as_string( V_BSTR( &vt ) );
                    sos::format( &buffer, str, "-9.999.999.999.999.990,00", ',' );
                } 
                catch( const Xc& ) {
                    hr = Put_bstr( V_BSTR( &vt ), false );
                    if( FAILED( hr ) )  { VariantClear( &vt ); goto ENDE; }
                }
/* greift nicht
                rtrim( &buffer );           // ::format lsst Blanks stehen
                buffer.append( '\0' );
                const char* p = buffer.char_ptr();
                while( *p == ' ' )  p++;
                put_area( Const_area( p ) );
*/
                put_string( c_str(buffer), false );

                VariantClear( &vt );
                break;
            }
        }
    }

ENDE: ;

    return hr;
}

//---------------------------------------------------------------------Constructor::embed

STDMETHODIMP Constructor::Embed( BSTR filename, Embed_flags embed_flags )
{
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = NOERROR;

    //Datei ffnen, vom rtf-Rahmen befreien und mit put_line ausgegeben.

    string      saved_source        = _site->_source;
    int         saved_first_lineno  = _site->_first_lineno;

    try {
        string                      fn     = bstr_as_string( filename );
        string                      script = file_as_string( _site->text_filename(fn,embed_flags) );
        Sos_ptr<Embed_synthesizer>  synth  = SOS_NEW( Embed_synthesizer( _site, this ) );
        Sos_ptr<Parser>             parser = Parser::create( _site, synth, c_str(script), fn );

        _site->_source = fn;
        _site->_first_lineno = 0;

        parser->set_constructor( this );
        parser->_include = true;
        parser->_allowed_properties = embed_flags? (rtf::Flags)embed_flags : rtf::flag_prop_para | rtf::flag_prop_char;
        parser->_restore_properties = true;
        parser->init();
        parser->parse( c_str(script) );

        _site->_source = saved_source;
        _site->_first_lineno = saved_first_lineno;
    }
    catch( Xc& x )
    {
        _site->_source = saved_source;
        _site->_first_lineno = saved_first_lineno;

        x.insert( "embed" );
        hr = _set_excepinfo( x, "hostWare.Constructor::embed" );
    }
    catch( const exception& x )
    {
        _site->_source = saved_source;
        _site->_first_lineno = saved_first_lineno;
        hr = _set_excepinfo( x, "hostWare.Constructor::embed" );
    }

    return hr;
}

//-------------------------------------------------------Script_synthesizer::Script_synthesizer

Script_synthesizer::Script_synthesizer( Site* site, Area* buffer )
: 
    _zero_(this+1),
    _buffer( buffer )
{ 
    _site = site;
}

//-----------------------------------------------------------------Script_synthesizer::put_text

void Script_synthesizer::put_text( const string& txt, bool raw )
{
    const Byte* p     = (Byte*)c_str( txt );
    const Byte* p_end = p + txt.length();

    while( p < p_end )
    {
        const Byte* p0 = p;

        while( ( p[0] & 0x7F ) >= ' '  &&  p[0] != '"'  &&  p[0] != '\\' )  p++;
        if( p > p0 ) {
            _buffer->append( _indent );
            _buffer->append( '\t' );
            _buffer->append( _site->_script_constructor_prefix );
            _buffer->append( "put " );
            if( _site->_use_call_parentheses )  _buffer->append( '(' );
            _buffer->append( _site->_quote_char );
            _buffer->append( p0, p - p0 );
            _buffer->append( _site->_quote_char );
            _buffer->append( ',' );
            _buffer->append( '0' + raw );
            if( _site->_use_call_parentheses )  _buffer->append( ')' );
            if( _site->_use_semicolon )  _buffer->append( ';' );
            _buffer->append( "\r\n" );
            _next_lineno++;
        }

        if( p == p_end )  break;

        _buffer->append( _indent );
        _buffer->append( '\t' );
        _buffer->append( _site->_script_constructor_prefix );
        _buffer->append( "put " );
        if( _site->_use_call_parentheses )  _buffer->append( '(' );

        if( _site->_use_backslash )
        {
            _buffer->append( '\"' );    // Bei Perl: "\x9", nicht '\x9'!  _site->_quote_char );
            _buffer->append( "\\x" );
            _buffer->append( printf_string( "%02X", (int)(uint)p[0] ) );
            _buffer->append( '\"' );    // Bei Perl: "\x9", nicht '\x9'!  _site->_quote_char );
        }
        else
        {
            _buffer->append( "chr(" );                      // Nur VBScript
            _buffer->append( as_string( (uint)p[0] ) );
            _buffer->append( ')' );
        }

        _buffer->append( ',' );
        _buffer->append( '0' + raw );

        if( _site->_use_call_parentheses )  _buffer->append( ')' );
        if( _site->_use_semicolon )  _buffer->append( ';' );
        _buffer->append( "\r\n" );
        _next_lineno++;

        p++;
    }
}

//------------------------------------------------------------------Script_synthesizer::put_expr

void Script_synthesizer::put_expr( const string& expr_, Ausgabeformat format )
{
    if( _site->_suppress_script )  
    {
        string expr = expr_;
        for( int i = 0; i < expr.length(); i++ )  if( (unsigned char)expr[i] < ' ' )  expr[i] = '_';
        expr = z::quoted_string( "$(" + expr + ")", _site->_quote_char, _site->_use_backslash? '\\' : _site->_quote_char );
        put_expr2( expr, format_std );
    }
    else
        put_expr2( expr_, format );
}

//-----------------------------------------------------------------Script_synthesizer::put_expr2

void Script_synthesizer::put_expr2( const string& expr, Ausgabeformat format )
{
    char    f [ 40 ];

    _buffer->append( _indent );
    _buffer->append( '\t' );
    _buffer->append( _site->_script_constructor_prefix );
    _buffer->append( "put_formatted " );

    if( _site->_use_call_parentheses )  _buffer->append( '(' );

    _buffer->append( expr );
    
    _buffer->append( ',' );
    sprintf( f, "%d", (int)format );
    _buffer->append( f );

    if( _site->_use_call_parentheses )  _buffer->append( ')' );
    if( _site->_use_semicolon        )  _buffer->append( ';' );

    _buffer->append( "\r\n" );

    for( int i = 0; i < expr.length(); i++ )  if( expr[i] == '\n' )  _next_lineno++;

    _next_lineno++;
}

//-----------------------------------------------------------Script_synthesizer::map_next_line
/*
void Script_synthesizer::map_next_line( int doc_no, zschimmer::rtf::Entity* e )
{
    _rtf_line_map[ _next_lineno ]._doc_no = doc_no;
    _rtf_line_map[ _next_lineno ]._entity = e;
}
*/
//---------------------------------------------------------Script_synthesizer::put_script_line

void Script_synthesizer::put_script_line( const string& text )
{
    if( _site->_suppress_script )
    {
        string text_ = text;
        for( int i = 0; i < text_.length(); i++ )  if( (unsigned char)text_[i] < ' ' )  text_[i] = '_';
        text_ = z::quoted_string( "<%" + text_ + "%>", _site->_quote_char, _site->_use_backslash? '\\' : _site->_quote_char );
        put_expr2( text_, format_std );
    }
    else
    {
        _buffer->append( text );
        _buffer->append( "\r\n" );

        for( int i = 0; i < text.length(); i++ )  if( text[i] == '\n' )  _next_lineno++;
        _next_lineno++;
    }
}

//---------------------------------------------------------Script_synthesizer::put_preprocessor

void Script_synthesizer::put_preprocessor( const string& stmt, bool text_script )
{
    bool        is_include        = false;
    bool        is_include_script = false;
    const char* p = c_str( stmt );
    while( isspace( *p ) )  p++;

    if( is_include = strncmpi( p, "include ",  8 ) == 0, is_include )  p += 8;
    else
    if( is_include_script = strncmpi( p, "include_script ", 15 ) == 0, is_include_script )  p += 15;

    if( is_include || is_include_script )
    {
        if( _include_nesting >= max_include_nesting )  throw_xc( "FACTORY-205" );  // Welche Position haben wir hier? throw_syntax_error( ..., POS );
        Increment_semaphore<int> sema ( &_include_nesting );    // _include_nesting++;

        include_( p, is_include && text_script );    // ruft rekursiv put_script_line fr verschachtelte #include
                                                     // #include in #include_script wird zu #include_script!
    }
    else 
        throw_xc( "FACTORY-203", c_str( stmt ) );
}

//----------------------------------------------------------------Script_synthesizer::rtf_prop

void Script_synthesizer::rtf_prop( int prop, int param )
{
    if( _buffer->size() - _buffer->length() < 100 )   _buffer->resize_min( _buffer->size() + 100000 );

    _buffer->append( _indent );
    _buffer->append( '\t' );
    _buffer->append( _site->_script_constructor_prefix );
    _buffer->append( "rtf_prop " );
    if( _site->_use_call_parentheses )  _buffer->append( '(' );

    char* p = _buffer->char_ptr() + _buffer->length();
    sprintf( p, "%d,%d", prop, param );
    _buffer->length( _buffer->length() + strlen( p ) );

    if( _site->_use_call_parentheses )  _buffer->append( ')' );
    if( _site->_use_semicolon )  _buffer->append( ';' );
    _buffer->append( "\r\n" );
    _next_lineno++;
}

//-----------------------------------------------------------------Script_synthesizer::rtf_put

void Script_synthesizer::rtf_put( int begin, int end )
{
    if( begin != end )
    {
        if( _buffer->size() - _buffer->length() < 100 )   _buffer->resize_min( _buffer->size() + 100000 );

        _buffer->append( _indent );
        _buffer->append( '\t' );
        _buffer->append( _site->_script_constructor_prefix );
        _buffer->append( "rtf_put " );
        if( _site->_use_call_parentheses )  _buffer->append( '(' );

        char* p = _buffer->char_ptr() + _buffer->length();
        sprintf( p, "%d,%d", begin, end );
        _buffer->length( _buffer->length() + strlen( p ) );

        if( _site->_use_call_parentheses )  _buffer->append( ')' );
        if( _site->_use_semicolon        )  _buffer->append( ';' );
        _buffer->append( "\r\n" );
        _next_lineno++;
    }
}

//----------------------------------------------------------------Script_synthesizer::rtf_open

void Script_synthesizer::rtf_open( int entity_ptr )
{
    if( _buffer->size() - _buffer->length() < 100 )   _buffer->resize_min( _buffer->size() + 100000 );

    _buffer->append( _indent );
    _buffer->append( '\t' );
    _buffer->append( _site->_script_constructor_prefix );
    _buffer->append( "rtf_open " );
    if( _site->_use_call_parentheses )  _buffer->append( '(' );

    char* p = _buffer->char_ptr() + _buffer->length();
    sprintf( p, "%d", entity_ptr );
    _buffer->length( _buffer->length() + strlen( p ) );

    if( _site->_use_call_parentheses )  _buffer->append( ')' );
    if( _site->_use_semicolon        )  _buffer->append( ';' );
    _buffer->append( "\r\n" );
    _next_lineno++;

    _indent += "\t";
}

//---------------------------------------------------------------Script_synthesizer::rtf_close

void Script_synthesizer::rtf_close()
{
    if( _buffer->size() - _buffer->length() < 100 )   _buffer->resize_min( _buffer->size() + 100000 );

    _indent = _indent.substr( 0, _indent.length() - 1 );
    _buffer->append( _indent );
    _buffer->append( '\t' );
    _buffer->append( _site->_script_constructor_prefix );
    _buffer->append( "rtf_close " );
    if( _site->_use_call_parentheses )  _buffer->append( "()" );
    if( _site->_use_semicolon        )  _buffer->append( ';' );
    _buffer->append( "\r\n" );
    _next_lineno++;
}

//------------------------------------------------------------Script_synthesizer::rtf_include_begin

void Script_synthesizer::rtf_include_begin()
{
    if( _buffer->size() - _buffer->length() < 100 )   _buffer->resize_min( _buffer->size() + 100000 );

    _buffer->append( _indent );
    _buffer->append( '\t' );
    _buffer->append( _site->_script_constructor_prefix );
    _buffer->append( "rtf_include_begin " );
    if( _site->_use_call_parentheses )  _buffer->append( "()" );
    if( _site->_use_semicolon        )  _buffer->append( ';' );
    _buffer->append( "\r\n" );
    _next_lineno++;

    _indent += "\t";
}

//--------------------------------------------------------------Script_synthesizer::rtf_include_end

void Script_synthesizer::rtf_include_end()
{
    if( _buffer->size() - _buffer->length() < 100 )   _buffer->resize_min( _buffer->size() + 100000 );

    _indent = _indent.erase( _indent.length() - 1 );
    _buffer->append( _indent );
    _buffer->append( '\t' );
    _buffer->append( _site->_script_constructor_prefix );
    _buffer->append( "rtf_include_end " );
    if( _site->_use_call_parentheses )  _buffer->append( "()" );
    if( _site->_use_semicolon        )  _buffer->append( ';' );
    _buffer->append( "\r\n" );
    _next_lineno++;
}

//-------------------------------------------------------------------Script_synthesizer::rtf_start

void Script_synthesizer::rtf_start()
{
    _buffer->append( '\t' );
    _buffer->append( _site->_script_constructor_prefix );
    _buffer->append( "rtf_start" );
    if( _site->_use_call_parentheses )  _buffer->append( "()" );
    if( _site->_use_semicolon )  _buffer->append( ';' );
    _buffer->append( "\r\n" );
    _next_lineno++;
}

//-------------------------------------------------------------------Script_synthesizer::rtf_finish

void Script_synthesizer::rtf_finish()
{
    _buffer->append( '\t' );
    _buffer->append( _site->_script_constructor_prefix );
    _buffer->append( "rtf_finish" );
    if( _site->_use_call_parentheses )  _buffer->append( "()" );
    if( _site->_use_semicolon )  _buffer->append( ';' );
    _buffer->append( "\r\n" );
    _next_lineno++;
}

//---------------------------------------------------------------Script_synthesizer::include_script

void Script_synthesizer::include_script( const string& script ) 
{
    const char* p = c_str(script);
    while( *p )
    {
        const char* nl = strchr( p, '\n' );
        if( nl )  nl++;
            else  nl = p + strlen( p );

        if( *p == '#' )  put_preprocessor( as_string( p+1, nl-p ), false );  // false: #include --> #include_script
                   else  _buffer->append( as_string( p, nl-p ) );

        p = nl;
    }
}

//------------------------------------------------------Script_synthesizer::include_file

void Script_synthesizer::include_file( const string& filename, bool text_include, Embed_flags embed_flags ) 
{
    const char sep = '\f'; // Formfeed als Trenner
    
    if( strchr( c_str( filename ), sep ) )     // Liste von durch Semikolon getrennten Dateinamen?
    {
        const char* p = c_str( filename );
        while( *p ) {
            const char* s;
            // Nchstes nicht gequotetes Semikolon suchen
            s = strchr( p, sep );
            // TODO: falls Escape-Zeichen gewnscht => Esc-Zeichen wieder entfernen
            //while( s && s > p && *(s-1) == esc ) s = strchr( s+1, sep );

            if( !s )  s = p + strlen( p );
            include_file( as_string( p, s - p ), text_include, embed_flags );
            p = s;
            if( *p == sep )  p++;
        }
        return;
    }

    Z_LOG2( "factory.processor", "#include " << filename << '\n' );

    if( empty( filename ) )  return;

    string  saved_source        = _site->_source;
    int     saved_first_lineno  = _site->_first_lineno;

    try
    {
        _site->_source = filename;
        _site->_first_lineno = 0;

        string script = file_as_string( _site->text_filename( filename, embed_flags ), SYSTEM_NL );
        
        if( text_include )  // Text mit Skript
        {
            rtf_include_begin();

            Sos_ptr<Parser> parser = Parser::create( _site, this, c_str(script), filename );

            parser->_main_parser = _site->_main_parser;      // Parser, der oberstes #include ausfhrt
            parser->set_constructor( _site->_constructor );
            parser->_include = true;
            parser->_allowed_properties = embed_flags? (rtf::Flags)embed_flags : rtf::flag_prop_para | rtf::flag_prop_char;
          //parser->_restore_properties = true;  // Das wirkt nicht, denn _main_doc wird noch nicht aufgebaut. Stattdessen put_properties=true bei sym_preprocessor
            parser->init();
            parser->parse( c_str(script) );

            rtf_include_end();
        }
        else                // Reines Skript, ohne Text
        {       
            include_script( script );
        }
    }
    catch( Xc& x )
    {
        _site->_source = saved_source;
        _site->_first_lineno = saved_first_lineno;

        x.insert( "#include" );
        throw;
    }
    catch( exception& )
    {
        _site->_source = saved_source;
        _site->_first_lineno = saved_first_lineno;
        throw;
    }

    _site->_source = saved_source;
    _site->_first_lineno = saved_first_lineno;
}

//-----------------------------------------------------------Script_synthesizer::include_

void Script_synthesizer::include_( const char* p, bool text_include )
{
    string      filename;
    string      expr;
 //?bool        factory_texts = false;
    Embed_flags embed_flags = (Embed_flags)0;

    while( *p == ' ' )  p++;

    if( *p == '"' )     // Stringkonstante
    {
        p++;
        while( p[0] ) {
            if( p[0] == '"' ) { p++; if( p[0] != '"' )  break; }
            filename += p[0];
            p++;
        }
    }
    else
    {
        while( *p  &&  *p != ',' )   expr += *p++;      // Sehr simpel: Ausdruck darf kein Komma enthalten
        expr.erase( length_without_trailing_spaces( c_str(expr) ) );
    }

    while( *p == ' ' )  p++;
    
    if( text_include  &&  p[0] == ',' )
    {
        p++;
        while( *p == ' ' )  p++;
        while( isdigit( (Byte)*p ) )  embed_flags = (Embed_flags)( 10*embed_flags + *p++ - '0' );
    }

    while( isspace(*p) )  p++;
    if( *p )  throw_xc( "FACTORY-204", p );


    if( !filename.empty() ) 
    {
        include_file( filename, text_include, embed_flags );
    } 
    else 
    {
        if( expr == "factory_texts"  &&  _site->_factory_texts_valid )    // VBScript 2.0 kann offenbar keine Expressions!
        {
            include_file( _site->_factory_texts, text_include, embed_flags );
        }
        else
        {
            if( !_site->_script_parse )  throw_xc( "FACTORY-204", p );

            HRESULT     hr;
            VARIANT     result;
            EXCEPINFO   excep;

            try 
            {
                SOS_DELETE( _site->_script_exception );
                VariantInit( &result );
                BSTR script_text = SysAllocString_string( expr );

                Z_LOG2( "factory.processor", "ParseScriptText( \"" << script_text << "\", SCRIPTTEXT_ISEXPRESSION )\n" );

                hr = _site->_script_parse->ParseScriptText( script_text,
                                                            NULL, 
                                                            NULL, 
                                                            NULL, 
                                                            0, 
                                                            0, 
                                                            SCRIPTTEXT_ISEXPRESSION,
                                                            &result,
                                                            &excep
                                                          );
                SysFreeString( script_text );  
                script_text = NULL;

                if( _site->_script_exception ) {
                    SysFreeString( excep.bstrSource );
                    SysFreeString( excep.bstrDescription );
                    SysFreeString( excep.bstrHelpFile );
                    throw *_site->_script_exception;
                }
                if( FAILED( hr ) )  throw_ole_excepinfo( hr, &excep, "IActiveScriptParse::ParseScriptText" ); 
            
                //Z_LOG2( "factory.processor", "ParseScriptText ok\n" );

                if( V_VT( &result ) & VT_ARRAY ) 
                {
                    long32              lower_bound;
                    long32              upper_bound;
                    int                 n;
                    VARIANT*            params = NULL;

                    hr = SafeArrayGetLBound( V_ARRAY( &result ), 1, &lower_bound );
                    if( FAILED( hr ) )  throw_ole( hr, "SafeArrayGetLBound" );

                    hr = SafeArrayGetUBound( V_ARRAY( &result ), 1, &upper_bound );
                    if( FAILED( hr ) )  throw_ole( hr, "SafeArrayGetUBound" );

                    n = upper_bound - lower_bound + 1;

                    hr = SafeArrayAccessData( V_ARRAY( &result ), (void __huge**)&params );
                    if( FAILED( hr ) )  throw_ole( hr, "SafeArrayAccessData" );

                    for( int i = 0; i < n; i++ ) 
                    {
                        include_file( variant_as_string( params[ i ] ), text_include, embed_flags );
                    }
                }
                else 
                {
                    include_file( variant_as_string( result ), text_include, embed_flags );
                }

                VariantClear( &result );
            }
            catch( const Xc& ) 
            {
                VariantClear( &result );
                throw;
            }
        }
    }
}

//------------------------------------------------------------Script_synthesizer::script_error
/*
void Script_synthesizer::script_error( const Xc& x )
{
    if( _site && _site->_constructor )  _site->_constructor->script_error( x );

    throw x;
}
*/
//--------------------------------------------------------Embed_synthesizer::Embed_synthesizer

Embed_synthesizer::Embed_synthesizer( Site* site, Constructor* constructor )
: 
    _zero_(this+1),
    _constructor(constructor)
{ 
    _site = site;
}

//-----------------------------------------------------------------Embed_synthesizer::put_text

void Embed_synthesizer::put_text( const string& txt, bool raw )
{
    _constructor->put_string( txt, raw );
}

//------------------------------------------------------------------Embed_synthesizer::rtf_put

void Embed_synthesizer::rtf_prop( int prop, int param )
{
    //Z_LOG2( "joacim", __PRETTY_FUNCTION__ << " \\" << ((Rtf_constructor*)+_constructor)->descr((rtf::Code)prop)->_name << param << "\n" );
    _constructor->rtf_prop( prop, param );
}

//------------------------------------------------------------------Embed_synthesizer::rtf_put

void Embed_synthesizer::rtf_put( int begin, int end )
{
    //for( rtf::Simple_entity* e = (rtf::Simple_entity*)begin; e != (rtf::Simple_entity*)end; e = e->_next )
    //    if( z::Log_ptr log = "joacim" )  log << __PRETTY_FUNCTION__ << " ", ((Rtf_constructor*)+_constructor)->print_entity( log, *e ), log << "\n";
    //if( end )  if( z::Log_ptr log = "joacim" )  log << __PRETTY_FUNCTION__ << " end=", ((Rtf_constructor*)+_constructor)->print_entity( log, *(rtf::Simple_entity*)end ), log << "\n";

    _constructor->rtf_put( begin, end );
}

//-----------------------------------------------------------------Embed_synthesizer::rtf_open

void Embed_synthesizer::rtf_open( int entity_ptr )
{
    _constructor->rtf_open( entity_ptr );
}

//-----------------------------------------------------------------Embed_synthesizer::rtf_close

void Embed_synthesizer::rtf_close()
{
    _constructor->rtf_close();
}

//------------------------------------------------------------------Embed_synthesizer::put_expr

void Embed_synthesizer::put_expr( const string& expr, Ausgabeformat format )
{
    HRESULT     hr;
    VARIANT     result;
    Excepinfo    excep;
    BSTR        script_text = SysAllocStringLen_char( expr.c_str(), expr.length() );

    SOS_DELETE( _site->_script_exception );
    VariantInit( &result );

    hr = _site->_script_parse->ParseScriptText( script_text,
                                                NULL, 
                                                NULL, 
                                                NULL, 
                                                0, 
                                                0, 
                                                SCRIPTTEXT_ISEXPRESSION,
                                                &result,
                                                &excep
                                              );

    SysFreeString( script_text ),  script_text = NULL;

    if( _site->_script_exception ) {
        SysFreeString( excep.bstrSource );
        SysFreeString( excep.bstrDescription );
        SysFreeString( excep.bstrHelpFile );
        throw *_site->_script_exception;
    }
    if( FAILED( hr ) )  throw_ole_excepinfo( hr, &excep, "IActiveScriptParse::ParseScriptText" ); 

    hr = _constructor->Put_formatted( &result, format );
    if( FAILED( hr ) )  throw_ole_excepinfo( hr, &excep, "Constructor::put" ); 
}

//-----------------------------------------------------------Embed_synthesizer::map_next_line
/*
void Embed_synthesizer::map_next_line( int, zschimmer::rtf::Entity* )
{
    throw_xc( "FACTORY-206" );
}
*/
//----------------------------------------------------------Embed_synthesizer::put_script_line

void Embed_synthesizer::put_script_line( const string& )
{
    throw_xc( "FACTORY-206" );
}

//-------------------------------------------------------------Embed_synthesizer::script_error
/*
void Embed_synthesizer::script_error( const Xc& x )
{
    throw x;
}
*/
//--------------------------------------------------------------------Embed_synthesizer::rtf_start

void Embed_synthesizer::rtf_start()
{
    throw_xc( "Embed_synthesizer::rtf_start" );
}

//-------------------------------------------------------------------Embed_synthesizer::rtf_finish

void Embed_synthesizer::rtf_finish()
{
    throw_xc( "Embed_synthesizer::rtf_finish" );
  //_constructor->finish_();
}

//----------------------------------------------------------Embed_synthesizer::put_preprocessor

void Embed_synthesizer::put_preprocessor( const string&, bool )
{
    throw_xc( "FACTORY-207" );
}

//-------------------------------------------------------------------------------Parser::create

Sos_ptr<Parser> Parser::create( Site* site, Synthesizer* synth, const char* script, const string& doc_name )
{
    Sos_ptr<Parser> parser;
    if( strncmp( c_str( script ), "{\\rtf1", 6 ) == 0 )  parser = SOS_NEW( Rtf_parser( &site->_processor->_rtf_context ) );
                                                   else  parser = SOS_NEW( Text_parser );

    parser->_site = site;
    parser->_synth = synth;
    parser->_document_name = doc_name;

    return parser;
}

//----------------------------------------------------------------------Parser::set_constructor

void Parser::set_constructor( Constructor* constructor )
{
    _constructor = constructor;

    if( _constructor )  _constructor->_parser = this;
}

//------------------------------------------------------------------------------Parser::~Parser

Parser::~Parser()
{
    /*
    if( _constructor ) {
        _constructor->Release();
        _constructor = NULL;
    }
    */
}

//-----------------------------------------------------------------------------------Site::Site

Site::Site( Processor* processor )
: 
    _zero_(this+1),
    _processor(processor),
    _script_size ( script_size )
{
#   ifdef SYSTEM_WIN
        set_engine_name( "VBScript" );
#    else
        set_engine_name( "PerlScript" );
#   endif

    _variables = new Hostware_variables;
    _template_filename_changed = true;
}
                             
//----------------------------------------------------------------------------------Site::~Site

Site::~Site()
{
    Z_LOG2( "factory.processor", "~Site\n" );
/*
    if( _constructor ) { 
        _constructor->Release(); 
        _constructor = NULL; 
    }
*/
}
                                     
//----------------------------------------------------------------------------------Site::init

void Site::init()
{
}
                             
//-----------------------------------------------------------------------Site::set_engine_name

void Site::set_engine_name( const string& name )
{
    _engine_name = name;    // z.B. "VBScript"

    _is_unix_perlscript = false;


    if( stricmp(c_str(_engine_name),"VBScript") == 0 ) 
    {
        _script_constructor_prefix = SCRIPT_CONSTRUCTOR_NAME ".";
        _quote_char                = '"';
        _quotes                    = "\"";
        _use_backslash             = false;
        _use_call_parentheses      = false;
        _use_semicolon             = false;
        _var_char                  = '&';
        _var_prefix                = 0;
        _comment                   = '\'';
    }
    else
    if( stricmp(c_str(_engine_name),"JavaScript") == 0 
     || stricmp(c_str(_engine_name),"JScript"   ) == 0  
     || stricmp(c_str(_engine_name),"ECMAScript") == 0 
     || stricmp(c_str(_engine_name),"Spidermonkey") == 0 ) 
    {
        _script_constructor_prefix = SCRIPT_CONSTRUCTOR_NAME ".";
        _quote_char                = '\'';
        _quotes                    = "\"'";
        _use_backslash             = true;
        _use_call_parentheses      = true;
        _use_semicolon             = true;
        _var_char                  = '$';
        _var_prefix                = 0;
        _comment                   = "//";
    }
    else
    if( stricmp(c_str(_engine_name),"PerlScript") == 0
     || stricmp(c_str(_engine_name),"Perl") == 0 ) 
    {
        _script_constructor_prefix = "$" SCRIPT_CONSTRUCTOR_NAME "->";
        _quote_char                = '\'';
        _quotes                    = "\"'`";
        _use_backslash             = true;
        _use_call_parentheses      = true;
        _use_semicolon             = true;
        _var_char                  = '$';
        _var_prefix                = '$';
        _comment                   = "//";

#       ifdef Z_UNIX
            _is_unix_perlscript = true;
#       endif
    }
}

//-----------------------------------------------------------------------------------Site::run

void Site::put_param( Sos_option_iterator* opt )
{
    opt->log();

    for(; !opt->end(); opt->next() ) 
    {
        if( opt->with_value( "log" )   )               log_start( opt->value() );
        else
        if( opt->with_value( "output" ) )              _output_filename = opt->value();
        else
        if( opt->with_value( "engine" ) )              set_engine_name( opt->value() );
        else
        if( opt->with_value( "constructor-prefix" ) )  _script_constructor_prefix = opt->value();
        else
        if( opt->with_value( "quote" ) )               _quote_char = opt->as_char();
        else
        if( opt->with_value( "quotes" ) )              _quotes = opt->value();
        else
        if( opt->with_value( "use-backslash" ) )       _use_backslash = opt->set();
        else
        if( opt->with_value( "use-call-parentheses" ) ) _use_call_parentheses = opt->set();
        else
        if( opt->with_value( "use-semicolon" ) )       _use_semicolon = opt->set();
        else
        if( opt->with_value( "var-char" ) )            _var_char = opt->as_char();
        else
        if( opt->with_value( "var-prefix" ) )          _var_prefix = opt->as_char();
        else
        if( opt->with_value( "comment" ) )             _comment = opt->value();
        else
        if( opt->with_value( "script-size" ) )         _script_size = opt->as_uintK();       // Mehr Zeichen darf das entstehende Skript nicht haben
        else
        if( opt->with_value( "factory-texts" ) )       { _factory_texts_valid = true;  _factory_texts = opt->value(); }        // Wert fr #include factory_texts
        else
        if( opt->with_value( "source-name" ) )         _source = opt->value();                // Fr Fehlermeldungen
        else
        if( opt->with_value( "first-lineno" ) )        _first_lineno = opt->as_int();         // Fr Fehlermeldungen
        else
        if( opt->flag      ( "rtf-single-lines" ) )    _rtf_single_lines = opt->set();        // Jede RTF-Anweisung auf eine eigene Zeile (fr diff)
        else
        if( opt->flag      ( "rtf-indent" ) )          _rtf_indent = opt->set()? 2 : 0 ;      // RTF-Dokument hbsch einrcken
        else
        if( opt->with_value( "rtf-indent" ) )          _rtf_indent = opt->as_int();           // RTF-Dokument hbsch einrcken
        else
        if( opt->flag      ( "suppress-script" ) )     _suppress_script = opt->set();         // Skript-Anweisungen ignorieren
        else
        if( opt->with_value( "list-rtf-codes" ) )      { rtf::print_rtf_table( &ofstream( opt->value().c_str() ) ); }
        else 
            throw_sos_option_error( *opt );
    }
}

//---------------------------------------------------------------------------Site::make_script

void Site::make_script( const string& source_text, const string& template_name )
{
    Z_LOGI2( "factory.processor", "Site::make_script\n" );

    Sos_ptr<Script_synthesizer> synth   = SOS_NEW( Script_synthesizer( this, &_script_buffer ) );

    _main_parser = Parser::create( this, synth, c_str(source_text), template_name );

    //Z_MUTEX( hostware_mutex )
    try
    {
        _script_buffer.allocate_min( _script_size ); // Puffer fr das grte Skript anlegen

        _main_parser->_allowed_properties = rtf::flag_prop;  // Alles durchlassen

        if( _different_templates_allowed )
        {
            if( _constructor )  _main_parser->_allowed_properties = ~rtf::flag_prop_doc;  // Beim zweiten Dokument Dokumenteigenschaften nicht durchlassen
            _main_parser->set_constructor( _constructor );
        }

        _main_parser->init();

        _constructor = _main_parser->_constructor;

        Z_LOGI2( "factory.processor", "_main_parser->parse( c_str(source_text) ) ...\n" );
        _main_parser->parse( c_str(source_text) );
        Z_LOG2( "factory.processor", "_main_parser->parse( c_str(source_text) ) OK\n" );

#       ifdef _DEBUG
#           ifdef SYSTEM_WIN
                string debug_filename = "c:/tmp/factory-script.txt";
                ofstream(debug_filename.c_str()) << _script_buffer;
#            else
                string debug_filename = "/tmp/factory-script.txt";
                //#{ ofstream f ("/tmp/factory-script.txt"); f << _script_buffer; }
                { FILE* f=fopen(debug_filename.c_str(),"w"); fwrite(_script_buffer.ptr(),_script_buffer.length(),1,f); fclose(f); }
#           endif
            LOG( debug_filename << " enthlt Skript\n" );
#       endif

        _script_bstr.assign( _script_buffer.char_ptr(), _script_buffer.length() );
        _script_buffer.free();

        _main_parser = NULL;
    }
    catch( const exception& )
    {
        _main_parser = NULL;
        _script_buffer.free();
        throw;
    }
}

//-----------------------------------------------------------------------------Site::preprocess

BSTR Site::preprocess( BSTR script_bstr )
{
    Z_LOGI2( "factory.processor", "Site::preprocess\n" );

    BSTR                        result_script = NULL;
    string                      source_script = bstr_as_string(script_bstr);
    Sos_ptr<Script_synthesizer> synth         = SOS_NEW( Script_synthesizer( this, &_script_buffer ) );

    try
    {
        _script_buffer.allocate_min( _script_size );  // Puffer fr das grte Skript anlegen
        _script_buffer.length( 0 );

        synth->include_script( source_script );

        result_script = SysAllocStringLen_char( _script_buffer.char_ptr(), _script_buffer.length() );
        _script_buffer.free();
    }
    catch( const exception& )
    {
        _script_buffer.free();
        throw;
    }

    return result_script;
}

//-----------------------------------------------------------------------------------Site::parse

void Site::parse( BSTR script_text, ::Scripttext_flags flags, VARIANT* result )
{
    HRESULT   hr;
    EXCEPINFO excep;        

    memset( &excep, 0, sizeof excep );

    Z_LOGI2( "factory.processor", "Site::parse  ParseScriptText\n" );

    if( !init_engine_called() )  init_engine();

    SOS_DELETE( _script_exception );

    hr = _script_parse->ParseScriptText( script_text,
                                         NULL, 
                                         NULL, 
                                         NULL, 
                                         0, 
                                         0, 
                                         flags,   // SCRIPTTEXT_ISVISIBLE, 
                                         result,  // Nur fr SCRIPTTEXT_ISEXPRESSION
                                         &excep
                                       );


    if( _script_exception ) {
        SysFreeString( excep.bstrSource );
        SysFreeString( excep.bstrDescription );
        SysFreeString( excep.bstrHelpFile );
        //if( _constructor )  _constructor->script_error( *_script_exception );
        throw *_script_exception;
    }

    if( FAILED( hr ) ) {
        if( !excep.bstrSource  ||  !excep.bstrSource[0] )  excep.bstrSource = SysAllocString_string( _source );
        throw_ole_excepinfo( hr, &excep, "IActiveScriptParse::ParseScriptText" ); 
    }
}

//----------------------------------------------------------------------------Site::start_engine

void Site::start_engine()
{
    HRESULT   hr;
    
    SOS_DELETE( _script_exception );

    // Skript starten:
    Z_LOGI2( "factory.processor", "Site::start_engine  SetScriptState( SCRIPTSTATE_STARTED )\n" );
    hr = _script->SetScriptState( SCRIPTSTATE_STARTED );

    if( _script_exception ) {
        //if( _constructor )  _constructor->script_error( *_script_exception );
        throw *_script_exception;
    }

    if( FAILED( hr ) )  throw_ole( hr, "" );  //, "IActiveScript::SetScriptState", "SCRIPTSTATE_STARTED" );
}

//----------------------------------------------------------------------------Site::close_engine

void Site::close_engine()
{
    _parameters_added = false;
    _variables = new Hostware_variables();   // 1818

    Script_site::close_engine();
}

//-------------------------------------------------------------------------Site::add_parameters

void Site::add_parameters()
{
    //if( _parameters_added )  throw_xc( "SOS-1429" );
    if( _parameters_added )  return;

    if( stricmp( _engine_name.c_str(), "vbscript" ) == 0 )
    {
        add_obj( z::qi_ptr<Ivariables2>( _variables ), Bstr( "factory_parameters" ), sif_isvisible | sif_globalmembers );        // Parameter als OBJEKTE bergeben (call-by-ref).
    }
    else
    {
        add_obj( z::qi_ptr<Ivariables2>( _variables ), Bstr( "factory_parameters" ), sif_isvisible );

        // Dann einzeln bergeben

        for( Hostware_variables::Map::iterator it = _variables->_map.begin(); it != _variables->_map.end(); it++ )
        {
            if( it->second )
            {
                if( !_is_unix_perlscript )
                {
                    add_obj( static_cast<Ivariable*>( it->second ), it->first, sif_isvisible );        // Parameter als OBJEKTE bergeben (call-by-ref).
                }
#ifdef Z_UNIX
                else
                {
                    HRESULT     hr;
                    string      stmt;
                    Bstr        name_bstr;
                    Variant     index_vt;
                    Variant     value_vt;

                    hr = it->second->get_Name( &name_bstr );                if(FAILED(hr)) throw_ole( hr, "Ihostware_variable::name" );
                    hr = it->second->get_Value( &index_vt, &value_vt );     if(FAILED(hr)) throw_ole( hr, "Ihostware_variable::value" );

                    if( value_vt.vt != VT_EMPTY )
                    {
                        if( _var_prefix )  stmt = _var_prefix;
                        stmt += bstr_as_string(name_bstr);
                        stmt += '=';
                    
    //fprintf(stderr,"Site::add_parameters value_vt.vt=%d\n", value_vt.vt );
                        if( zschimmer::com::variant_is_numeric(value_vt) || value_vt.vt == VT_BOOL )  
                            stmt += variant_as_string(value_vt);
                        else 
                            stmt += z::quoted_string( variant_as_string(value_vt), _quote_char, _use_backslash? '\\' : _quote_char );

                        if( _use_semicolon )  stmt += ';';
                    
                        Z_LOGI2( "factory.processor", "Site::add_parameters: " << stmt << "\n" );
                        parse( Bstr(stmt), (::Scripttext_flags)0, NULL );
                    }
                }
#endif
            }
        }
    }

    _parameters_added = true;
}

//-------------------------------------------------------------------------Site::connect_script
/*
void Site::connect_script()
{
    HRESULT   hr;

    LOGI( "Site::connect_script  SetScriptState( SCRIPTSTATE_CONNECTED )\n" );
    hr = _script->SetScriptState( SCRIPTSTATE_CONNECTED );
    if( FAILED( hr ) )  throw_ole( hr, "IActiveScript::SetScriptState", "SCRIPTSTATE_CONNECTED" );
}
*/

//--------------------------------------------------------------------------factory_normalized_path
// 1735

static string factory_normalized_path( const string& path )
{
    string result = z::lcase( path );

    for( uint i = 0; i < result.length(); i++ )
    {
        switch( result[i] )
        {
            case ' ':
            case '_':   result[i] = '-';    break;

            case '%':   result[i] = '@';    break;

            case '\\':  result[i] = '/';    break;

            default:;
        }
    }

    return result;
}

//---------------------------------------------------------------------------Site::text_filename

string Site::text_filename( const string& name, Embed_flags embed_flags )
{
    HRESULT hr;


    bool is_template = (int)embed_flags == 3;   // Sonst Textbaustein, die Factory-Datenbank macht immer noch Unterschiede
    bool is_head     = (int)embed_flags == 7;   // Zwischen Textbaustein, Vorlage und Briefbogen.
                                                

    // Wenn filename mit '|' beginnt, wird der Rest geliefert.
    // Sonst wird _factory_idispatch->"text_filename"( filename ) dynamisch ber IDispatch gerufen.

    //if( is_filename( fn )  &&  !is_absolute_filename( fn ) ) {
    //    fn = "file " + _site->_embed_path + fn;
    //}

    if( c_str( name )[ 0 ] == '|' ) 
    {
        return c_str( name ) + 1;
    }

    if( length( name ) >= 3   
    && (    name[1] == ':'  && ( name[2] == '\\'  ||  name[2] == '/' ) )                        // "c:\"
       || ( name[0] == '\\' || name[0] == '/' ) && ( name[1] == '\\' || name[1] == '/' ) )      // "\\unc"
    {
        return name;
    }


    if( _factory_idispatch  &&  _factory_document_idispatch )
    {
        //if( !_factory_text_filename_dispid ) 
        //{
            Bstr        text_filename_method     = "text_filename";
            Bstr        template_filename_method = "template_filename";
            Bstr        head_filename_method     = "head_filename";

            BSTR        method                   = is_head    ? head_filename_method  :
                                                   is_template? template_filename_method 
                                                              : text_filename_method;
            DISPID      dispid                   = DISPID_UNKNOWN;
            IDispatch*  idispatch                = NULL;


            // Die neue documentfactory.dll (C++) implementiert die Methoden text_filename und template_filename im Objekt Document (fr die Mehrsprachigkeit)
            hr = _factory_document_idispatch->GetIDsOfNames( IID_NULL, &method, 1, STANDARD_LCID, &dispid );
            if( !FAILED( hr ) )
            {
                idispatch = _factory_document_idispatch;
            }
            else
            {
                hr = _factory_idispatch->GetIDsOfNames( IID_NULL, &method, 1, LOCALE_SYSTEM_DEFAULT, &dispid );
                if( !FAILED(hr) )
                {
                    idispatch = _factory_idispatch;
                }
            }
            //_factory_text_filename_dispid = dispid;
        //}

        if( idispatch )
        {
            DISPPARAMS  dispparams;
            Variant     variant_array[ 1 ];
            Variant     result;
            EXCEPINFO   excepinfo;  memset( &excepinfo, 0, sizeof excepinfo );
            uint        error;

            
            variant_array[ 0 ] = SysAllocString_string(name);

            dispparams.rgvarg            = variant_array;     // Array of arguments.
            dispparams.rgdispidNamedArgs = NULL;	          // Dispatch IDs of named arguments.
            dispparams.cArgs             = 1;                 // Number of arguments.
            dispparams.cNamedArgs        = 0;                 // Number of named arguments.

            hr = idispatch->Invoke( dispid, IID_NULL, LOCALE_SYSTEM_DEFAULT, DISPATCH_METHOD, &dispparams, &result, &excepinfo, &error );
            if( FAILED( hr ) )  throw_ole_excepinfo( hr, &excepinfo, "Factory::Invoke", bstr_as_string(method).c_str() );

            result.ChangeType( VT_BSTR );
            return bstr_as_string( V_BSTR( &result ) );
        }
    }


    if( _db_name != "" )    // Fr Andreas's Java-Factory. 1735
    {
        int id;

        {
            z::S id_select;
            id_select << "select `" << ( is_template? "template" : "id" ) << "`";
            id_select << "  from " << ( is_head? "lf_heads" : is_template? "lf_templates" : "lf_texts" ) << "  where ";
            id_select << "`normalized_path`=" << z::sql::quoted( factory_normalized_path( name ) );

            Any_file file ( "-in " + _db_name + id_select );
            if( file.eof() )  throw_xc( "SOS-1463", name );

            id = file.get_record().as_int( 0 );
        }


        int     blob_object_code = is_template? 1 : 2;
        int     blob_nr_code     = 1;
      //string  language         = "de";
        z::S    blob_select;        

        blob_select << "-binary " << _db_name << " -table=LF_BLOBS -blob=VALUE";
        blob_select << " where `object`=" << blob_object_code;
        blob_select <<   " and `id`="     << id;
        blob_select <<   " and `nr`="     << blob_nr_code;

        /*
        if( language != "" )
        {
            S name;
            name << blob_object_code << "/" << id << "/" << blob_nr_code;
            if( rest != "" )  name << "/" << rest;
            blob_select  << " and `name`="   << sql::quoted( blob_name( blob_object_code, id, blob_nr_code, rest ) );
        }
        */

        return blob_select;
    }


    return name;
}

//------------------------------------------------------------------------------Processor::_methods
#ifndef SYSTEM_HAS_COM

const Com_method Processor::_methods[] =
{ 
    { DISPATCH_PROPERTYPUT,  1, "param"                     , (Com_method_ptr)&Processor::put_Param                 , VT_EMPTY    , { VT_BSTR } },
    { DISPATCH_METHOD     ,  2, "make_script"               , (Com_method_ptr)&Processor::Init_engine               , VT_BSTR     , { VT_BSTR } },
    { DISPATCH_METHOD     ,  3, "init_engine"               , (Com_method_ptr)&Processor::Init_engine               },
    { DISPATCH_METHOD     ,  4, "close_engine"              , (Com_method_ptr)&Processor::Close_engine              },
    { DISPATCH_METHOD     ,  5, "global_object"             , (Com_method_ptr)&Processor::get_Global_object         , VT_DISPATCH }, 
    { DISPATCH_METHOD     ,  6, "add_obj"                   , (Com_method_ptr)&Processor::Add_obj                   , VT_EMPTY    , { VT_DISPATCH, VT_BSTR, VT_INT }, 1 }, 
    { DISPATCH_METHOD     ,  7, "parse"                     , (Com_method_ptr)&Processor::Parse                     , VT_VARIANT  , { VT_BSTR, VT_INT } },
    { DISPATCH_METHOD     ,  8, "start_engine"              , (Com_method_ptr)&Processor::Start_engine              },
  //{ DISPATCH_PROPERTYGET,  9, "script"                    , (Com_method_ptr)&Processor::get_Script                , VT_UNKNOWN },
    { DISPATCH_PROPERTYGET, 10, "idispatch"                 , (Com_method_ptr)&Processor::get_Idispatch             , VT_DISPATCH },
    { DISPATCH_METHOD     , 11, "name_exists"               , (Com_method_ptr)&Processor::Name_exists               , VT_BOOL     , { VT_BSTR } },
    { DISPATCH_METHOD     , 12, "preprocess"                , (Com_method_ptr)&Processor::Preprocess                , VT_BSTR     , { VT_BSTR } },
    { DISPATCH_PROPERTYGET, 13, "error_filename"            , (Com_method_ptr)&Processor::get_Error_filename        , VT_BSTR     },
    { DISPATCH_PROPERTYGET, 14, "error_document"            , (Com_method_ptr)&Processor::get_Error_document        , VT_BSTR     },
    { DISPATCH_PROPERTYPUT, 15, "language"                  , (Com_method_ptr)&Processor::put_Language              , VT_EMPTY    , { VT_BSTR } },
    { DISPATCH_PROPERTYGET, 15, "language"                  , (Com_method_ptr)&Processor::get_Language              , VT_BSTR },
    { DISPATCH_PROPERTYPUT, 16, "template_filename"         , (Com_method_ptr)&Processor::put_Template_filename     , VT_EMPTY    , { VT_BSTR } },
    { DISPATCH_PROPERTYGET, 16, "template_filename"         , (Com_method_ptr)&Processor::get_Template_filename     , VT_BSTR },
    { DISPATCH_PROPERTYPUT, 17, "document_filename"         , (Com_method_ptr)&Processor::put_Document_filename     , VT_EMPTY    , { VT_BSTR } },
    { DISPATCH_PROPERTYGET, 17, "document_filename"         , (Com_method_ptr)&Processor::get_Document_filename     , VT_BSTR },
    { DISPATCH_PROPERTYPUT, 18, "head_filename"             , (Com_method_ptr)&Processor::put_Head_filename         , VT_EMPTY    , { VT_BSTR } },
    { DISPATCH_PROPERTYGET, 18, "head_filename"             , (Com_method_ptr)&Processor::get_Head_filename         , VT_BSTR },
  //{ DISPATCH_PROPERTYPUTREF, 19, "parameters"             , (Com_method_ptr)&Processor::putref_Parameters         , VT_EMPTY    , { VT_DISPATCH } },
  //{ DISPATCH_PROPERTYGET, 19, "parameters"                , (Com_method_ptr)&Processor::get_parameters            , VT_DISPATCH },
    { DISPATCH_PROPERTYPUT, 20, "parameter"                 , (Com_method_ptr)&Processor::put_Parameter             , VT_EMPTY    , { VT_BSTR, VT_VARIANT|VT_BYREF } },
  //{ DISPATCH_PROPERTYGET, 20, "parameter"                 , (Com_method_ptr)&Processor::get_Parameter             , VT_DISPATCH , { VT_BSTR } },
    { DISPATCH_METHOD     , 21, "process"                   , (Com_method_ptr)&Processor::Process                   },
  //                        22   eval
  //                        23   close
    { DISPATCH_PROPERTYPUT, 24, "collect"                   , (Com_method_ptr)&Processor::put_Collect               , VT_EMPTY    , { VT_INT } },
    { DISPATCH_PROPERTYGET, 24, "collect"                   , (Com_method_ptr)&Processor::get_Collect               , VT_INT      },
    { DISPATCH_PROPERTYGET, 24, "document_head_modified"    , (Com_method_ptr)&Processor::get_Document_head_modified, VT_BOOL     },
    { DISPATCH_PROPERTYGET, 25, "document_copied"           , (Com_method_ptr)&Processor::get_Document_copied       , VT_BOOL     },
    { DISPATCH_METHOD     , 26, "close_output_file"         , (Com_method_ptr)&Processor::Close_output_file         },
    { DISPATCH_PROPERTYGET, 27, "collected_documents_count" , (Com_method_ptr)&Processor::get_Collected_documents_count, VT_INT    },
    { DISPATCH_PROPERTYPUT, 28, "template_dir"              , (Com_method_ptr)&Processor::put_Template_dir          , VT_EMPTY    , { VT_BSTR } },
    { DISPATCH_PROPERTYGET, 28, "template_dir"              , (Com_method_ptr)&Processor::get_Template_dir          , VT_BSTR },
    { DISPATCH_PROPERTYPUT, 29, "document_dir"              , (Com_method_ptr)&Processor::put_Document_dir          , VT_EMPTY    , { VT_BSTR } },
    { DISPATCH_PROPERTYGET, 29, "document_dir"              , (Com_method_ptr)&Processor::get_Document_dir          , VT_BSTR },
    { DISPATCH_PROPERTYGET, 30, "real_document_filename"    , (Com_method_ptr)&Processor::get_Real_document_filename, VT_BSTR },
    { DISPATCH_PROPERTYGET, 31, "last_step"                 , (Com_method_ptr)&Processor::get_Last_step             , VT_BSTR },
    { DISPATCH_PROPERTYGET, 32, "script_text"               , (Com_method_ptr)&Processor::get_Script_text           , VT_BSTR },
    { DISPATCH_METHOD     , 33, "add_parameters"            , (Com_method_ptr)&Processor::Add_parameters            },
    { DISPATCH_PROPERTYGET, 34, "document_written"          , (Com_method_ptr)&Processor::get_Document_written      , VT_BOOL },
    { DISPATCH_PROPERTYPUT, 35, "db_name"                   , (Com_method_ptr)&Processor::put_Db_name               , VT_EMPTY    , { VT_BSTR } },
    { DISPATCH_PROPERTYGET, 35, "db_name"                   , (Com_method_ptr)&Processor::get_Db_name               , VT_BSTR     },
    { DISPATCH_PROPERTYPUT, 36, "merge_documents"           , (Com_method_ptr)&Processor::put_Merge_documents       , VT_EMPTY    , { VT_BOOL } },
    { DISPATCH_PROPERTYGET, 36, "merge_documents"           , (Com_method_ptr)&Processor::get_Merge_documents       , VT_BOOL     },
    { 0 }
};

#endif
//-------------------------------------------------------------------------Processor::Processor

Processor::Processor()
:
    Sos_ole_object( &factory_processor_class, this ),
    _zero_(this+1)
{
    _site = new Site( this );
    if( !_site )  throw_no_memory_error();
    _site->AddRef();

    _site->init();
}        

//----------------------------------------------------------------------------Processor::~Processor

Processor::~Processor()
{
    Z_LOGI2( "factory.processor", "~Processor\n" );

    try
    {
        Close();

        if( _site ) 
        {
            _site->close_engine();                         // Zirkel auflsen
            _site->Release();  
            _site = NULL;
        }
    }
    catch( const exception& x )  { LOG( __PRETTY_FUNCTION__ << "(): " << x << "\n" ); }
}        

//-------------------------------------------------------------------Processor::put_Merge_documents

STDMETHODIMP Processor::put_Merge_documents( VARIANT_BOOL b )
{ 
    _site->_different_templates_allowed = b != 0;   
    _rtf_context._change_doc_to_section_properties = _site->_different_templates_allowed;     // \landscape durch \lndscpsxn fr jeden Abschnitt ersetzen.
    
    return S_OK; 
}

//-----------------------------------------------------------------------------Processor::put_param

HRESULT Processor::put_Param( BSTR param )
{
    HRESULT hr = NOERROR;

    try {
        //LOGI( "Processor::param := " << param << '\n' );

        Sos_option_iterator opt ( bstr_as_string(param) );
        _site->put_param( &opt );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::param" ); }

    return hr;
}

//-----------------------------------------------------------------------Processor::init_engine

HRESULT Processor::Init_engine()
{
    Z_LOGI2( "factory.processor", "Processor::init_engine \n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = NOERROR;

    try {
        _site->init_engine();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::init_engine" ); }

    return hr;
}

//----------------------------------------------------------------------Processor::close_engine

HRESULT Processor::Close_engine()
{
    Z_LOGI2( "factory.processor", "Processor::close_engine \n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = NOERROR;

    try {
        _site->close_engine();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::close_engine" ); }

    return hr;
}

//----------------------------------------------------------------------Processor::start_engine

HRESULT Processor::Start_engine()
{
    Z_LOGI2( "factory.processor", "Processor::start_engine \n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = NOERROR;

    try {
        _site->start_engine();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::start_engine" ); }

    return hr;
}

//-----------------------------------------------------------------------Processor::get_script

STDMETHODIMP Processor::get_Script( IUnknown** script )
{
    Z_LOGI2( "factory.processor", "Processor::get_script\n" );
    Z_COM_MY_THREAD_ONLY;

    *script = _site->_script;
    if( *script )  (*script)->AddRef(); 
    return NOERROR;
}

//-----------------------------------------------------------------Processor::get_global_object

STDMETHODIMP Processor::get_Global_object( IDispatch** o )
{
    Z_LOGI2( "factory.processor", "Processor::get_global_object\n" );
    Z_COM_MY_THREAD_ONLY;

    HRESULT hr = NOERROR;

    try 
    {
        if( !_site->_constructor )  throw_xc( "FACTORY-114" );
        hr = _site->_constructor->iunknown()->QueryInterface( IID_IDispatch, (void**)o );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::global_object" ); }


    return hr;
}

//---------------------------------------------------------------------------Processor::add_obj

HRESULT Processor::Add_obj( IDispatch* obj, BSTR name_bstr, Script_item_flags flags )
{
    HRESULT hr = NOERROR;

    string name = bstr_as_string(name_bstr);

    Z_LOGI2( "factory.processor", "Processor::add_obj " << name << '\n' );
    Z_COM_MY_THREAD_ONLY;

    try 
    {
        if( name.empty() )  throw_xc( "FACTORY-306" );

        if( !_site->init_engine_called() )  _site->init_engine();

        if( strcmpi( c_str(name), factory_manager_name ) == 0 ) {         // Wird fr Constructor::embed verwendet
            _site->_factory_idispatch = obj;
        }

        if( name == factory_document_name )
            _site->_factory_document_idispatch = obj;


        _site->add_obj( obj, name_bstr, flags );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::add_obj" ); }

    return hr;
}

//-----------------------------------------------------------------------Processor::make_script

HRESULT Processor::Make_script( BSTR rtf_text, BSTR* script_text )
{
    HRESULT hr = NOERROR;
        
    Z_LOGI2( "factory.processor", "Processor::make_script\n ");
    Z_COM_MY_THREAD_ONLY;

    try {

        _site->make_script( bstr_as_string( rtf_text ), "" );

        hr = _site->_script_bstr.CopyTo( script_text );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::make_script" ); }

    return hr;
}

//-----------------------------------------------------------------------Processor::get_script_text

HRESULT Processor::get_Script_text( BSTR* script_text )
{
    HRESULT hr = NOERROR;
        
    Z_LOGI2( "factory.processor", "Processor::get_script_text\n ");
    Z_COM_MY_THREAD_ONLY;

    try 
    {
        prepare_process();

        hr = _site->_script_bstr.CopyTo( script_text );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::script_text" ); }

    return hr;
}

//-----------------------------------------------------------------------Processor::preprocess

HRESULT Processor::Preprocess( BSTR script, BSTR* result_script )
{
    HRESULT hr = NOERROR;
        
    Z_LOGI2( "factory.processor", "Processor::preprocess\n ");
    Z_COM_MY_THREAD_ONLY;

    try {
        *result_script = _site->preprocess( script );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::preprocess" ); }

    return hr;
}

//-----------------------------------------------------------------------------Processor::parse

HRESULT Processor::Parse( BSTR script_text, ::Scripttext_flags flags, VARIANT* result )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    Z_MUTEX( hostware_mutex )
    try {
        _site->parse( script_text, flags, result );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::parse" ); }

    return hr;
}

//-------------------------------------------------------------------------Processor::idispatch

HRESULT Processor::get_Idispatch( BSTR name, IDispatch** result )
{
    HRESULT hr;

    Z_LOG2( "factory.processor", "Processor::get_idispatch " << name << '\n' );

    Z_COM_MY_THREAD_ONLY;

    hr = _site->_script->GetScriptDispatch( name && name[0]? name : NULL, result );

    return hr;
}

//-----------------------------------------------------------------------Processor::name_exists
// Vgl. Script_object::obj_name_exists

HRESULT Processor::Name_exists( BSTR sub_name, VARIANT_BOOL* result )
{
    HRESULT     hr;
    OLECHAR*    names [ 1 ];
    IDispatch*  idispatch = NULL;
    DISPID      dispid;
    Bstr        object_name_bstr;
    OLECHAR*    punkt;

    Z_LOG2( "factory.processor", "Processor::name_exists " << sub_name << '\n' );

    punkt = (OLECHAR*)wmemchr( sub_name, L'.', SysStringLen( sub_name ) );
    if( punkt ) {
        object_name_bstr.assign( sub_name, punkt - sub_name );
        names[0] = punkt + 1;
    }
    else 
    {
        names[0] = sub_name;
    }

    hr = get_Idispatch( object_name_bstr, &idispatch );
    if( FAILED( hr ) )  return hr;

    hr = idispatch->GetIDsOfNames( IID_NULL, names, 1, LOCALE_SYSTEM_DEFAULT, &dispid );
    idispatch->Release();  
    idispatch = NULL;

    if( FAILED( hr ) ) 
    {
        if( hr != DISP_E_UNKNOWNNAME )  return hr;
        hr = NOERROR;
        *result = 0;
    } 
    else 
    {
        *result = -1;
    }

    return hr;
}

//--------------------------------------------------------------------Processor::get_error_filename

STDMETHODIMP Processor::get_Error_filename( BSTR* filename_bstr )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try
    {
        if( !_site->_constructor )  //throw_xc( "FACTORY-116 ");
        {
            *filename_bstr = NULL;
            return FALSE;
        }

        if( _site->_script_exception ) 
        {
            _site->_template_filename_changed = true;
            _site->_constructor->set_error( *_site->_script_exception  );
            SOS_DELETE( _site->_script_exception );
        }

        *filename_bstr = SysAllocString_string( _site->_constructor->error_filename() );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::error_filename" ); }

    return hr;
}

//--------------------------------------------------------------------Processor::get_error_document

STDMETHODIMP Processor::get_Error_document( BSTR* doc_bstr )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try
    {
        if( !_site->_constructor )  //throw_xc( "FACTORY-116 ");
        {
            *doc_bstr = NULL;
            return FALSE;
        }

        if( _site->_script_exception ) 
        {
            _site->_template_filename_changed = true;
            _site->_constructor->set_error( *_site->_script_exception  );
            SOS_DELETE( _site->_script_exception );
        }

        *doc_bstr = SysAllocString_string( _site->_constructor->error_document() );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::error_document" ); }

    return hr;
}

//--------------------------------------------------------------------------Processor::put_language

STDMETHODIMP Processor::put_Language( BSTR language_bstr )
{
    if( !_site )  return E_POINTER;

    Z_COM_MY_THREAD_ONLY;

    try
    {
        string language = bstr_as_string( language_bstr );
        //fprintf( stderr, "Processor::put_language %s\n", language.c_str() );
        if( !language.empty() )  _language_set = true;
        _site->set_engine_name( language );
    }
    catch( const exception& x )  { return _set_excepinfo( x, "hostWare.Factory_processor::language" ); }

    return NOERROR;
}

//--------------------------------------------------------------------------Processor::get_language

STDMETHODIMP Processor::get_Language( BSTR* result )
{
    if( !_site )  return E_POINTER;

    *result = SysAllocString_string( _site->_engine_name );

    return NOERROR;
}

//-----------------------------------------------------------------Processor::put_template_filename

STDMETHODIMP Processor::put_Template_filename( BSTR template_filename_bstr )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try
    {
        _site->_document_written = false;

        string template_filename = make_absolute_filename( _template_dir, bstr_as_string( template_filename_bstr ) );

        _site->_template_filename_changed |= template_filename.empty()  ||  _site->_template_filename != template_filename;
        
        if( _site->_template_filename_changed  &&  !_site->_different_templates_allowed )  _site->close_output_file();

        _site->_template_filename = template_filename;
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::template_filename" ); }

    return hr;
}

//-----------------------------------------------------------------Processor::get_template_filename

STDMETHODIMP Processor::get_Template_filename( BSTR* result )
{
    Z_COM_MY_THREAD_ONLY;

    *result = SysAllocString_string( _site->_template_filename );
    return NOERROR;
}

//---------------------------------------------------------------------Processor::put_head_filename

STDMETHODIMP Processor::put_Head_filename( BSTR head_filename_bstr )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try
    {
        _site->_document_written = false;

        string head_filename = make_absolute_filename( _template_dir, bstr_as_string( head_filename_bstr ) );

        _site->_template_filename_changed |= _site->_head_filename != head_filename;
        
        if( _site->_template_filename_changed  &&  !_site->_different_templates_allowed )  _site->close_output_file();

        _site->_head_filename = head_filename;
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::template_filename" ); }

    return hr;
}

//-----------------------------------------------------------------Processor::put_document_filename

STDMETHODIMP Processor::put_Document_filename( BSTR document_filename_bstr )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try
    {
        string document_filename = make_absolute_filename( _document_dir, bstr_as_string( document_filename_bstr ) );
        
        if( _site->_output_filename != document_filename )  
        {
            _site->close_output_file();
            _site->_document_nr = 0;
        }

        _site->_output_filename = document_filename;

        if( _site->_output_filename == "" )  throw_xc( "FACTORY-307" );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::document_filename" ); }

    return hr;
}

//-----------------------------------------------------------------Processor::get_document_filename

STDMETHODIMP Processor::get_Document_filename( BSTR* result )
{
    Z_COM_MY_THREAD_ONLY;

    *result = SysAllocString_string( _site->_output_filename );
    return NOERROR;
}

//---------------------------------------------------------------------Processor::putref_parameters

STDMETHODIMP Processor::putref_Parameters( Ivariables* variables )
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try
    {
        ptr<Hostware_variables> v;

        if( !variables )  return E_POINTER;

        HRESULT hr = variables->QueryInterface( IID_Variables, v.void_pp() );
        if( FAILED(hr) )  return hr;

        if( _site->_parameters_added )  throw_xc( "SOS-1429" );

        _site->_variables = v;
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::parameters" ); }

    return hr;
}

//------------------------------------------------------------------------Processor::get_parameters

STDMETHODIMP Processor::get_Parameters( Ivariables** result )
{
    *result = _site->_variables;
    (*result)->AddRef();
    return NOERROR;
}

//-------------------------------------------------------------------------Processor::put_parameter

STDMETHODIMP Processor::put_Parameter( BSTR name, VARIANT* value )
{
    //string s = variant_as_string(*value);
    //fprintf( stderr, "Processor::put_parameter %s, %s\n", bstr_as_string(name).c_str(), s.c_str() );

    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try
    {
        if( _site->_parameters_added )  throw_xc( "SOS-1429" );
        hr = _site->_variables->put_Value( name, value );
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::error_document" ); }

    return hr;
}

//-------------------------------------------------------------------------Processor::get_parameter

STDMETHODIMP Processor::get_Parameter( BSTR name, Ivariable** result )
{
    Z_COM_MY_THREAD_ONLY;

    return _site->_variables->get_Value( name, result );
}

//------------------------------------------------------------------------Processor::add_parameters
// Wird auch von process() gerufen.
// Laut Andreas werden von parse() die mit set_parameter() gesetzten Variablen nicht erkannt. 
// Deshalb knnen die Parameter hier mit add_parameters() der Scripting Engine bergeben werden.
// Fr PerlScript und VBScript aber nicht nachvollziehbar, die erkennen auch nachtrglich gesetzte Variablen.
// jz 10.12.03

STDMETHODIMP Processor::Add_parameters()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        add_parameters_();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::add_parameters" ); }

    return hr;
}

//-----------------------------------------------------------------------Processor::add_parameters_

void Processor::add_parameters_()
{
    //if( !_add_parameters_called )
    {
        string rtf_filename;

        if( !_site->init_engine_called() )  _site->init_engine();

        if( _site->_template_filename_changed ) 
        {
            if( !_site->_head_filename.empty() )
            {
                Bstr    factory_template = "factory_template";
                Bstr    bstr = _site->_template_filename.c_str();
                Variant v = (BSTR)bstr;
                _site->_variables->put_Value( factory_template, &v );
            }
        }

        _site->add_parameters();

        //_add_parameters_called = true;
    }
}

//-----------------------------------------------------------------------Processor::prepare_process

void Processor::prepare_process()
{
    if( !_prepare_process_called )
    {
        _site->_document_written = false;

        add_parameters_();

        if( _site->_template_filename_changed ) 
        {
            string rtf_filename;

            if( _site->_head_filename.empty() )  rtf_filename = _site->_template_filename;
                                           else  rtf_filename = _site->_head_filename;

            _last_step = "make_script";
            _site->make_script( file_as_string( rtf_filename ), rtf_filename ); 
            _last_step = "";

            _site->_template_filename_changed = false;
        }

        _prepare_process_called = true;
    }
}

//-------------------------------------------------------------------------------Processor::process

void Processor::process()
{
    HRESULT hr;

    Z_LOGI2( "factory.processor", "Processor::process\n" );

    try 
    {
        ptr<IDispatch>  constructor;
        Variant         result;

        if( _site->_output_filename == "" )  throw_xc( "FACTORY-307" );

        prepare_process();
        _prepare_process_called = false;    // Fr den nchsten Aufruf von process() (Serienfactory)

        hr = get_Global_object( constructor.pp() );
        if( FAILED(hr) )  throw_ole( hr, "Processor::global_object" );

        Bstr constructor_bstr = SCRIPT_CONSTRUCTOR_NAME;
        _site->add_obj( constructor, constructor_bstr, Script_item_flags( sif_isvisible ) );

        // Gelegentlich ausbauen:
        Bstr global_bstr = "factory_global";
        _site->add_obj( constructor, global_bstr, Script_item_flags( sif_isvisible | ( _site->_is_unix_perlscript? 0 : sif_globalmembers ) ) );


        if( _site->_is_unix_perlscript )   // sif_globalmembers funktioniert nicht. Also embed() hier bekanntgeben:
        {
            _site->parse( Bstr("sub embed { $" SCRIPT_CONSTRUCTOR_NAME "->embed(@_); }"), ::scripttext_isvisible, &result );
        }


        if( !_site->_is_unix_perlscript )  _site->start_engine();

        _last_step = "template";
        if( _site->_source.empty() )  _site->_source = "Skript aus " + _site->_template_filename;
        _site->parse( _site->_script_bstr, ::scripttext_isvisible, &result );
        _site->_source = "";
        _last_step = "";
        
        if( _site->_is_unix_perlscript )  _site->start_engine();       // Beim direkten Aufruf von Perl darf nur _ein_ perl_parse() ausgefhrt werden.

        _site->close_engine();

        //_site->_template_filename_changed = false;
    }
    catch( const exception& )
    { 
        _site->close_engine(); 
        throw; 
    }
}

//-------------------------------------------------------------------------------Processor::process

STDMETHODIMP Processor::Process()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;
    try 
    {
        process();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::process" ); }

/*
static _CrtMemState memstate;
static bool state_filled = false;
if( state_filled )  _CrtMemDumpAllObjectsSince(&memstate);
_CrtMemCheckpoint(&memstate), state_filled = true;
*/

    return hr;
}

//---------------------------------------------------------------------------------Processor::Close

STDMETHODIMP Processor::Close()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    //HRESULT hr2 = Close_output_file();        Nicht automatisch schreiben. Im Fehlerfall entscheidet der Aufrufer

    try 
    {
        Z_LOGI2( "factory.processor", "Processor::close\n" );

        if( _site ) 
        {
            _site->close_engine();
            if( _site->_constructor )  _site->_constructor->close();
        }
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::close" ); }

    return hr;
}

//---------------------------------------------------------------------Processor::close_output_file

STDMETHODIMP Processor::Close_output_file()
{
    HRESULT hr = NOERROR;

    Z_COM_MY_THREAD_ONLY;

    try 
    {
        if( _site )  _site->close_output_file();
    }
    catch( const exception& x )  { hr = _set_excepinfo( x, "hostWare.Factory_processor::close_output_file" ); }

    return hr;
}

//------------------------------------------------------------Processor::get_document_head_modified

STDMETHODIMP Processor::get_Document_head_modified( VARIANT_BOOL* b )
{ 
    *b = false;

    Z_COM_MY_THREAD_ONLY;

    if( _site  &&  _site->_constructor  &&  typeid( *_site->_constructor ) == typeid(Rtf_constructor) ) 
    {
        Rtf_constructor* c = static_cast<Rtf_constructor*>( +_site->_constructor );
        *b = c->_doc->_doc_properties_changed;
    }

    return NOERROR;
}

//----------------------------------------------------------------Processor::check_document_written
/*
void Processor::check_document_written( int next_record_nr )
{
    if( _document_has_to_be_written  &&  _processor->collected_documents_count() == 0 )
    {
        if( _output_object )
        {
            if( com_name_exists( _output_object, "factory_output_written" ) )
            {
                Z_LOG2( "factory.processor", "factory_output_written()\n" );
                com_call( _output_object, "factory_output_written", Variant( Bstr( _processor->real_document_filename().c_str() ) ), 
                                                                    Variant( _processor->last_collected_documents_count() ) );
            }
            _output_object = NULL;
        }

        if( _rerun_file  &&  !_rerun )  write_rerun_number( next_record_nr );

        _document_has_to_be_written = false;
    }
}
*/
//----------------------------------------------------------------------Processor::document_written

STDMETHODIMP Processor::get_Document_written( VARIANT_BOOL* result )
{
    HRESULT hr = S_OK;

    Z_COM_MY_THREAD_ONLY;

    *result = _site && _site->_document_written;
/*

    try
    {
        Z_LOG2( "factory.processor", __PRETTY_FUNCTION__ << '\n' );

        hr = E_NOTIMPL;
    }
    catch( const exception&  x )  { return _set_excepinfo( x, __FUNCTION__ ); }
    catch( const _com_error& x )  { return _set_excepinfo( x, __FUNCTION__ ); }
*/
    return hr;
}

//------------------------------------------------------------------------Processor::connect_script
/*
HRESULT Processor::connect_script()
{
    HRESULT hr = NOERROR;

    try {
        _site->connect_script();
    }
    catch( const Xc& x )  { hr = _set_excepinfo( x ); }

    return hr;
}
*/

} //namespace factory
} //namespace sos

//-------------------------------------------------------------------------create_factory_processor

HRESULT Create_factory_processor( const CLSID& clsid, IUnknown** result )
{                                                                            
    if( clsid != CLSID_Factory_processor )  return CLASS_E_CLASSNOTAVAILABLE;
    *result = (IUnknown*)(Ifactory_processor*) new sos::factory::Processor;
    return S_OK;
}

//-------------------------------------------------------------------------------------------------

