// $Id: factory_processor.h 3546 2005-04-03 08:30:29Z jz $

// 1735


#ifndef __SOS_FACTORY_H
#define __SOS_FACTORY_H

#include "../hostole/hostole.h"

//#ifdef SYSTEM_WIN
//
//#   if defined _DEBUG
//#       import "../../hostole/debug/hostole.tlb"   no_namespace raw_interfaces_only named_guids
//#    else
//#       import "../../hostole/release/hostole.tlb" no_namespace raw_interfaces_only named_guids
//#   endif
//
//#endif

#include "../kram/scripting_engine.h"
#include "../kram/sosscrpt.h"
#include "../kram/sosopt.h"
#include "../zschimmer/rtf/rtf.h"
#include "../hostole/variables.h"

namespace sos {

extern Typelib_descr hostole_typelib;

namespace factory {

struct Site;
struct Synthesizer;
struct Parser;
struct Processor;

//-----------------------------------------------------------------------------------Constructor
// Objekt, das vom Skript benutzt wird, um das Dokument zu generieren.
// embed() wird von diesem Objekt bereitgestellt.
// Oberklasse fr Text_constructor, Rtf_constructor

struct Constructor : Sos_ole_object, 
                     z::My_thread_only
{
    void*                       operator new            ( size_t size )                         { return sos_alloc( size, "factory::Constructor" ); }
    void                        operator delete         ( void* ptr )                           { sos_free( ptr ); }


                                Constructor             ( Ole_class_descr*, IUnknown* this_ );
    virtual                    ~Constructor             ()  {}

    STDMETHODIMP                Put_bstr                ( BSTR, VARIANT_BOOL raw );
    STDMETHODIMP                Put                     ( VARIANT*, VARIANT_BOOL raw );
    STDMETHODIMP                Put_formatted           ( VARIANT*, Ausgabeformat format );
    STDMETHODIMP                Embed                   ( BSTR, enum Embed_flags = (Embed_flags)0 );

    virtual void                close                   ()                          {}
    virtual void                rtf_prop                ( int, int )                { throw_xc( "FACTORY-211" ); }
    virtual void                rtf_put                 ( int, int )                { throw_xc( "FACTORY-211" ); }
    virtual void                rtf_open                ( int )                     { throw_xc( "FACTORY-211" ); }
    virtual void                rtf_close               ()                          { throw_xc( "FACTORY-211" ); }
    virtual void                start                   ()                          = 0;
    virtual void                finish                  ()                          = 0;
    virtual void                set_error               ( const Xc& )               {}            
    virtual string              error_filename          ()                          { return ""; }          // throw_xc( "FACTORY-115" ); 
    virtual string              error_document          ()                          { return ""; }          // throw_xc( "FACTORY-115" ); 
    virtual void                close_output_file       ()                          {}

    IDispatch*                  idispatch               ()                          { return dynamic_cast<IDispatch*>( iunknown() ); }

    virtual void                put_string              ( const string&, bool raw ) = 0;

    Fill_zero                  _zero_;
    Site*                      _site;
    Parser*                    _parser;
    bool                       _ok;                     // Ohne Exception bis zum Ende durchgelaufen?
};

//-----------------------------------------------------------------------------------------Parser
// Erkennt <%...%>, &(...) und #prprozessor.
// Oberklasse fr Text_parser, Rtf_parser

struct Parser : Sos_self_deleting//Sos_object
{
    static Sos_ptr<Parser>      create                  ( Site*, Synthesizer*, const char* first_bytes_of_script,
                                                          const string& doc_name = "" );


                                Parser                  ()                          : _zero_(this+1) {}
    virtual                    ~Parser                  ();

    void                        set_constructor         ( Constructor* );
    virtual void                init                    () = 0;
    virtual void                merge                   ()                                          {}
    virtual void                parse                   ( const char* script_text ) = 0;

    Fill_zero                  _zero_;
    Site*                      _site;
    Sos_ptr<Synthesizer>       _synth;
    Parser*                    _main_parser;            // Parser, der oberstes #include ausfhrt, sonst NULL (auch bei embed)
    ptr<Constructor>           _constructor;
    bool                       _include;
    zschimmer::rtf::Flags      _allowed_properties;
    bool                       _restore_properties;     // Anfngliche Eigenschaften am Ende wiederherstellen
    string                     _document_name;
};


//-----------------------------------------------------------------------------------Synthesizer

struct Synthesizer : Sos_self_deleting,
                     z::My_thread_only
{
    struct Length
    {
                                Length                  ( int length, int lineno )                : _length(length), _next_lineno(lineno) {}

        int                    _length;
        int                    _next_lineno;
    };


                                Synthesizer             () : _zero_(this+1) { obj_const_name( "Factory::Synthesizer" ); }

    virtual bool                is_script_synthesizer   () const                                    { return false; }

    virtual void                put_script_line         ( const Sos_string& script_code )           = 0;    // Generiert direkt Skript-Code
    virtual void                put_preprocessor        ( const Sos_string& statement, bool text_script = true ) = 0;
    virtual void                put_text                ( const Sos_string& string, bool raw )      = 0;    // Text zu RTF: "{" -> "\{", Lnge beliebig
    virtual void                put_expr                ( const Sos_string&, Ausgabeformat )        = 0;
    virtual void                rtf_prop                ( int, int )                                {}
    virtual void                rtf_put                 ( int, int )                                { throw_xc( "FACTORY-211" ); }
    virtual void                rtf_open                ( int )                                     { throw_xc( "FACTORY-211" ); }
    virtual void                rtf_close               ()                                          { throw_xc( "FACTORY-211" ); }
    virtual void                rtf_start               ()                                          = 0;
    virtual void                rtf_finish              ()                                          = 0;

    virtual Length              length                  ()                                          { throw_xc( "Synthesizer::length" ); return Length(0,0); }
    virtual void                truncate                ( const Length& )                           { throw_xc( "Synthesizer::truncate" ); }


    Fill_zero                  _zero_;
    Site*                      _site;
    int                        _include_nesting;        // Verschachtelungsgrad der #include
    int                        _next_lineno;            // Fr Script_synthesizer::put_script_line(), sonst 0
};

//---------------------------------------------------------------------------Script_synthesizer
// Der Script_synthesizer stellt das Skript zusammen.
// put_expr() generiert eine Skript-Anweisung mit dem Ausdruck
// put_preprocessor() interpretiert #include

struct Script_synthesizer : Synthesizer
{
                                Script_synthesizer      ( Site*, Area* );

    virtual bool                is_script_synthesizer   () const                                    { return true; }

    void                        put_text                ( const Sos_string& string, bool raw );
    void                        put_expr                ( const Sos_string&, Ausgabeformat );
    void                        put_expr2               ( const Sos_string&, Ausgabeformat );
    void                        put_script_line         ( const Sos_string& script_code );              // Generiert direkt Skript-Code
    void                        put_preprocessor        ( const Sos_string& statement, bool text_include = true );
    void                        rtf_prop                ( int prop, int value );
    void                        rtf_put                 ( int start, int ende );
    void                        rtf_start               ();
    void                        rtf_finish              ();
    void                        rtf_open                ( int entity_ptr );
    void                        rtf_close               ();
    void                        rtf_include_begin       ();
    void                        rtf_include_end         ();

    
    void                        include_                ( const char*, bool text_include );  //  #include, #include_script
    void                        include_file            ( const Sos_string&, bool text_include, Embed_flags );
    void                        include_script          ( const Sos_string& script );

    Length                      length                  ()                                          { return Length( _buffer->length(), _next_lineno ); }
    void                        truncate                ( const Length& length )                    { _buffer->length( length._length );  _next_lineno = length._next_lineno; }


    Fill_zero                  _zero_;
    string                     _indent;
    Area* const                _buffer;                 // ASCII-Skript (VBScript) = Site::_script_buffer
};

//----------------------------------------------------------------------------Embed_synthesizer
// Fr Constructor::embed()
// Whrend das mit Script_synthesizer zusammengestellte Skript abluft kann embed() ausgefhrt werden.
// Der embed-Text wird mit dieser Klasse eingebettet.
// Embed_synthesizer generiert kein Skript, sondern fhrt die Anweisungen _sofort_ aus, 
// indem sie an den Constructor bergeben werden.
// put_expr() ruft direkt die Scripting Engine zur Auswertung des Ausdrucks aus.
// put_script_line() und put_preprocessor() sind gesperrt.

struct Embed_synthesizer : Synthesizer
{
                                Embed_synthesizer       ( Site*, Constructor* );

    void                        put_text                ( const Sos_string& string, bool raw );
    void                        put_expr                ( const Sos_string&, Ausgabeformat );
    void                        put_script_line         ( const Sos_string& script_code );                  // Generiert direkt Skript-Code
    void                        put_preprocessor        ( const Sos_string& statement, bool text_script );
    void                        rtf_prop                ( int prop, int value );
    void                        rtf_put                 ( int, int );
    void                        rtf_open                ( int entity_ptr );
    void                        rtf_close               ();
    void                        rtf_start               ();
    void                        rtf_finish              ();
    
    void                        include_script          ( const char* );                                // #include ...
    void                        include_script_file     ( const Sos_string& );



    Fill_zero                  _zero_;
    ptr<Constructor>           _constructor;
};

//-----------------------------------------------------------------------------------------Site
// Site fr Scripting Engine.
// Das Objekt hlt alles zusammmen.

struct Site : Script_site,
              z::My_thread_only
{
    void*                       operator new            ( size_t size )                                 { return sos_alloc( size, "Site" ); }
    void                        operator delete         ( void* ptr )                                   { sos_free( ptr ); }

                                Site                    ( Processor* );
                               ~Site                    ();

    void                        init                    ();
    void                        set_engine_name         ( const string& );
    void                        start_engine            ();
    void                        close_engine            ();
    void                        add_parameters          ();
    void                        put_param               ( Sos_option_iterator* );
    void                        make_script             ( const string& rtf, const string& template_name );    // Konvertiert rtf nach Skript
    BSTR                        preprocess              ( BSTR );
    void                        parse                   ( BSTR script_text, ::Scripttext_flags, VARIANT* );  // Skript in die Scripting Engine einlesen 
    Sos_string                  text_filename           ( const Sos_string& name, Embed_flags );
    void                        close_output_file       ()                                              { if( _constructor )  _constructor->close_output_file(); }


    Fill_zero                  _zero_; 
    Sos_ptr<Parser>            _main_parser;            // Parser des Hauptdokuments
    ptr<Constructor>           _constructor;

    bool                       _is_unix_perlscript;

    string                     _script_constructor_prefix;  // z.B: "factory_global."
    char                       _quote_char;             // Zeichen fr String
    string                     _quotes;                 // Alle Zeichen fr alle Arten von Strings (Perl: "'`, VBScript: ").
    bool                       _use_backslash;          // Escape-Zeichen in Strings
    bool                       _use_call_parentheses;   // Bei Unterprogrammaufruf Parameter klammern
    bool                       _use_semicolon;          // Skriptsprache verlangt Semikolon am Ende einer Anweisung
    char                       _var_char;               // '$' (oder '&')
    char                       _var_prefix;             // In Perl: '$'
    bool                       _rtf_single_lines;       // rtf::Doc._single_lines (Jede RTF-Anweisung auf ne eigene Zeile)
    int                        _rtf_indent;             // rtf::Doc._rtf_intent   (RTF-Dokument hbsch einrcken)
    string                     _comment;
    IDispatch*                 _factory_idispatch;          // (alt) VB-Object Factory mit Methode text_filename(path) fr embed/#include
    IDispatch*                 _factory_document_idispatch; // (neu) C++--Object Document mit Methode text_filename(path) fr embed/#include
    string                     _db_name;                // Fr Andreas' Java-Factory (1735)
  //DISPID                     _factory_text_filename_dispid;
    Sos_string                 _output_filename;        // Ausgabe des fertigen RTF-Textes (durch die Scripting Enging), mit ### fr lfd Nr
    Sos_string                 _real_output_filename;
  //bool                       _output_filename_changed;// Wenn sich der Name nicht gendert hat, alle Dokumente zu einem Dokument zusammenfassen
    int                        _collect;
    Dynamic_area               _script_buffer;
    int4                       _script_size;
    Bstr                       _script_bstr;
    Sos_string                 _factory_texts;          // Weil VBScript 2.0 keine Expressions auswertet, wird die Variable hier gespeichert
    Bool                       _factory_texts_valid;
    ptr<Hostware_variables>    _variables;       
    bool                       _parameters_added;
    string                     _template_filename;
    bool                       _template_filename_changed;  // Auch wenn _head_filename wechselte
    string                     _head_filename;
    int                        _collected_documents_count;      // Bei zusammengefassten Dokumenten die bisherige Anzahl
    int                        _last_collected_documents_count; // Anzahl der zusammengefassten Dokumente in der letzten Ausgabedatei
    int                        _document_nr;            // Nummer des Dokuments, fr #### in _output_filename
    bool                       _document_copied;        // Falls wegen genderten Dokumentkopfs das Dokument umkopiert werden musste
    bool                       _document_written;
    bool                       _suppress_script;
    bool                       _different_templates_allowed;
    Processor*                 _processor;
};

//----------------------------------------------------------------------------------------Processor

struct Processor : Ifactory_processor,
                   Sos_ole_object,
                   z::My_thread_only
{
    void*                       operator new            ( size_t size )                             { return sos_alloc( size, "factory::Processor" ); }
    void                        operator delete         ( void* ptr )                               { sos_free( ptr ); }

                                Processor               ();
    virtual                    ~Processor               ();

    USE_SOS_OLE_OBJECT

    STDMETHODIMP            put_Param                   ( BSTR );
    STDMETHODIMP                Make_script             ( BSTR rtf_text, BSTR* script_text );       // Konvertiert rtf nach Skript
    STDMETHODIMP                Init_engine             ();
    STDMETHODIMP                Close_engine            ();
    STDMETHODIMP                Start_engine            ();
    STDMETHODIMP            get_Global_object           ( IDispatch** );
    STDMETHODIMP                Add_obj                 ( IDispatch*, BSTR, Script_item_flags );
    STDMETHODIMP                Parse                   ( BSTR script, ::Scripttext_flags, VARIANT* );
    STDMETHODIMP                Eval                    ( BSTR source, ::Scripttext_flags flags, VARIANT* result ) { return Parse( source, ::Scripttext_flags( flags | ::scripttext_isexpression ), result ); }
    STDMETHODIMP                Parse_file              ( BSTR filename );           // Skript in die Scripting Engine einlesen 
  //STDMETHODIMP                Connect_script          ();
    STDMETHODIMP            get_Script                  ( IUnknown** );     // Besser: IActiveScript*
    STDMETHODIMP            get_Idispatch               ( BSTR name, IDispatch** );
    STDMETHODIMP                Name_exists             ( BSTR sub_name, VARIANT_BOOL* );
    STDMETHODIMP                Preprocess              ( BSTR script, BSTR* result_script );
    STDMETHODIMP            get_Error_filename          ( BSTR* filename );
    STDMETHODIMP            get_Error_document          ( BSTR* doc );

    STDMETHODIMP            put_Language                ( BSTR );
    STDMETHODIMP            get_Language                ( BSTR* );
    STDMETHODIMP            get_Template_dir            ( BSTR* s )                                 { *s = SysAllocString_string( _template_dir ); return NOERROR; }
    STDMETHODIMP            put_Template_dir            ( BSTR s )                                  { _template_dir = bstr_as_string( s ); return NOERROR; }
    STDMETHODIMP            put_Template_filename       ( BSTR );
    STDMETHODIMP            get_Template_filename       ( BSTR* );
    STDMETHODIMP            put_Head_filename           ( BSTR );
    STDMETHODIMP            get_Head_filename           ( BSTR* s )                                 { *s = SysAllocString_string( _site->_head_filename ); return NOERROR; }
    STDMETHODIMP            get_Document_dir            ( BSTR* s )                                 { *s = SysAllocString_string( _document_dir ); return NOERROR; }
    STDMETHODIMP            put_Document_dir            ( BSTR s )                                  { _document_dir = bstr_as_string( s ); return NOERROR; }
    STDMETHODIMP            put_Document_filename       ( BSTR );
    STDMETHODIMP            get_Document_filename       ( BSTR* );
    string                      real_document_filename  ()                                          { return _site->_real_output_filename; }
    STDMETHODIMP            get_Real_document_filename  ( BSTR* s )                                 { *s = SysAllocString_string( real_document_filename() ); return NOERROR; }
    STDMETHODIMP         putref_Parameters              ( Ivariables* );
    STDMETHODIMP            get_Parameters              ( Ivariables** );
    STDMETHODIMP            put_Parameter               ( BSTR name, VARIANT* );
    STDMETHODIMP            get_Parameter               ( BSTR name, Ivariable** );
    void                        add_parameters_         ();
    STDMETHODIMP                Add_parameters          ();
    void                        process                 ();
    STDMETHODIMP                Process                 ();
    STDMETHODIMP                Close                   ();
    STDMETHODIMP            put_Collect                 ( int collect )                             { _site->_collect = collect; return NOERROR; }
    STDMETHODIMP            get_Collect                 ( int* collect )                            { *collect = _site->_collect; return NOERROR; }
    STDMETHODIMP                Close_output_file       ();
    int                         last_collected_documents_count()                                    { return _site->_last_collected_documents_count; }
    int                         collected_documents_count()                                         { return _site->_collected_documents_count; }
    STDMETHODIMP            get_Collected_documents_count( int* count )                             { *count = collected_documents_count(); return NOERROR; }
    STDMETHODIMP            get_Document_head_modified   ( VARIANT_BOOL* );
    STDMETHODIMP            get_Document_copied         ( VARIANT_BOOL* b )                         { *b = _site->_document_copied; return NOERROR; }
    STDMETHODIMP            get_Last_step               ( BSTR* s )                                 { *s = SysAllocString_string(_last_step); return NOERROR; }
    STDMETHODIMP            get_Script_text             ( BSTR* );                                  // Konvertiert rtf nach Skript
    void                        prepare_process         ();
    STDMETHODIMP            get_Document_written        ( VARIANT_BOOL* );
    STDMETHODIMP            put_Db_name                 ( BSTR db_name )                            { _site->_db_name = z::com::string_from_bstr( db_name ) + " ";  return S_OK; }
    STDMETHODIMP            get_Db_name                 ( BSTR* result )                            { return z::com::String_to_bstr( _site->_db_name, result ); }
    STDMETHODIMP            put_Merge_documents         ( VARIANT_BOOL b );
    STDMETHODIMP            get_Merge_documents         ( VARIANT_BOOL* result )                   { *result = _site->_different_templates_allowed;  return S_OK; }


    Fill_zero                  _zero_;
    Site*                      _site;
    bool                       _language_set;
    string                     _template_dir;
    string                     _document_dir;
    string                     _last_step;
    bool                       _prepare_process_called;
    bool                       _add_parameters_called;

    zschimmer::rtf::Rtf_context  _rtf_context;
};

} //namespace sos
} //namespace factory

#endif
