// $Id: java_Process_with_pid.cxx 4016 2005-11-09 11:57:30Z jz $

#include "hostjava_common.h"
#include "../zschimmer/z_process.h"

using namespace std;
using namespace zschimmer;
using namespace zschimmer::com;
using namespace zschimmer::java;
using namespace sos::hostjava;

//---------------------------------------------------------------------------Process_with_pid.close

extern "C"
JNIEXPORT void JNICALL Java_sos_hostware_Process_1with_1pid_close( JNIEnv* jenv, jobject, jlong process_ptr )
{
    try
    {
        Process* process = (Process*)(size_t)process_ptr;

        if( process )  process->Release();
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }
}

//---------------------------------------------------------------------------Process_with_pid.start

extern "C"
JNIEXPORT jlong JNICALL Java_sos_hostware_Process_1with_1pid_start( JNIEnv* jenv, jobject, jstring cmd_line_jstr )
{
    jlong result = 0;

    try
    {
        ptr<Process> process = Z_NEW( Process );

        process->start( string_from_jstring( jenv, cmd_line_jstr ) );

        result = (jlong)(size_t)process.take();
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }

    return result;
}

//---------------------------------------------------------------------Process_with_pid.start_array

extern "C"
JNIEXPORT jlong JNICALL Java_sos_hostware_Process_1with_1pid_start_1array( JNIEnv* jenv, jobject, jobjectArray cmd_array_j )
{
    jlong result = 0;

    try
    {
        ptr<Process> process = Z_NEW( Process );
        int len = jenv->GetArrayLength( cmd_array_j );
        vector<string> cmd_array ( len );

        for( int i = 0; i < len; i++ )
        {
            jstring jparam = static_cast<jstring>( jenv->GetObjectArrayElement( cmd_array_j, i ) );
            cmd_array[ i ] = string_from_jstring( jenv, jparam );
            jenv->DeleteLocalRef( jparam );
        }

        process->start( cmd_array );

        result = (jlong)(size_t)process.take();
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }

    return result;
}

//-----------------------------------------------------------------------------Process_with_pid.pid

extern "C"
JNIEXPORT jint JNICALL Java_sos_hostware_Process_1with_1pid_pid( JNIEnv* jenv, jobject, jlong process_ptr )
{
    jint result = 0;

    try
    {
        Process* process = (Process*)(size_t)process_ptr;
        if( !process )  { set_NullPointerException( jenv, "hostjava" );  return result; }

        result = process->pid();
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }

    return result;
}

//----------------------------------------------------------------------Process_with_pid.terminated

extern "C"
JNIEXPORT jboolean JNICALL Java_sos_hostware_Process_1with_1pid_terminated( JNIEnv* jenv, jobject, jlong process_ptr )
{
    jboolean result = false;

    try
    {
        Process* process = (Process*)(size_t)process_ptr;
        if( !process )  { set_NullPointerException( jenv, "hostjava" );  return result; }

        result = process->terminated();
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }

    return result;
}

//-------------------------------------------------------------------------Process_with_pid.destroy

extern "C"
JNIEXPORT void JNICALL Java_sos_hostware_Process_1with_1pid_destroy( JNIEnv* jenv, jobject, jlong process_ptr )
{
    try
    {
        Process* process = (Process*)(size_t)process_ptr;
        if( !process )  { set_NullPointerException( jenv, "hostjava" );  return; }

        process->try_kill();
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }
}

//-------------------------------------------------------------------------Process_with_pid.waitFor

extern "C"
JNIEXPORT void JNICALL Java_sos_hostware_Process_1with_1pid_waitFor( JNIEnv* jenv, jobject, jlong process_ptr )
{
    try
    {
        Process* process = (Process*)(size_t)process_ptr;
        if( !process )  { set_NullPointerException( jenv, "hostjava" );  return; }

        process->wait();
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }
}

//-----------------------------------------------------------------------Process_with_pid.exitValue

extern "C"
JNIEXPORT jint JNICALL Java_sos_hostware_Process_1with_1pid_exitValue( JNIEnv* jenv, jobject, jlong process_ptr )
{
    jint result = 0;

    try
    {
        Process* process = (Process*)(size_t)process_ptr;
        if( !process )  { set_NullPointerException( jenv, "hostjava" );  return result; }

        result = process->exit_code();
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }

    return result;
}

//------------------------------------------------------------------------Process_with_pid.kill_pid

extern "C"
JNIEXPORT void JNICALL Java_sos_hostware_Process_1with_1pid_kill_1pid( JNIEnv* jenv, jclass, jint pid )
{
    try
    {
        kill_process_immediately( pid );
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }
}

//--------------------------------------------------------------------Process_with_pid.try_kill_pid

extern "C"
JNIEXPORT jboolean JNICALL Java_sos_hostware_Process_1with_1pid_try_1kill_1pid( JNIEnv* jenv, jclass, jint pid )
{
    jboolean result = false;

    try
    {
        result = try_kill_process_immediately( pid );
    }
    catch( const exception&  x ) { set_java_exception( jenv, x ); }
    catch( const _com_error& x ) { set_java_exception( jenv, x ); }

    return result;
}

//--------------------------------------------------------------------------------------------const

namespace sos {
namespace hostjava {

const static JNINativeMethod native_methods[] = 
{
    // sos.hostware.Process_with_pid
    { "close"                   , "(J)V"                    , (void*)Java_sos_hostware_Process_1with_1pid_close },
    { "start"                   , "(Ljava/lang/String;)J"   , (void*)Java_sos_hostware_Process_1with_1pid_start },
    { "start_array"             , "([Ljava/lang/String;)J"  , (void*)Java_sos_hostware_Process_1with_1pid_start_1array },
    { "pid"                     , "(J)I"                    , (void*)Java_sos_hostware_Process_1with_1pid_pid },
    { "terminated"              , "(J)Z"                    , (void*)Java_sos_hostware_Process_1with_1pid_terminated },
    { "destroy"                 , "(J)V"                    , (void*)Java_sos_hostware_Process_1with_1pid_destroy },
    { "waitFor"                 , "(J)I"                    , (void*)Java_sos_hostware_Process_1with_1pid_waitFor },
    { "exitValue"               , "(J)I"                    , (void*)Java_sos_hostware_Process_1with_1pid_exitValue },
    { "kill_pid"                , "(I)V"                    , (void*)Java_sos_hostware_Process_1with_1pid_kill_1pid },
    { "try_kill_pid"            , "(I)Z"                    , (void*)Java_sos_hostware_Process_1with_1pid_try_1kill_1pid },
};

//------------------------------------------------------------------------------------init_hostjava

void init_hostjava_process_with_pid( java::Vm* java_vm )
{
    register_natives( java_vm, "sos/hostware/Process_with_pid", native_methods, NO_OF( native_methods ) );
}

} //namespace hostjava
} //namespace sos

//-------------------------------------------------------------------------------------------------

