<?
# $Id: hostware_dynobj.inc 3483 2005-03-24 10:17:34Z jz $
# Joacim Zschimmer                /                  Software- und Organisations-Service GmbH, 2000

/**
* Klasse fr Datensatzobjekte
*
* Die Klasse beschreibt einen Datensatz mit mehreren Feldern.
* Die Objekte dieser Klasse mssen mit destruct() geschlossen werden.
*
* @copyright    Joacim, Zschimmer, SOS GmbH
* @author       Joacim Zschimmer <j@zsch.de>
* @since        1.0-2002/07/08
* @access       public
* @package      CONNECTION
*/

class Hostware_Dynobj
{
    /**
    * Konstruktor
    *
    * @param    string $handle 
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function Hostware_Dynobj( $handle )
    {
        $this->dynobj_handle = $handle;
    }


    /**
    * Destruktor
    *
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function destruct()
    {
        global $hostware_session_nr;

        if( $this->dynobj_handle )
        {
            hostware_set_session( $hostware_session_nr );
            hostware_dynobj__delete( $this->dynobj_handle );
            $this->dynobj_handle = 0;
        }
    }


    /**
    * Prfung auf leeren Datensatz
    *
    * @return   boolean liefert True, wenn der Datensatz leer ist, sonst False
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function is_empty()
    {
        return hostware_dynobj_empty( $this->dynobj_handle );
    }


    /**
    * Wert eines Felds im Datensatz liefern
    *
    * @param    string $name Name des Felds
    * @return   string liefert den Wert des Felds
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function field( $name )
    {
        return hostware_dynobj_field( $this->dynobj_handle, $name );
    }


    /**
    * Anzahl Felder im Datensatz liefern
    *
    * @return   integer liefert die Anzahl Felder
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function field_count()
    {
        return hostware_dynobj_field_count( $this->dynobj_handle );
    }


    /**
    * Namen eines Felds im Datensatz liefern
    *
    * @param    integer $nr lfd. Nr des Felds im Datensatz
    * @return   string liefert den Namen des Felds
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function field_name( $nr )
    {
        return hostware_dynobj_field_name( $this->dynobj_handle, $nr );
    }


    /**
    * Wert eines Felds im Datensatz setzen
    *
    *
    * @param    string   $name Feldname
    * @param    string   $value Wert
    * @return   boolean Fehlerzustand
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function set_field( $name, $value )
    {
        return hostware_dynobj_set_field( $this->dynobj_handle, $name, $value );
    }


    /**
    * Feld im Datensatz hinzufgen
    *
    *
    * @param    string   $name Feldname
    * @param    string   $type Feldtyp, zugelassene Werte sind ...
    * @return   boolean Fehlerzustand
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function add_field( $name, $type )
    {
        return hostware_dynobj_add_field( $this->dynobj_handle, $name, $type );
    }


    /**
    * Felder des Datensatzes als XML-Elemente liefern
    *
    * Die Felder werden in der Form <feld1>wert1</feld1> geliefert
    *
    * @return   string XML-String der Felder als Elemente
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function xml()
    {
        return hostware_dynobj_xml( $this->dynobj_handle, '' );
    }


    /**
    * Felder des Datensatzes aus XML-Elementen setzen
    *
    * Die Felder werden in der Form <feld1>wert1</feld1> aus dem Eingabestring gesetzt
    *
    * @param    string XML-String, der Felder als Elemente enthlt
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function set_xml( $xml_text )
    {
        hostware_dynobj_set_xml( $this->dynobj_handle, '', $xml_text );
    }


    var $dynobj_handle;
};

?>
