<?
# $Id: hostware_file.inc 3483 2005-03-24 10:17:34Z jz $
# Joacim Zschimmer                /                  Software- und Organisations-Service GmbH, 2000

#------------------------------------------------------------------------------hostware_session_end
#
#function hostware_session_end()
#{
#    global $hostware_session_nr;
#
#    if( isset( $hostware_session_nr ) )
#    {
#        hostware_session__delete( $hostware_session_nr );
#        unset( $hostware_session_nr );
#    }
#}
#
#register_shutdown_function( 'hostware_session_end' );
#

/** ID der aktuellen Sitzung */
$hostware_session_nr = hostware_session__new(); 

/**
* Klasse fr Dateien
*
* Die Klasse beschreibt eine offene Datei
* Die Objekte dieser Klasse mssen mit destruct() geschlossen werden.
*
* @copyright    Joacim, Zschimmer, SOS GmbH
* @author       Joacim Zschimmer <j@zsch.de>
* @since        1.0-2002/07/08
* @access       public
* @package      CONNECTION
*/

class Hostware_File
{
    /**
    * Konstruktor
    *
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */   

    function Hostware_File()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        $this->file_handle = hostware_file__new();
    }


    /**
    * Destruktor
    *
    * @access   public  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */
    
    function destruct()
    {
        global $hostware_session_nr;

        if( $this->file_handle )
        {
            hostware_set_session( $hostware_session_nr );
            hostware_file__delete( $this->file_handle );
            $this->file_handle = 0;
        }
    }


    /**
    * Destruktor
    *
    * @access   private  
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function destructor_Hostware_file()
    {
        global $hostware_session_nr;

        if( $this->file_handle )
        {
            hostware_set_session( $hostware_session_nr );
            hostware_file__delete( $this->file_handle );
        }
    }


    /**
    * Parsieren eines SQL-Statements
    *
    * @param    string $filename hostWare-Dateiname mit SQL-Statement
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function prepare( $filename )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_prepare( $this->file_handle, $filename );
    }


    /**
    * Zuweisen von Abfrageparametern an ein SQL-Statement
    *
    * @param    integer $number lfd. Nr. des Parameters (in der Reihenfolge der ? im SQL-Statement)
    * @param    string  $value Wert des Parameters
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function set_parameter( $number, $value )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_set_parameter( $this->file_handle, $number, $value );
    }


    /**
    * Datei ffnen
    *
    * @param    string $filename hostWare-Dateiname
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function open( $filename = '' )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_open( $this->file_handle, $filename );
    }


    /**
    * SQL-Statement ausfhren
    *
    * Das Statement muss zuvor mit prepare() erzeugt worden sein
    *
    * @see      prepare()
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function execute()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_execute( $this->file_handle );
    }


    /**
    * SQL-Cursor schlieen
    *
    * Das Result-Set einer SQL-Abfrage wird gelscht
    *
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function close_cursor()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_close_cursor( $this->file_handle );
    }
    

    /**
    * Datei schlieen
    *
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function close()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_close( $this->file_handle );
    }
    

    /**
    * Liefert den Zustand einer Datei
    *
    * @return   boolean True = geffnet, False = geschlossen
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function opened()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return hostware_file_opened( $this->file_handle );
    }


    /**
    * Leeren Datensatz erzeugen
    *
    * Liefert einen leeren Satz, dessen Felder anschlieend gefllt und mit put() geschrieben werden knnen
    *
    * @see      set_field(), put()
    * @return   object Hostware_dynobj Datensatzobjekt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function create_record()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return new Hostware_dynobj( hostware_file_create_record( $this->file_handle ) );
    }


    /**
    * Leeren Schlsselsatzobjekt erzeugen
    *
    * @return   object Hostware_dynobj Schlsselsatzobjekt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function create_key()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return new Hostware_dynobj( hostware_file_create_key( $this->file_handle ) );
    }


    /**
    * Zeile lesen
    *
    * @return   string Datensatz als String
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function get_line()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return hostware_file_get_line( $this->file_handle );
    }


    /**
    * Nchsten Datensatz lesen
    *
    * @return   object Hostware_dynobj Datensatz
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function get()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return new Hostware_dynobj( hostware_file_get( $this->file_handle ) );
    }


    /**
    * Prfung auf Dateiende
    *
    * @return   boolean Liefert True falls Dateiende erreicht, sonst False
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function eof()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return hostware_file_eof( $this->file_handle );
    }


    /**
    * Dateizeiger zurcksetzen
    *
    * Nach dem sequentiellen Lesen von Dateien wird der Zeiger auf Dateianfang eingestellt
    *
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function rewind()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_rewind( $this->file_handle );
    }


    /**
    * Zeitle bzw. Datensatz schreiben
    *
    * Fr satzorientierte Dateien wird ein Datensatzobjekt erwartet,
    * fr Textdateien oder Datenbanken wird ein String erwartet
    *
    * @param    object Hostware_dynobj $record Datensatzobjekt oder einfacher String 
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function put_line( $record )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_put_line( $this->file_handle, $record );
    }


    /**
    * Datensatz schreiben
    *
    * Die Methode schreibt das Datensatzobjekt in eine satzorientierte Datei
    *
    * @param    object Hostware_dynobj Datensatzobjekt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function put( $record )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        if( is_object( $record ) )  { hostware_file_put( $this->file_handle, $record->dynobj_handle ); }
                              else  { $this->put_line( $record ); }
    }


    /**
    * Dateizeiger in indexsequentiellen Dateien positionieren
    *
    * Der Dateizeiger wird auf den Satz mit dem angegebenen Schlssel eingestellt.
    * Der Schlssel kann ein Schlsselsatzobjekt (bei segmentierten Schlsseln) oder ein einfacher String sein
    *
    * @param    object Hostware_dynobj $key Schlsselsatzobjekt oder String
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function set_key( $key )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_set_key( $this->file_handle, $this->_get_key_param( $key ) );
    }


    /**
    * Satz aus indexsequentieller Datei entfernen
    *
    * Der Satz mit diesem Schlssel wird entfernt oder eine exception ausgelst, falls der Satz nicht vorhanden ist
    *
    * @param    object Hostware_dynobj $key Schlsselsatzobjekt oder String
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function delete_key( $key )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_delete_key( $this->file_handle, $this->_get_key_param( $key ) );
    }


    /**
    * Satz aus indexsequentieller Datei lesen
    *
    * liefert ein Datensatzobjekt oder lst eine exception aus, falls der Satz nicht gefunden wird
    *
    * @param    object Hostware_dynobj $key Schlsselsatzobjekt oder String
    * @return   object Hostware_dynobj Datensatzobjekt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function get_key( $key )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return new Hostware_dynobj( hostware_file_get_key( $this->file_handle, $this->_get_key_param( $key ) ) );
    }


    /**
    * Satz in eine indexsequentielle Datei oder Datenbank einfgen
    *
    * Schreibt einen Satz mit den Feldern des Datensatzobjekts.
    * Ist der Satz bereits vorhanden, wird eine exception ausgelst
    *
    * @param    object Hostware_dynobj $record Datensatzobjekt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function insert( $record )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_insert( $this->file_handle, $record->dynobj_handle );
    }


    /**
    * Satz in einer indexsequentiellen Datei oder Datenbank speichern
    *
    * Schreibt einen Satz mit den Feldern des Datensatzobjekts.
    * Ist der Satz bereits vorhanden, wird er berschrieben, sonst eingefgt
    *
    * @param    object Hostware_dynobj $record Datensatzobjekt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function store( $record )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_store( $this->file_handle, $record->dynobj_handle );
    }


    /**
    * Satz in eine indexsequentielle Datei oder Datenbank zurckschreiben
    *
    * Schreibt einen zuvor gelesenen Satz mit den Feldern des Datensatzobjekts.
    *
    * @param    object Hostware_dynobj $record Datensatzobjekt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function update( $record )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_update( $this->file_handle, $record->dynobj_handle );
    }


    /**
    * Satz in eine indexsequentielle Datei oder Datenbank zurckschreiben
    *
    * Schreibt einen zuvor gelesenen Satz mit den Feldern des Datensatzobjekts.
    * Bei SQL-Datenbanken wird das SQL-Statment nicht parsiert, sondern sofort ans DBMS geschickt
    *
    * @param    object Hostware_dynobj $record Datensatzobjekt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function update_direct( $record )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_update_direct( $this->file_handle, $record->dynobj_handle );
    }


    /**
    * Datumsformat fr die Darstellung in Feldern vereinbaren
    *
    * hostWare liest bzw. schreibt in Felder vom Datentyp date in der vereinbarten Reprsentation.
    * Beispiel: dd.mm.yyyy, dd-mon-yy
    *
    * @param    string $format Formatstring des Datums
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function set_date_format( $format )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_set_date_format( $this->file_handle, $format );
    }


    /**
    * Zeitformat fr die Darstellung in Feldern vereinbaren
    *
    * hostWare liest bzw. schreibt in Felder vom Datentyp timestamp in der vereinbarten Reprsentation.
    * Bei SQL-Datenbanken entspricht der Datentyp date einem timestamp
    * Beispiel: dd.mm.yyyy HH:MM:SS
    *
    * @param    string $format Formatstring des Zeitstempels
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function set_date_time_format( $format )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_set_date_time_format( $this->file_handle, $format );
    }


    /**
    * Dezimaldarstellung fr die Darstellung in Feldern vereinbaren
    *
    * hostWare liest bzw. schreibt in Felder vom Datentyp double bzw. float in der vereinbarten Reprsentation.
    *
    * @param    string $character Dezimalsymbol, z.B. der Punkt
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function set_decimal_symbol( $character )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_set_decimal_symbol( $this->file_handle, $character );
    }


    /**
    * Leere Felder des Datensatzes als Nullwerte in Datenbanken schreiben
    *
    * Ein leerer Wert (empty, null) wird als nichtiger SQL-NULL-Wert gespeichert
    *
    * @param    boolean $bool_value True, wenn leere Werte als SQL-Null gespeichert werden sollen
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function set_write_empty_as_null( $bool_value )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        hostware_file_set_write_empty_as_null( $this->file_handle, $bool_value );
    }


    /**
    * Anzahl Felder einer satzorientierten Datei oder Datenbanktabelle liefern
    *
    * @return   integer Anzahl Felder der Datei oder Datenbanktabelle
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function field_count()
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return hostware_file_field_count( $this->file_handle );
    }


    /**
    * Feldnamen einer satzorientierten Datei oder Datenbanktabelle liefern
    *
    * @param    integer lfd. Nr. des Felds im Datensatz
    * @return   string Name des Felds im Datensatz
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function field_name( $field_number )
    {
        global $hostware_session_nr;
        hostware_set_session( $hostware_session_nr );

        return hostware_file_field_name( $this->file_handle, $field_number );
    }


    /**
    * Schlsselsatz bzw. Schlsselwert liefern
    *
    * @param    object $key Schlsselsatz
    * @return   object Hostware_dynobj Handle fr Schlsselsatz
    * @access   private
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

    function _get_key_param( $key )
    {
        if( is_object( $key ) )  { return $key->dynobj_handle; }
                           else  { return $key; }
    }


    var $file_handle;
};

#-------------------------------------------------------------------------------hostware_get_single

    /**
    * Einen Satz einer Datei lesen
    *
    * Liefert die einzige Zeile der angegeben Datei als Hostware_dynobj.
    * Vorzugsweise fr SQL-Select.
    *
    * @param    string $filename hostWare-Dateiname
    * @return   object Datensatz
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

function hostware_get_single( $filename )
{
    global $hostware_session_nr;
    hostware_set_session( $hostware_session_nr );

    if( func_num_args() == 1 )
    {
        $dynobj_handle = hostware_get_single_( $filename );
    }
    else
    if( func_num_args() == 2 )
    {
        $param = func_get_arg(1);
        $dynobj_handle = hostware_get_single_( $filename, $param );
    }
    else
    {
        for( $i = 1; $i < func_num_args(); $i++ )  { $parameters[] = func_get_arg($i); }
        $dynobj_handle = hostware_get_single_( $filename, $parameters );
    }

    return new Hostware_dynobj( $dynobj_handle );
}

#-------------------------------------------------------------------------hostware_get_single_value

    /**
    * Ein Feld oder eine Zeile einer Datei lesen
    *
    * Liefert ein Feld oder eine Zeile als einfachen Wert (String, Integer, Double o..)
    *
    * @see      hostware_get_single()
    * @param    string $filename hostWare-Dateiname
    * @return   string Wert des Felds oder der Zeile
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

function hostware_get_single_value( $filename )
{
    global $hostware_session_nr;
    hostware_set_session( $hostware_session_nr );

    $parameters = array();
    for( $i = 1; $i < func_num_args(); $i++ )  { $parameters[] = func_get_arg($i); }
    
    $o = hostware_get_single( $filename, $parameters );

    if( !$o->is_empty() ) 
    { 
        $result = $o->field(0); 
        $o->destruct();
        return $result;
    }
    else
    {
        $o->destruct();
    }
}

#--------------------------------------------------------------------------------hostware_get_array

    /**
    * alle Stze einer Datei in einem Array von Datenstzen liefern
    *
    * Liefert ein Feld oder eine Zeile als einfachen Wert (String, Integer, Double o..)
    *
    * @see      hostware_get_single()
    * @param    string $filename hostWare-Dateiname
    * @return   array Jedes Array-Element enthlt einen Datensatz
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

function hostware_get_array( $filename )
{
    global $hostware_session_nr;
    hostware_set_session( $hostware_session_nr );

    $parameters = array();
    $array = array();
    
    for( $i = 1; $i < func_num_args(); $i++ )  { $parameters[] = func_get_arg($i); }
    $handle_array = hostware_get_array_( $filename, $parameters );

    for( $i = 0; $i < count($handle_array); $i++ )  { $array[$i] = new Hostware_dynobj( $handle_array[$i] ); }
    return $array;
}

#--------------------------------------------------------------------------hostware_get_array_value

    /**
    * Liefert alle Stze einer Datei in einem Array von einfachen Typen (String, Integer, Double o..)
    *
    * Die Datei liefert untypisierte Stze oder Stze mit genau einem Feld (z.B. ein SQL-Select).
    *
    * @see      hostware_get_array()
    * @param    string $filename hostWare-Dateiname
    * @return   array Jedes Array-Element enthlt einen einfachen Datentyp
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

function hostware_get_array_value( $filename )
{
    global $hostware_session_nr;
    hostware_set_session( $hostware_session_nr );

    $parameters = array();
    $array = array();

    for( $i = 1; $i < func_num_args(); $i++ )  { $parameters[] = func_get_arg($i); }
    $handle_array = hostware_get_array_( $filename, $parameters );

    for( $i = 0; $i < count($handle_array); $i++ )  
    { 
        $array[$i] = hostware_dynobj_field( $handle_array[$i], 0 ); 
        hostware_dynobj__delete( $handle_array[$i] );
    }

    return $array;
}

#-----------------------------------------------------------------------------------hostware_as_xml

    /**
    * Liefert das XML-Format eines Strings oder Datensatzes
    *
    * Bei Verwendung eines Datensatzes werden die Felder in der Form <feld1>wert1</feld1> geliefert.
    * Die Methode konvertiert Umlaute/Sonderzeichen xml-konform. 
    *
    * @param    string $value_or_dynobj einfacher String oder Datensatzobjekt
    * @return   string liefert den Eingabestring im XML-Format
    * @access   public
    * @author   Joacim Zschimmer <j@zsch.de>
    * @version  1.0-2002/07/09
    */

function hostware_as_xml( $value_or_dynobj )
{
    global $hostware_session_nr;
    hostware_set_session( $hostware_session_nr );

    if( gettype( $value_or_dynobj ) == 'object' )
    {
        return $value_or_dynobj->xml();
    }
    else
    {   
        $options = '';
        if( func_num_args() >= 2 )  { $options = func_get_arg(1); }
        return hostware_as_xml_( $value_or_dynobj, $options );
    }
}

?>
