#include "precomp.h"
//#define MODULE_NAME "sosmain0"
//#define COPYRIGHT   "1997 SOS GmbH Berlin"
//#define AUTHOR      "Joacim Zschimmer"

#include "../kram/sysdep.h"
#include "../kram/sysxcept.h"

#include "../kram/sosstrng.h"
#include "../kram/sos.h"
#include "../kram/sosappl.h"    // Sos_static
#include "../kram/sosprof.h"
#include "../kram/log.h"

#ifdef _MSC_VER
#   include "../zschimmer/com.h"
#   include "com_simple_standards.h"
#endif

using namespace std;
namespace sos {


extern int sos_main( int, char** );    // Wird von der Applikation implementiert.


int sos_main0( int argc, char** argv )
{
    Sos_appl appl ( false );   // init im try-Block
    int rc = 0;

    try {
        appl.init();

        for( int i = 1; i < argc; i++ )   // -sos.ini=FILENAME ?
        {
            const char* p = argv[ i ];

            if( strncmp( p, "-sos.ini=", 9 ) == 0 ) 
            {
                //if( !strchr( p + 9, ' ' ) )                     // Nicht bei "-sos.ini=filename bla ...", Blanks im Dateinamen nicht erlaubt
                if( p[9] == '"'  &&  p[ strlen( p ) - 1 ] == '"' )
                {
                    set_sos_ini_filename( string( p + 10, strlen( p + 10 ) - 1 ) );
                }
                else
                {
                    set_sos_ini_filename( p + 9 );
                }
            }
        }

        rc = sos_main( argc, argv );

        if( rc ) {
            //SHOW_ERR( "Programm-Rckgabewert: " << rc );
        }
    }

    catch( const Xc& x )
    {
        SHOW_ERR( "Fehler " << x );
        rc = 9999;
    }

    catch( const Xc_base& )
    {
        SHOW_ERR( "Fehler " /*<< x*/ );
        rc = 9999;
    }

    catch( const zschimmer::Xc& x )
    {
        SHOW_ERR( "Fehler " << x );
        rc = 9999;
    }

    catch( const exception& x )
    {
#       if defined __BORLANDC__
            SHOW_ERR( "Exception " << __throwExceptionName
                   << "( "         << x.why().c_str()
                   << " ) in "     << __throwFileName
                   << ", Zeile "   << __throwLineNumber );
//#    elif defined SYSTEM_SOLARIS
     //     SHOW_ERR( "Unbekannte Exception: " << x.why() );
#       else
            SHOW_ERR( "Exception exception(\"" << exception_text( x ) << "\")" );
#       endif

        rc = 9999;
    }

    catch( const char* text )
    {
#       if defined __BORLANDC__
            SHOW_ERR( "Exception " << __throwExceptionName
                   << text
                   << " in "       << __throwFileName
                   << ", Zeile "   << __throwLineNumber );
#        else
            SHOW_ERR( "Exception (char*) " << text );
#       endif
    }

#ifdef _MSC_VER
    catch( const _com_error& x )
    {
        SHOW_MSG( "Fehler HRESULT=" << as_hex_string( x.Error() ) << " source=" << bstr_as_string(x.Source()) << ", " << bstr_as_string(x.Description()) );
        rc = x.Error();
    }
#endif

#if !defined _MSC_VER       // Fr MSVC++ ist auch ein Absturz eine Exception
    catch( ... )
    {
#       if defined __BORLANDC__
            SHOW_ERR( "Exception " << __throwExceptionName
                      << " in "       << __throwFileName
                      << ", Zeile "   << __throwLineNumber );
#		 elif defined SYSTEM_GNU
            SHOW_ERR( "Unbekannte Exception" );
#        elif defined SYSTEM_SOLARIS
            //SHOW_ERR( "Unbekannte Exception: " << exception_name() );
            SHOW_ERR( "Unbekannte Exception: " );
#        endif
        rc = 9999;
    }
#endif

    return rc;
}

} //namespace sos
